package ZipTie::Adapters::CheckPoint::HTTPS;

use strict;
use warnings;
use 5.010_000;

use lib '/opt/qradar/lib/perl';
use JSON;

use ZipTie::Adapters::CheckPoint::HTTPS::Parsers qw(parse_system parse_chassis parse_interfaces parse_object_groups parse_filters parse_static_routing);
use ZipTie::Adapters::CheckPoint::HTTPS::Utils qw(create_connection get_server_by_ip get_server_by_name get_response_objects remote_run_script run_script check_response_for_expired_session decode_json_error_handler logout);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);

use ZipTie::ConnectionPath;
use ZipTie::Credentials;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::RESTAPI;

my $LOGGER = ZipTie::Logger::get_logger();

sub backup {
	my ($package, $backup_xml) = @_;

	# Turn the backup operation XML document into ZipTie::ConnectionPath
	my ($connection_path) = ZipTie::Typer::translate_document($backup_xml, 'connectionPath');

	my $mgmt_ip = $connection_path->get_management_server_ip_address();
	my $ip = $connection_path->get_ip_address();

	# Create RESTAPI object for communicating with the Management API
	my $connection = create_connection($connection_path);

	my $responses = get_responses($connection, $ip, $mgmt_ip);

	# Set up model printing object
	my $filehandle = get_model_filehandle('Check Point HTTPS', $connection_path->get_ip_address());
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, 'common');
	$printer->open_model();

	# Parse and print model
	parse_system($responses, $printer);
	parse_chassis($responses, $printer);
	parse_object_groups($responses, $printer);
	parse_filters($responses, $printer);
	parse_interfaces($responses, $printer);
	parse_static_routing($responses, $printer);

	# Log out of the API session
	logout($connection);

	$printer->close_model();
	close_model_filehandle($filehandle);

	return;
}

sub get_responses {
	my ($connection, $ip, $mgmt_ip) = @_;
	my $responses;

	$responses->{ip} = $ip;

	# Ordered list of API endpoint calls
	my @commands = (
		'show-gateways-and-servers',
		'show-hosts',
		'show-networks',
		'show-address-ranges',
		'show-groups',
		'show-groups-with-exclusion',
		'show-services-tcp',
		'show-services-udp',
		'show-service-groups',
		'show-packages'
	);

	# Call the API for each endpoint
	for my $command ( @commands ) {
		$responses->{$command} = get_response_objects($connection, $command);
	}

	# Callbacks for API queries we do not know the parameters of until parsing has started
	$responses->{'show-access-rulebase'} = sub {
		my $rulebase = shift;
		my $body = { name => $rulebase, limit => 500, 'details-level' => 'full', 'use-object-dictionary' => 'false' };

		my $response = get_response_objects($connection, "show-access-rulebase", $body);

		return $response;
	};

	$responses->{'show-nat-rulebase'} = sub {
		my $package = shift;
		my $body = { package => $package, limit => 500, 'details-level' => 'full', 'use-object-dictionary' => 'false' };

		my $response = get_response_objects($connection, "show-nat-rulebase", $body);

		return $response;
	};

	# Not all configuration is available directly from the mgmt API, such as the hostname
	# Before running any calls to run-script we need to find the target management server
	my $mgmt_server = get_server_by_ip($responses, $mgmt_ip);

	# If this is a cluster we want to execute the remote scripts on a cluster member IP
	my $this_server = get_server_by_ip($responses, $ip);
	my $type = $this_server->{type};
	my $cluster_uid;
	if( $type eq 'CpmiGatewayCluster' ) {
		my $cluster_members = $this_server->{'cluster-member-names'};
		if( $cluster_members ) { # TODO test cluster with one member
			for my $member_name ( @{ $cluster_members } ) {
				my $cluster_member = get_server_by_name($responses, $member_name);
				if( $cluster_member->{'ipv4-address'} ) {
					$ip = $cluster_member->{'ipv4-address'};
					$cluster_uid = $cluster_member->{uid};
					last;
				}
			}
		}
	}

	if( $mgmt_server->{version} gt 'R80' and $this_server->{version} ge 'R80.10' ) {
		my $uid = $cluster_uid // $this_server->{uid};
		$responses->{hostname} = run_script($connection, $uid, 'hostname');
		$responses->{ip_address} = run_script($connection, $uid, 'ip address');
		$responses->{route} = run_script($connection, $uid, 'route -n');
	} else {
		$responses->{hostname} = remote_run_script($connection, $mgmt_server, $ip, 'hostname');
		$responses->{ip_address} = remote_run_script($connection, $mgmt_server, $ip, 'ip address');
		$responses->{route} = remote_run_script($connection, $mgmt_server, $ip, 'route -n');
	}

	return $responses;
}

1;
