package ZipTie::Adapters::CheckPoint::HTTPS::Utils;

use strict;
use warnings;
use 5.010_000;

use lib '/opt/qradar/lib/perl';
use Data::Dumper;
use JSON;
use MIME::Base64;

use ZipTie::Logger;
use ZipTie::RESTAPI;

use Exporter qw(import);
our @EXPORT_OK = qw(create_connection get_server_by_ip get_server_by_name get_policy_by_name get_network_group_set get_protocol_for_service get_object_group_name_from_uid get_response_objects remote_run_script check_response_for_expired_session decode_json_error_handler run_script logout);

my $LOGGER = ZipTie::Logger::get_logger();

my $CONNECTION_PATH;
sub create_connection {
	my ($connection_path) = @_;

	# May need to recall this sub if session expires mid backup so store connection path
	$CONNECTION_PATH = $connection_path if $connection_path;
	$connection_path = $CONNECTION_PATH if not $connection_path;

	# Create RESTAPI object for communicating with the Management API
	my $connection = ZipTie::RESTAPI->new($connection_path, '/web_api/');

	# Set header declaring we are sending JSON
	$connection->{headers} = [ 'Content-Type' => 'application/json' ];

	# Authentication and get the Session ID
	my $credentials = $connection_path->get_credentials();
	my $authentication_body = {
		user => $credentials->{username},
		password => $credentials->{password},
		'session-timeout' => 900
	};
	$LOGGER->debug("Using username: $credentials->{username}");
	if( $credentials->{enableUsername} ) {
		$authentication_body->{domain} = $credentials->{enableUsername};
		$LOGGER->debug("Using domain  : $credentials->{enableUsername}");
	}
	$authentication_body = encode_json($authentication_body);

	# Store the Session ID in a header to be passed each consequent call
	my $sid = $connection->obtain_token_as_header(
		'login',                                # API endpoint
		$authentication_body,                   # POST body (encoded to JSON)
		\&_parse_sid,                           # Sub that parses SID from response
		'x-chkp-sid'                            # Header name for SID
	);

	return $connection;
}

# Parses Session ID from authentication request
sub _parse_sid {
	my ($json) = @_;
	$json = decode_json_error_handler($json, '_parse_sid');
	
	if( not $json->{sid} ) {
		$LOGGER->debug(Dumper($json));
		$LOGGER->fatal("Could not authenticate with SMS. Please check credentials.\nSMS error message: $json->{message}");
	}

	return $json->{sid}; 
}

# $in must have json encoded response of 'show-gateways-and-servers' to search
# $ip is the IP address of the managed firewall to return
sub get_server_by_ip {
	my ($in, $ip) = @_;
	my $servers = $in->{'show-gateways-and-servers'};

	for my $server ( @{ $servers } ) {
		if( $server->{'ipv4-address'} eq $ip ) {
			return $server;
		}
	}
	$LOGGER->debug("Could not retrieve server with IP $ip from 'show-gateways-and-servers'.");
	$LOGGER->fatal("Check this device was not discovered from the multi-domain server IP.\nDiscover from the domain management server virtual IP.");

	return;
}

# $in must have json encoded response of 'show-gateways-and-servers' to search
# $name is the name of the managed firewall to return
sub get_server_by_name {
	my ($in, $name) = @_;
	my $servers = $in->{'show-gateways-and-servers'};

	for my $server ( @{ $servers } ) {
		if( $server->{name} eq $name ) {
			return $server;
		}
	}
	$LOGGER->debug("Could not retrieve server with name $name from 'show-gateways-and-servers'.");
	$LOGGER->fatal("Check this device was not discovered from the multi-domain server IP.\nDiscover from the domain management server virtual IP.");

	return;
}

sub get_policy_by_name {
	my ($in, $name) = @_;
	my $packages = $in->{'show-packages'};

	for my $package ( @{ $packages } ) {
		if( $package->{name} eq $name ) {
			return $package;
		}
	}

	$LOGGER->debug("WARNING: Could not retrieve policy with name '$name' from 'show-packages'");
	return;
}

sub get_network_group_set {
        my ($name, $network_groups) = @_;
        my $set;

        if( exists $network_groups->{$name} ) {
                my %network_group = %{ $network_groups->{$name} };
                $set = Net::CIDR::Set->new();
                for my $key ( keys %network_group ) {
                        my $value = $network_group{$key};
                        if( $key eq 'objectGroupReference' ) {
                                for my $ref ( @{ $value } ) {
                                        my $nested_set = get_network_group_set($ref, $network_groups);
                                        if( $nested_set ) {
                                                $set->merge($nested_set);
                                        }
                                }
                        } elsif( $key eq 'host' ) {
                                for my $host ( @{ $value } ) {
                                        $set->add($host);
                                }
                        } elsif( $key eq 'subnet' ) {
                                for my $subnet ( @{ $value } ) {
                                        my $ip = $subnet->{address};
                                        my $mask = $subnet->{mask};
                                        $set->add("$ip/$mask");
                                }
                        } elsif( $key eq 'range' ) {
                                for my $range ( @{ $value } ) {
                                        my $start = $range->{startAddress};
                                        my $end = $range->{endAddress};
                                        $set->add("$start-$end");
                                }
                        }
                }
        }

        return $set;
}

sub get_protocol_for_service {
	my ($services, $id) = @_;
	my %services = %{ $services };

	for my $element ( 'portRange', 'portExpression' ) {
		if( $services{$id}->{$element} ) {
			for my $port_model ( @{ $services{$id}->{$element} } ) {
				return $port_model->{protocol};
			}
		} elsif( $services{$id . '__src'} and $services{$id . '__src'}->{$element} ) {
			for my $source_port_model ( @{ $services{$id . '__src'}->{$element} } ) {
				return $source_port_model->{protocol};
			}
		}
	}

	$LOGGER->debug("WARNING: Could not find protocol for service '$id'");
	return;
}

sub get_object_group_name_from_uid {
	my ($object_groups, $uid) = @_;

	for my $name ( keys %{ $object_groups } ) {
		if( $object_groups->{$name}->{description} eq $uid ) {
			return $name;
		} 
	}

	$LOGGER->debug("WARNING: Could not find object group with UID '$uid'");
	return;
}

# If there are more objects than can be displayed in one go (max 500)
# then call more than once and flatten into one array
sub get_response_objects {
	my ($connection, $command, $custom_body) = @_;

	my $name = $command;
	$name =~ s/-/ /g;
	my $body = $custom_body // { limit => 500, 'details-level' => 'full' };
	my $to = 0;
	my $total = 0;

	my @objects = ();
	do {
		GET_RESPONSE:
		my $response = $connection->post($name, $command, encode_json($body));
		$response = decode_json_error_handler($response, 'get_response_objects');

		# Session may expire, check and resend request if so
		my $new_connection = check_response_for_expired_session($response);
		if( $new_connection ) {
			$connection = $new_connection;
			goto GET_RESPONSE;
		}

		my $objects_ref = $response->{objects} // $response->{packages} // $response->{rulebase} // [];
		push @objects, @{ $objects_ref };
		$total = $response->{total};
		$to = $response->{to} // $total; # exit if 'to' is not set
		$body->{offset} = $to;
	} while( $to != $total );

	return \@objects;
}

# Workaround for running expert scripts on managed devices.
# Has three API calls to run-script:
# 1. execute the remote command and copy to a remote file
# 2. copy the remote file to the local management blade
# 3. read the local file contents
sub remote_run_script {
	my ($connection, $mgmt_server, $ip, $script) = @_;

	my $dir  = '/home/admin/';
	my $file = $dir . "QRM-$ip.txt";
	my $mgmt_name = $mgmt_server->{name};
	my $mgmt_uid = $mgmt_server->{uid};

	# Need to know if we are in a multi-domain environment to know to send extra
	# commands to switch to the domain context before sending cprid_util commands.
	# 'SMC User' is the domain name when not in a multi-domain deployment.
	my $switch_to_domain = '';
	if( $mgmt_server->{domain}->{name} ne 'SMC User' ) {
		$switch_to_domain = "source /etc/profile.d/CP.sh; mdsenv $mgmt_name; ";
	}

	my $remote_script = $switch_to_domain . "cprid_util rexec -verbose -server $ip -rcmd $script";
	my $response = run_script($connection, $mgmt_uid, $remote_script);

	chomp $response;

	return $response;
}

# Convience method for API calls to run-script
sub run_script {
	my ($connection, $target, $script) = @_;

	# Create script body and convert to json
	my $body = {
		'script-name' => "QRM - $script",
		'script' => $script,
		'targets' => [ $target ]
	};
	my $body_json = encode_json($body);

	# Run script and convert reponse to json
	RUN_SCRIPT:
	my $response = $connection->post($script, 'run-script', $body_json);
	$response = decode_json_error_handler($response, 'run_script');

	# Session may expire, check and resend request if so
	my $new_connection = check_response_for_expired_session($response);
	if( $new_connection ) {
		$connection = $new_connection;
		goto RUN_SCRIPT;
	}

	# SMS user needs access to run one time scripts, most likely to appear when read only user is used
	if( not $response->{tasks} ) {
		$LOGGER->debug("ERROR running one time script.");
                $LOGGER->fatal("Failed to Run One Time Script. Possible user configuration issue.\nPlease check user has permission to 'Run One Time Scripts'.\nIf the management server is on R80.10 or above and the gateway is under R80.10 a profile with Super User permissions is required.");
	}

	# Wait for the script to finish
	my @tasks = @{ $response->{tasks} };
	my $task = $tasks[0];
	$response = _wait_for_task($connection, $task->{'task-id'}, $script);

	return $response;
}

# Waits for a task on the management blade to complete then returns its output
sub _wait_for_task {
	my ($connection, $id, $script) = @_;
	my $body = {
		'details-level' => 'full',
		'task-id' => $id
	};
	my $body_json = encode_json($body);

	# Request task details until it completes
	my $task;
	do {
		sleep 3; # Give task a chance to complete

		SHOW_TASK:
		my $response = $connection->post("Wait for task '$id'", 'show-task', $body_json);
		$response = decode_json_error_handler($response, '_wait_for_task');

		# Session may expire, check and resend request if so
		my $new_connection = check_response_for_expired_session($response);
		if( $new_connection ) {
			$connection = $new_connection;
			goto SHOW_TASK;
		}

		my @tasks = @{ $response->{tasks} };
		$task = $tasks[0];

		if( $task->{status} =~ /failed/ ) {
			$LOGGER->fatal("Failed to execute remote script: $script\n\nPlease verify gateway is connected to the Management Server");
		}
	} while( $task->{status} !~ /succeeded/ );

	# Get the base64 encoded response of this task's script
	my $response = $task->{'task-details'}[0]->{responseMessage};
	$response = decode_base64($response);

	return $response;
}

sub decode_json_error_handler {
	my ($json, $subroutine) = @_;
	
	eval {
		$json = decode_json($json);
	};
	
	if ($@) {
		$LOGGER->debug("ERROR decoding JSON response in $subroutine:\n$@");
		$LOGGER->fatal("Failed to read API response, Possible device connection issue. Please check credentials and protocols.");
	}

	return $json;
}

sub check_response_for_expired_session {
	my ($response) = @_;
	my $connection;

	if( $response->{message} and $response->{message} =~ /Session may be expired/ ) {
		$LOGGER->debug('Session has expired, creating a new one.');
		$connection = create_connection();
	}

	return $connection;
}

sub logout {
	my ($connection) = @_;
	$connection->post('logout', 'logout', '{}');
	return;
}

1;
