package ZipTie::Adapters::Cisco::IOS;

use strict;
use warnings;
use 5.010_000;

use ZipTie::Adapters::BaseAdapter;
use ZipTie::Adapters::Cisco::IOS::AutoLogin;
use ZipTie::Adapters::Cisco::IOS::GetStartupConfig qw(get_startup_config);
use ZipTie::Adapters::Cisco::IOS::Parsers qw(parse_system parse_chassis create_config parse_static_routes parse_vlans parse_interfaces);
use ZipTie::Adapters::Cisco::IOS::Parsers qw(print_interfaces parse_filters parse_object_groups parse_dynamic_routes);
use ZipTie::Adapters::Cisco::IOS::Disconnect qw(disconnect);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::CLIProtocolFactory;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

use Exporter qw(import);
our @EXPORT_OK = qw(clean_responses);

my $LOGGER = ZipTie::Logger::get_logger();

# Specifies that this adapter is a subclass of ZipTie::Adapters::BaseAdapter
our @ISA = qw(ZipTie::Adapters::BaseAdapter);

sub backup {
	my $package_name = shift;

	# Retrieve the operation XML document that contains all of the IP, protocol, credential, and file server information
	# that is needed to successfully backup a device.
	my $backup_doc = shift;
	my $responses  = {};

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ($connection_path) = ZipTie::Typer::translate_document($backup_doc, 'connectionPath');

	# Connect to the device and capture the ZipTie::CLIProtocol that is created as a result of the connection.
	# Also be sure to capture the device prompt that is returned as a result of a successful connection.
	my ($cli_protocol, $enable_prompt_regex) = _connect($connection_path);

	# Attempt to disable pagination
	my $termLen = $cli_protocol->send_and_wait_for('terminal length 0', $enable_prompt_regex);

	# Grab an output filehandle for the model
	my $filehandle = get_model_filehandle('IOS', $connection_path->get_ip_address());

	# initialize the model printer
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, 'cisco', 'http://www.ziptie.org/model/cisco/1.0 cisco.xsd');
	$printer->open_model();

	$responses = get_command_responses($cli_protocol, $enable_prompt_regex, $connection_path);
	clean_responses ($responses);
	
	parse_system($responses, $printer);
	parse_chassis($responses, $printer);
	create_config($responses, $printer);
	parse_interfaces($responses);
	parse_dynamic_routes($responses, $printer);
	parse_object_groups($responses, $printer);
	parse_filters($responses, $printer);
	print_interfaces($responses, $printer);
	parse_static_routes($responses, $printer);
	parse_vlans($responses, $printer);

	$printer->close_model;

	# Close the model output file handle
	close_model_filehandle($filehandle);
}

sub _connect {
	my $connection_path = shift;

	# Create a new CLI protocol object
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	my $enable_prompt_regex = ZipTie::Adapters::Cisco::IOS::AutoLogin::execute($cli_protocol, $connection_path);

	# Store the found prompt as "enablePrompt" on the specified CLI protocol.
	$cli_protocol->set_prompt_by_name('enablePrompt', $enable_prompt_regex);

	# Return the created ZipTie::CLIProtocol object and the enable prompt encountered after successfully connecting to a device.
	return ($cli_protocol, $enable_prompt_regex);
}

sub get_command_responses {
	my ($cli_protocol, $enable_prompt_regex, $connection_path) = @_;
	
	my $responses;

	my %commands = (
		version        => 'show version',
		interfaces     => 'show interfaces',
		access_lists   => 'show access-lists',
		standby        => 'show standby',
		ip_route       => 'show ip route | exclude ^B',
		ip_route_bgp   => 'show ip route bgp | include 0.0.0.0/0',
		# following only supported on IOS 12.4+
		object_group   => 'show object-group',
		vlan           => 'show vlan',
	);

	for my $key ( keys %commands ) {
		my $command = $commands{$key};
		$responses->{$key} = $cli_protocol->send_and_wait_for($command, $enable_prompt_regex);
	}

	$responses->{startup_config} = get_startup_config($cli_protocol, $connection_path);

	disconnect($cli_protocol);
	
	return $responses;
}

sub clean_responses {
	my $responses = shift;

	foreach my $command ( keys %$responses ) {
		$responses->{$command} =~ s/[\b]\s?//mig;
		$responses->{$command} =~ s/^\s--More--(\s+)?//mg;
	}
}

1;
