package ZipTie::Adapters::Cisco::IOS::Parsers;

use strict;
use warnings;
use 5.010_000;

use Data::Dumper;
$Data::Dumper::Terse = 1; # Do not print $VAR1 before each dumped variable
$Data::Dumper::Sortkeys = 1; # Sort hash keys in dumped hashes

use MIME::Base64 qw(encode_base64);

use ZipTie::Adapters::Cisco::IOS::AutoLogin;
use ZipTie::Adapters::Cisco::IOS::AddressTranslationParser qw(parseAddressTranslation);
use ZipTie::Adapters::Cisco::IOS::Utils qw(get_full_interface_name);
use ZipTie::Adapters::Utils qw(mask_to_bits strip_mac create_network_element seconds_since_epoch get_mask get_port_number trim get_interface_type get_crep net_addr_to_mask is_contiguous_mask get_any_address_encoding invert_mask);
use ZipTie::Addressing::Subnet;
use ZipTie::Addressing::RouteFilter;
use ZipTie::Logger;
use ZipTie::RemoteGatewayResolution;

my $LOGGER = ZipTie::Logger::get_logger();

use Exporter qw(import);
our @EXPORT_OK = qw(parse_system create_config parse_chassis parse_interfaces print_interfaces parse_dynamic_routes parse_static_routes parse_vlans parse_filters parse_object_groups);

my $REMOTE_GATEWAY_RESOLVER = ZipTie::RemoteGatewayResolution->new();

my $IP_REGEX = get_crep('cipm');

sub parse_system {
	my ($in, $out) = @_;

	my ($system_name) = $in->{startup_config} =~ /^hostname (\S+)/m;
	$out->print_element( 'core:systemName', $system_name );

	$out->open_element('core:osInfo');
	if( $in->{version} =~ /^System image file is "([^"]+)/m ) {
		$out->print_element( 'core:fileName', $1 );
	}

	$out->print_element( 'core:make', 'Cisco' );
	if( $in->{version} =~ /^(IOS.+?),/m ) {
		$out->print_element( 'core:name', $1 );
	}

	if( $in->{version} =~ /^(?:Cisco\s+)?IOS.+Version\s+(\S[^\s,]+)/mi ) {
		$out->print_element( 'core:version', $1 );
	} elsif( $in->{version} =~ /^Version\s+V(\d+\.\d+\.\d+)/mi ) {
		$out->print_element( 'core:version', $1 );
	}

	$out->print_element( 'core:osType', 'IOS' );

	# Call to generate the manifest of phases that the device uses sequentially on a given packet.
	# This gives the calling application a hint as to how to simulate device operation.
	_generate_packet_processing_declaration($out);

	$out->close_element('core:osInfo');

	if( $in->{version} =~ /ROM:\s+System\s+Bootstrap[,]\s+Version\s+([^\s,]+)/i
	or  $in->{version} =~ /ROM:\s+Bootstrap\s+program\s+is\s+([\S ]+)/i
	or  $in->{version} =~ /BOOTLDR:\s+\S+\s+Boot\s+Loader.*?Version\s+([^\s,]+)/i
	or  $in->{version} =~ /^ROM:\s+TinyROM\s+version\s+([^\s]+)/mi
	or  $in->{version} =~ /^ROM:\s+([^\s]+)/mi ) {
		$out->print_element( 'core:biosVersion', $1 );
	}

	if( $in->{version} =~ /\b(cat|(WS|ME)-C)\d{4}|catalyst|CIGESM/i ) {
		$out->print_element( 'core:deviceType', 'Switch' );
	} elsif( $in->{version} =~ /\bC1200\b|\bAIR/ ) {
		$out->print_element( 'core:deviceType', 'Wireless Access Point' );
	} else {
		$out->print_element( 'core:deviceType', 'Router' );
	}

	return;
}

sub create_config {
	my ($in, $out) = @_;

	my $repository;
	$repository->{'core:name'} = "/";

	if( defined $in->{startup_config} ) {
		my $startup;
		$startup->{'core:name'}       = 'startup-config';
		$startup->{'core:textBlob'}   = encode_base64( _clean_config_text( $in->{startup_config} ) );
		$startup->{'core:mediaType'}  = 'text/plain';
		$startup->{'core:context'}    = 'boot';
		$startup->{'core:promotable'} = 'true';

		push @{ $repository->{'core:config'} }, $startup;
	}

	$out->print_element('core:configRepository', $repository);

	return;
}

sub parse_chassis {
	my ($in, $out) = @_;

	my $serial_number;
	my $model;
	for my $line ( split /\n/, $in->{version} ) {
		if( $line =~ /^IOS\s+\(tm\)\s+c(16|17|26|29)\d\d(XL)?\s+Software.*/i ) {
			$model = $1;
		}

		if( defined $model and $model == 29 ) {
			if( $line =~ /System\s+serial\s+number\s*:\s+(\S+)/i ) {
				$serial_number = $1;
			}
		} elsif( $line =~ /^Processor\s+board\s+ID\s+(\S+(\s+\(\S+\))??)/i ) {
			$serial_number = $1;
		}
	}

	my $model_number;
	if( $in->{version} =~ /^(?:Product\/)?Model number\s*:\s*(\S+)/mi
	or  $in->{version} =~ /^cisco\s+((?:WS-C|Cat|AS|C|VG)?\d{3,4}\S*\b)\S*\b/mi
	or  $in->{version} =~ /^cisco\s+((?:SOHO)?\d{2}\S*\b)\S*\b/mi
	or  $in->{version} =~ /^cisco\s+((?:uBR)?\d{3}\S*\b)\S*\b/mi
	or  $in->{version} =~ /^cisco\s+(\S+)(?:\s\(\S+\))?\s*processor.+?(?=with)/mi
	or  $in->{version} =~ /^cisco\s+(\S+)\s.+?Voice\sLinecard.+?\sprocessor(?:\s\(.+?\))? with/mi
	or  $in->{version} =~ /^cisco\s+catalyst\s+(\d{3,4})\s/mi
	or  $in->{version} =~ /^cisco\s+CISCO(\d+)\s+\(R\d+\)\s+processor\s+\(revision/mi ) {
		$model_number = $1;
	} elsif( $in->{version} =~ /^cisco CISCO(\S+)\s\(revision/mi ) {
		$model_number = "CISCO$1"
	}

	my $chassis = {
		'core:asset' => {
			'core:assetType' => 'Chassis',
			'core:factoryinfo' => {
				'core:make' => 'Cisco'
			}
		}
	};
	$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = $serial_number if $serial_number;
	$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $model_number if $model_number;

	my $description;
	if( $in->{version} =~ /(^Cisco\s+(?:Internetwork\s+Operating\s+System|IOS)\s+Software.+^Compiled.+?$)/msi ) {
		$description = $1;
	} elsif( $in->{version} =~ /(Cisco Catalyst .*? Enterprise Edition Software.+Copyright.+?$)/msi ) {
		$description = $1;
	}
	$chassis->{'core:description'} = $description if $description;

	if( $in->{version} =~ /Ethernet\s+(MAC)?\s*Address:\s+(\S+)/msi ) {
		my $mac = $2;
		$mac =~ s/[^\da-z]//gi;
		$chassis->{macAddress} = $mac;
	}

	$out->print_element( 'chassis', $chassis );

	return;
}

sub _generate_packet_processing_declaration {
	my $out = shift;

	$out->open_element("core:packetProcessingOrdering");
	for my $full_phase (
		'layer2-bridging',
		'packet-filtering/source',
		'routing',
		'address-translation',
		'packet-filtering/destination' ) {

		my ($phase, $position) = split '/', $full_phase;
		$out->open_element('core:packetProcessingElement');
		$out->print_element('core:packetProcessingPhase', $phase);
		if( defined $position ) {
			$out->print_element( 'core:processingPoint', $position);
		}		
		$out->close_element( 'core:packetProcessingElement');
	}
	$out->close_element("core:packetProcessingOrdering");

	return;
}

sub parse_dynamic_routes {
	my ($in, $out) = @_;
	if( not $in->{dynamic_routes} ) {

		# Concat the ip route response for BGP routes to 0.0.0.0
		$in->{ip_route} .= $in->{ip_route_bgp};
		_parse_route_data($in);
	}

	if( $in->{dynamic_routes} ) {
		$out->print_element('dynamicRoutes', { dynamicRoute => $in->{dynamic_routes} } );
	}

	return;
}

sub parse_static_routes {
	my ($in, $out) = @_;

	if( not $in->{static_routes} ) {
		_parse_route_data($in);
	}

	if( $in->{static_routes} ) {

		# Append the crypto map routes
		if( $in->{_crypto_map_routes} ) {
			push @{ $in->{static_routes } }, @{ $in->{_crypto_map_routes} };
		}

		$out->print_element('staticRoutes', { staticRoute => $in->{static_routes } });
	}

	return;
}

sub _parse_route_data {
	my ($in)= @_;
	my %protocol_codes = ( D => 'eigrp', O => 'ospf', R => 'rip', B => 'bgp', i => 'is-is' );

	# Parse the default route, it can be defined in a few places
	_parse_last_resort_route($in);

	# Build regex string for valid interface names
	my $interface_names = join '|', @{ $in->{interface_names} };
	$interface_names .= '|Null0';    # name for discard interface
	$interface_names =~ s/\//\\\//g; # escape /
	$interface_names =~ s/\./\\./g;  # escape .

	# Remove non-ascii characters, can get utf8 no-break space
	$in->{ip_route} =~ s/[^[:ascii:]]//g;

	while( $in->{ip_route} =~ /
		^(
		(\S\*?)\s*\S{0,2}\s+     # Protocol code (secondary code optional)
		($IP_REGEX)(\/(\d+))?\s+ # Destination address (mask optional)
		(.+?)?                   # Capture group for multiple next hops (optional)
		)(?=^\S)/xsmg
	) {
		my ($raw_route, $code, $address, $mask, $next_hops) = ($1, $2, $3, $5 // 32, $6);

		# There can be multiple next hops
		my @prefs;
		my @gateways;
		my @interfaces;
		while( $next_hops =~ /(\[(\d+)\/\d+\]\s+via\s+($IP_REGEX))?(.+?\s+($interface_names)\s+)?/g ) {
			my ($pref, $gateway, $interface) = ($2, $3, $5);
			push @prefs, $pref if $pref;
			push @gateways, $gateway if $gateway;
			push @interfaces, $interface if $interface;
		}

		$LOGGER->verbose("Parsing route: $raw_route");

		my $other_details = {};

		my $is_default = $code =~ s/\*$//;

		# Static route
		if( $code eq 'S' ) {
			if( @interfaces and @interfaces ~~ /Null\d+/i ) {
				$other_details->{discard} = 'true';
			}

		# Dynamic route
		} else {
			my $protocol = $protocol_codes{$code};
			if( not $protocol ) {
				if( $code ne 'C' and $code ne 'L' ) {
					$LOGGER->debug("WARNING: Dropping route '$raw_route', as protocol code '$code' is not supported.");
				}
				next;
			}

			$other_details = { dynamicRoute => 1, protocol => $protocol };
		}

		for( my $i = 0; $i < scalar @gateways or $i < scalar @interfaces; ++$i ) {
			$REMOTE_GATEWAY_RESOLVER->add_route($address, $mask, $gateways[$i], $prefs[$i], $is_default, $other_details);
		}
	}

	# Resolve all remote next hop addresses
	$REMOTE_GATEWAY_RESOLVER->resolve_remote_gateways();

	# Get the resolved routes and store for printing
	my ($static_routes, $dynamic_routes) = $REMOTE_GATEWAY_RESOLVER->get_routes();
	$in->{static_routes} = $static_routes;
	$in->{dynamic_routes} = $dynamic_routes;

	return;
}

sub _parse_last_resort_route {
	my ($in) = @_;
	my $subnets = $in->{interface_mapper};

	my $PREF = 200;
	my $IS_DEFAULT = 'true';

	my $default_network;
	if( $in->{startup_config} =~ /^ip\s+default-network\s+($IP_REGEX)/im ) {
		$default_network = $1;
	}

	# Now look for a static route to the default net
	if( $default_network and $in->{startup_config} =~ /^ip\s+route\s+$default_network\s+($IP_REGEX|\d+)\s+(\S+)/mi) {
		my ($mask, $gateway) = ($1, $2);
		$REMOTE_GATEWAY_RESOLVER->add_route($default_network, mask_to_bits($mask), $gateway, $PREF, $IS_DEFAULT);

	# If not that, look for ip default-gateway, which should only be used in the absence of routing
	# Failing that, look for a route to 0.0.0.0.  This is the same semantically as ip default-network, but
	# has some limitations, such as the fact that IGRP/EIGRP will only route to or propagate it under certain
	# circumstances.
	# Failing that, look in "show ip route" to handle an injected default route, from BGP for example.
	} elsif( $in->{startup_config} =~ /^ip\s+default-gateway\s+($IP_REGEX)\s*$/im
	     or  $in->{startup_config} =~ /^ip\s+route\s+0\.0\.0\.0\s+\d+\.\S+\s+(\d+\.\S+)/im
	     or  $in->{ip_route} =~ /Gateway\sof\slast\sresort\sis\s($IP_REGEX)\sto\snetwork\s0\.0\.0\.0/i ) {
		my $gateway = $1;
		$REMOTE_GATEWAY_RESOLVER->add_route('0.0.0.0', 0, $gateway, $PREF, $IS_DEFAULT);
	}

	return;
}

sub parse_vlans {
	my ($in, $out) = @_;

	if( $in->{vlan} =~ /^VLAN\s+Name\s+Status\s+Ports(.+?)^\s*$/msi ) {

		# create a variable with only section 2 data
		my $section2;
		while ( $in->{vlan} =~	/^VLAN\s+Type\s+SAID\s+MTU\s+Parent\s+RingNo\s+BridgeNo\s+Stp\s+BrdgMode\s+Trans1\s+Trans2(.+?)^\s*$/msig ) {
		    $section2 .= $1;
		}

		my $section1      = $1 . "\nEND";
		my $headerPrinted = 0;
		while( $section1 =~ /^(\d+)\s+(\S+)\s+(\S+)(.*?)(?=^\b)/msg ) {
			$out->open_element("vlans") if not $headerPrinted;
			$headerPrinted = 1;

			my $number = $1;
			my $name   = $2;
			my $status = $3;
			my $ports  = $4;

			my $vlan = { number => $number, name => $name, enabled => 'false' };
			if( $status =~ /^act/i ) {
				$vlan->{enabled} = 'true';
			}

			# Add any vlan pseudo interfaces to the correct vlan as an interfaceMember.
			if( defined $in->{interfacesData}->{"Vlan$number"} ) {
				push @{$vlan->{interfaceMember}}, "Vlan$number";
			}

			if ($ports) {
				for my $line ( split /\n/, $ports ) {
					for my $port ( split /,/, $line ) {
						# Make sure there is at least one non white space character in the port name
						if ($port =~ /\S/) {
							push @{ $vlan->{interfaceMember} }, get_full_interface_name($port);
						}
					}
				}
			}

		    if( $section2 =~ /^$number+\s+(\S+)\s+(\d+)\s+(\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+(\d+)\s+(\d+)/m ) {
			$vlan->{implementationType} = $1;
			$vlan->{said}               = $2;
			$vlan->{mtu}                = $3;
			$vlan->{parent}             = $4 if ( $4 ne '-' );
			$vlan->{ringNumber}         = $5 if ( $5 ne '-' );
			$vlan->{bridgeNumber}       = $6 if ( $6 ne '-' );
			$vlan->{bridgeMode}         = $7 if ( $7 ne '-' );
			$vlan->{translationBridge1} = $8;
			$vlan->{translationBridge2} = $9;
		    }

		    $out->print_element( 'vlan', $vlan );
		}
		$out->close_element('vlans') if $headerPrinted;
	} else {

		# MSFC output
		my $headerPrinted = 0;
		while ( $in->{vlan} =~ /Virtual LAN ID:\s+(\d+)\s+(.+?)(?=^\S)/msig ) {
			$out->open_element('vlans') if not $headerPrinted;
			$headerPrinted = 1;
			my $vlan = {
				number  => $1,
				name    => "vlan$1",
				enabled => 'true',
			};
			my $blob = $2;
			while( $blob =~ /VLAN Trunk Interface:\s+(\S+)/gi ) {
				push @{ $vlan->{interfaceMember} }, get_full_interface_name($1);
			}
			$out->print_element( 'vlan', $vlan );
		}
		$out->close_element('vlans') if $headerPrinted;
	}

	return;
}

sub parse_filters {
 	my ($in, $out) = @_;
	my $timeRanges = parse_time_range($in);
	my $lineNumber   = 0;
	my $processOrder = 1;  # starts at 1
	my $listSignatureAssembly = undef;
	my $currentList;
	my $natFilterListInstances;

	$out->open_element('filterLists');
	
	# Parse NAT.  This involves using the AddressTranslationParser to drive population of an AddressTranslation
	# instance allocated here.  We need to do a pass first, however, to detect where the zones map
	
	# Every combination of source and dest interface provide a separate filter list for the 
	# NAT section of the FilterLists element.  They are split into those specifying rules corresponding strictly
	# to source-address-mapped packet flows, and into destination-mapped address packet flows.  One of those is
	# going to be an implied byproduct of the NAT rule as found in the running-config.
	#
	if( not $in->{interfaceZones} ) {
		$in->{interfaceZones} = _parse_nat_interfaces($in);
	}
	
	# only define NAT if we see a hint it's present
	if( $in->{interfaceZones} ) {
		$LOGGER->debug('Parsing NAT');
		$natFilterListInstances = parseAddressTranslation( { in => $in, interfaces => $in->{interfaceZones},
						interface_mapper => \&_mapIfNameToPrimaryAddr, mapper_context => $in->{interface_mapper} } );
				
		if( $natFilterListInstances ) {		
			# this maps phase position for enclosing filterListsSection to a set of groups of
			#     - directional orientation of enclosed rules
			#     - the argument that should be used to call AddressTranslation::generateTranslationList with them
			 
			# +++wayne FIXME change this interface to have the second arg to generateTranslationList
			# be called "source" or "destination", since it's a completely overloaded mess

			my $generateRulePatterns = {
				before => [
					["source", "destination"],
					["dest", "source"]
				],
				after => [
					["source", "source"],
					["dest", "destination"]
				]
			};
				
			for my $phasePos ( keys %{$generateRulePatterns} ) {
				my $fListGenerated = 0;
				
				for my $rulesTuple ( @{$generateRulePatterns->{$phasePos}} ) {
					my ($rulesDirection, $genXlDirection) = @{$rulesTuple};
					
					for my $natFilterList ( @{$natFilterListInstances} ) {
						if( $natFilterList->getTranslationName() =~ /$rulesDirection/ ) {
							if( not $fListGenerated ) { 
								$out->open_element('filterListsSection');
								$fListGenerated = 1;
							}
							$natFilterList->resetForNewFiltersSection();
							$natFilterList->generateTranslationList($out, $genXlDirection);
						}
					}
					
				}
				if( $fListGenerated ) {
					$out->open_element('phaseDiscriminator');
					$out->print_element('processingPhase', { 'core:packetProcessingPhase' => "routing"});
					$out->print_element('processingPhasePosition', $phasePos);
					$out->close_element('phaseDiscriminator');
					$out->close_element('filterListsSection');					
				}
			}
		}
		# else, no 'ip nat' statements defined in running-config

		delete $in->{interfaceZones};
	}

	$out->open_element('filterListsSection');

	# hash of acls names to an ordered array of rule hits counts (aka matches)
	# $all_acl_matches->{my_acl}->[ 0, 10, 1234 ];
	my $all_acl_matches = _parse_rule_matches($in);
	my $acl_matches;

	for my $line ( split /\n/, $in->{startup_config} ) {
		$lineNumber++;

		if( $line =~ /^ip access-list\s+(extended|dynamic|standard)\s+(.+\b)/ ) {
		    if ($currentList) {
			$processOrder = 1;
			_close_list( $currentList, $out );
			undef $currentList;
			undef $acl_matches;
		    }
		    $currentList = {
			type     => $1,
			protocol => 'ip',
			name     => $2,
		    };
                    if( $all_acl_matches and $all_acl_matches->{$2} ) {
                        $acl_matches = $all_acl_matches->{$2};
                    }

		    $out->open_element('filterList configPath="running-config"');
		    $out->print_element("defaultAction", "deny");
		    
		    # Filter list's are set as being tracked with signature assembly
		    $listSignatureAssembly = $out->attach_signature_assembly('filterList', "ztSignature");
		} elsif( $line =~ /^ipv6 access-list\s+(\b.+\b)/ ) {
			$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping IPv6 access list: $1");
		} elsif( $line =~ /^access-list\s+(\d+)(\s+.+)/ ) {
		    my $aclNumber = $1;
		    my $aceBody   = $2;
		    if ( $currentList && $currentList->{name} ne $aclNumber ) {
			_close_list( $currentList, $out );
			undef $currentList;
			undef $acl_matches;
		    }

		    if ( !$currentList ) {
			$processOrder = 1;
			$currentList  = {
			    protocol => 'ip',
			    name     => $aclNumber,
			    type     => 'standard',
			};
                        if( $all_acl_matches and $all_acl_matches->{$aclNumber} ) {
                            $acl_matches = $all_acl_matches->{$aclNumber};
                        }
			$out->open_element('filterList configPath="running-config"');
			$out->print_element("defaultAction", "deny");
			$listSignatureAssembly = $out->attach_signature_assembly('filterList', "ztSignature");						
		    }
		    if ( $line !~ /\s+remark/i ) {
			if ( ($aclNumber < 100) || (($aclNumber >= 1300) && ($aclNumber <= 1399)) ) {
			    _standard_term( $aceBody, $processOrder, $lineNumber,
					    $timeRanges, $acl_matches, $out );
			}
			else { # technically, 100-199, 2600-2699
			    _extended_term( $aceBody, $processOrder, $lineNumber,
					    $timeRanges, $acl_matches, $out );
			}
			$processOrder++;
		    }
		}
		elsif ( $line =~ /^\S/ ) {
		    if ($currentList) {
			$processOrder = 1;
			_close_list( $currentList, $out );
		    }
		    undef $currentList;
                    undef $acl_matches;
		}
		elsif ( $currentList && $line !~ /^\s*remark/ ) {
			if( $currentList->{type} =~ /standard/i ) {
				_standard_term( $line, $processOrder, $lineNumber, $timeRanges, $acl_matches, $out );

			} elsif( $currentList->{type} =~ /extended/i ) {
				_extended_term( $line, $processOrder, $lineNumber, $timeRanges, $acl_matches, $out );

			} elsif( $currentList->{type} =~ /ipv6/i ) {
				$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping IPv6 rule: $line");
			}
			$processOrder++;
		}
	}
	$out->close_element('filterListsSection');
	
	# Now, we need to generate the filterLists section that is complete with the references for input
	# filter lists and output filter lists
	my $filterRefs; # indexed by ingress/egress hash ref
	my $phaseDiscrim;
	
	for my $fAssignIf ( keys %{$in->{filter_assignments}} ) {
		for my $assignedFilter (@{$in->{filter_assignments}->{$fAssignIf}}) {
			my $filterRefInst = { name => $assignedFilter->{filterName}, direction => $assignedFilter->{direction}};
			
			if ($filterRefInst->{'direction'} eq 'ingress') {
				push @{$filterRefInst->{appliesTo}}, { source => $fAssignIf};
			} else {
				push @{$filterRefInst->{appliesTo}}, { destination => $fAssignIf};
			}
			
			push @{$filterRefs->{$assignedFilter->{direction}}}, $filterRefInst;
		}	
	} 	
	
	# We should just be able to write the input/output filtering assignment sections in bulk now
	$out->open_element('filterListsSection');
	
	for my $filterRefInst ( @{$filterRefs->{ingress}} ) {
		$out->print_element('filterListReference', $filterRefInst);
	}
	
	my $source_phase = {
		processingPhase => {
			'core:packetProcessingPhase' => 'packet-filtering',
			'core:processingPoint' => 'source'
		}
	};
	$out->print_element('phaseDiscriminator', $source_phase);
	$out->close_element('filterListsSection');

	$out->open_element('filterListsSection');
	
	for my $filterRefInst ( @{$filterRefs->{egress}} ) {
		$out->print_element('filterListReference', $filterRefInst);
	}
	
	my $dest_phase = {
		processingPhase => {
			'core:packetProcessingPhase' => 'packet-filtering',
			'core:processingPoint' => 'destination'
		}
	};
	$out->print_element('phaseDiscriminator', $dest_phase);
	$out->close_element('filterListsSection');
	$out->close_element('filterLists');

	return;
}

sub _parse_rule_matches {
	my ($in) = @_;
	my $matches;

	return unless $in->{access_lists};

	while( $in->{access_lists} =~ /^(Standard|Extended)\s+IP\s+access\s+list\s+(\S+)\s*\n(.+?)(?=^\S)/gmis ) {
		my ($acl_name, $blob) = ($2, $3);
		for my $line ( split /\n/, $blob ) {
			my $count = 0;
			if( $line =~ /\((\d+)\s+matches\)\s*$/i ) {
				$count = $1;
			}
			push @{ $matches->{$acl_name} }, $count;
		}
	}

	return $matches;
}

sub _close_list {
	my ( $list, $out ) = @_;
	$out->print_element( 'mode', 'stateless' );
	# we don't generate the list name into the signature
	$out->print_element( 'name', $list->{name}, {suppressFromSignatureAssembly => 1});
	$out->close_element('filterList');
	return;
}

# process a standard term:
sub _standard_term {
	my ( $acl_line, $term_process_order, $lineNumber, $time_range, $acl_matches, $out ) = @_;

	my $thisterm = { processOrder => $term_process_order, };
	$thisterm->{'-attributes'} = { lineStart => $lineNumber, lineEnd => $lineNumber, };

	if( $acl_matches ) {
		my @acl_hits = @{ $acl_matches };
		my $hits = $acl_hits[$term_process_order - 1];
		$thisterm->{'-attributes'}->{hits} = $hits // 0;
	}

	if( $acl_line =~ /^\s*(no permit|permit|deny)\s*(tcp|udp)?\s+any/i ) {
		my $primaryAction = $1;
		if( $primaryAction eq 'no permit' ) {
			$primaryAction = 'deny';
		}
		$thisterm->{primaryAction} = $primaryAction;
		$thisterm->{protocol} = $2 if $2;

		push @{ $thisterm->{sourceIpAddr}->{network} }, get_any_address_encoding();

	} elsif( $acl_line =~ /^\s*(no permit|permit|deny)\s*(ip|tcp|udp)?\s+(\d+\.\d+\.\d+\.\d+)\s+(\d+\.\d+\.\d+\.\d+)/i ) {
		my $primaryAction = $1;
		if( $primaryAction eq 'no permit' ) {
			$primaryAction = 'deny';
		}
		$thisterm->{primaryAction} = $primaryAction;
		$thisterm->{protocol} = $2 if $2;

		push @{ $thisterm->{sourceIpAddr}->{network} }, create_network_element( $3, $4, 1 );

	} elsif( $acl_line =~ /^\s*(no permit|permit|deny)\s*(ip|tcp|udp)?\s+host\s+(\d+\.\d+\.\d+\.\d+)/i ) {
		my $primaryAction = $1;
		if ( $primaryAction eq 'no permit' ) {
			$primaryAction = 'deny';
		}
		$thisterm->{primaryAction} = $primaryAction;
		$thisterm->{protocol} = $2 if $2;
		my $thisIPAddress = { host => $3 };

		push @{ $thisterm->{sourceIpAddr} }, $thisIPAddress;

	} elsif( $acl_line =~ /^\s*(no permit|permit|deny)\s*(ip|tcp|udp)?\s+(\d+\.\d+\.\d+\.\d+),\s+wildcard\s+bits\s+(\d+\.\d+\.\d+\.\d+)/i ) {
		my $primaryAction = $1;
		if( $primaryAction eq 'no permit' ) {
			$primaryAction = 'deny';
		}
		$thisterm->{primaryAction} = $primaryAction;
		$thisterm->{protocol} = $2 if $2;
		push @{ $thisterm->{sourceIpAddr}->{network} }, create_network_element( $3, $4, 1 );
	} elsif( $acl_line =~ /^\s*(no permit|permit|deny)\s*(ip|tcp|udp)?\s+(\d+\.\d+\.\d+\.\d+)/i ) {
		my $primaryAction = $1;
		if( $primaryAction eq 'no permit' ) {
			$primaryAction = 'deny';
		}
		$thisterm->{primaryAction} = $primaryAction;
		$thisterm->{protocol} = $2 if $2;
		my $thisIPAddress = { host => $3, };
		push @{ $thisterm->{sourceIpAddr} }, $thisIPAddress;
	}

	$thisterm->{log} = 'false';
	if( $acl_line =~ /\s+log\s*$/ ) {
		$thisterm->{log} = 'true';
	}

	# look for the time-range instruction
	if( $acl_line =~ /\btime-range\s+(\S+)/i ) {
		$thisterm->{timeAllowed} = $time_range->{$1};
	}
	$out->print_element( 'filterEntry', $thisterm );
	return;
}

# Process an extended term
sub _extended_term {

	my ( $acl_line, $term_process_order, $lineNumber, $time_range, $acl_matches, $out) = @_;

	my $thisterm = {
		processOrder => $term_process_order,
		'-attributes' => { lineStart => $lineNumber, lineEnd => $lineNumber, },
	};

	my $ipaddresses = [];

	my $hits = 0;
        if( $acl_matches ) {
            my @acl_hits = @{ $acl_matches };
            $hits = $acl_hits[$term_process_order - 1] // 0;
        }
	$thisterm->{'-attributes'}->{hits} = $hits;

	# permit/deny and dynamic acl settings
	my $primaryAction;
	if ( $acl_line =~ /^\s*(no permit|deny|permit)/i ) {
		$primaryAction = $1;
	} elsif ( $acl_line =~ /^\s*dynamic\s+(\S+)\s+timeout\s+(\S+)\s+(no permit|deny|permit)/i ) {
		$primaryAction = $3;
	} elsif ( $acl_line =~ /^\s*dynamic\s+(\S+)\s+(no permit|deny|permit)/i ) {
		$primaryAction = $2;
	} else {
		return;
	}

	if ( $primaryAction eq 'no permit' ) {
		$primaryAction = 'deny';
	}
	$thisterm->{primaryAction} = $primaryAction;

	# Set term protocol. Keep in mind it can have been set as an object-group reference.
	my $termProtocol = "";
	my $ogSvcObjRef = undef;
	my $protoArgTxt = "";

	if ( $acl_line =~ /^\s*(no permit|deny|permit)\s+(object-group\s+(\S+))\s+/i ) {
		# Object group reference
		# This is the entire protocol term argument, including "object-group"
		$protoArgTxt = $2;
		$ogSvcObjRef = $3;
	} elsif ( $acl_line =~ /^\s*(no permit|deny|permit)\s+[0-9a-f]{4}\.[0-9a-f]{4}\.[0-9a-f]{4}\s+/i ) {
		$LOGGER->debug("WARNING: Layer 2 address matching not supported. Dropping rule at line $lineNumber in 'show startup-config'");
		return;
	} elsif ( $acl_line =~ /^\s*(no permit|deny|permit)\s+0x\d{4}\s+/i ) {
		$LOGGER->debug("WARNING: Ethernet type matching not supported. Dropping rule at line $lineNumber in 'show startup-config'");
		return;
	} elsif ( $acl_line =~ /^\s*(no permit|deny|permit)\s+(\S+)\s+/i ) {
		$termProtocol = $2;
		$protoArgTxt = $termProtocol;
	}

	# Choose source address format
	my $srctext = "";
	my $termSrcAddress = ();

	if ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+any)(?:\s+any)?\s+/i ) {
		$srctext = $1;
		push @{ $termSrcAddress->{network} }, get_any_address_encoding();
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+object-group\s+(\S+))/i )	{
		# This is a network object group reference
		$srctext = $1;
		push @{$termSrcAddress->{objectGroupReference}}, $3;
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+(addrgroup\s+.+))/i ) {
		# This is a network object group reference but in "addrgroup" form..there could be multiple object groups after it.
		$srctext = $1;
		my $agBlob = $3;

		# Multiple groups could follow
		while ($agBlob =~ /addrgroup\s+(\S+)/gi) {
			push @{$termSrcAddress->{objectGroupReference}}, $1;
		}
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+(\d+\.\d+\.\d+\.\d+)\s+(\d+\.\d+\.\d+\.\d+))\s+/i ) {
		$srctext = $1;
		push @{ $termSrcAddress->{network} },
		create_network_element( $3, $4, 1 );
	} elsif ($acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+([a-f\d:]+)\/(\d+))\s+/i ) {
		$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping network element creation: $acl_line");
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+(\d+\.\d+\.\d+\.\d+),\s+wildcard\s+bits\s+(\d+\.\d+\.\d+\.\d+))\s+/i ) {
		$srctext = $1;
		push @{ $termSrcAddress->{network} },
		create_network_element( $3, $4, 1 );
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+host\s+(\d+\.\d+\.\d+\.\d+|[a-f\d:]+))\s+/i ) {
		$srctext = $1;	
		push @{ $termSrcAddress->{host} }, $3;
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+(\d+\.\d+\.\d+\.\d+))\s+/i ) {
		$srctext = $1;
		push @{ $termSrcAddress->{host} }, $3;
	}

	# Now look for the src port
	my $srcport        = "";
	my $termSrcService = ();
	if ( $acl_line =~ /$srctext\s+((lt|gt|eq|neq)\s+(\S+))/i ) {
		$srcport = $1;

		my $operator = $2;
		if ( $operator =~ /neq/ ) {
			$operator = "ne";
		}

		# Note: this is the locus of port-map support, pulled in momentary short-term indecision
		# as to how to handle its support and testing (see PR 29487)
		my $snPort = $3;

		# If we didn't get a service mapping out of that, try it as a numeric or trivially "built in" port
		if ( _int_port($snPort) =~ /^\d+$/ ) {
			my $thisport = {
				port     => _int_port($snPort),
				operator => $operator,
			};
			push @{ $termSrcService->{portExpression} }, $thisport;
		} else {
			# We have to bypass mapping support for this port qualification
			$LOGGER->debug('Unable to parse or map presumed filter entry source port/service: ' . _int_port($snPort) );
		}

	} elsif ( $acl_line =~ /$srctext\s+(range\s+(\S+)[-\s](\S+))/i ) {
		$srcport = $1;
		my $thisport = {
			portStart => _int_port($2),
			portEnd   => _int_port($3),
		};
		push @{ $termSrcService->{portRange} }, $thisport;
	}

	# Destination address
	my $dsttext = "";
	my $termDestAddr = ();

	if( $acl_line =~ /$srctext\s*$srcport\s*(any(?:\s+any)?)/i ) {
		$dsttext = $1;
		push @{ $termDestAddr->{network} }, get_any_address_encoding();
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*(object-group\s+(\S+))/i ) {
		$dsttext = $1;
		push @{ $termDestAddr->{objectGroupReference}}, $2;
	} elsif( $acl_line =~ /$srctext\s*$srcport\s+(addrgroup\s+.+)/i ) {
		# This is a network object group reference but in "addrgroup" form..there could be multiple object groups after it.
		$dsttext = $1;

		# Multiple groups could follow
		while ($dsttext =~ /addrgroup\s+(\S+)/gi) {
			push @{$termDestAddr->{objectGroupReference}}, $1;
		}
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*((\d+\.\d+\.\d+\.\d+)\s+(\d+\.\d+\.\d+\.\d+))/i ) {
		$dsttext = $1;
		push @{ $termDestAddr->{network} }, create_network_element( $2, $3, 1 );
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*(([a-f\d:]+)\/(\d+))/i ) {
		$dsttext = $1;
		push @{ $termDestAddr->{network} }, create_network_element( $2, $3, 1 );
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*((\d+\.\d+\.\d+\.\d+),\s+wildcard\s+bits\s+(\d+\.\d+\.\d+\.\d+))/i ) {
		$dsttext = $1;
		push @{ $termDestAddr->{network} }, create_network_element( $2, $3, 1 );
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*(host\s+(\d+\.\d+\.\d+\.\d+|[a-f\d:]+))/i ) {
		$dsttext = $1;
		push @{$termDestAddr->{host}}, $2;
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*(\d+\.\d+\.\d+\.\d+|[a-f\d:]+)/i ) {
		$dsttext = $1;
		push @{$termDestAddr->{host}}, $1;
	}

	# now look for the dst port
	my $dstport            = "";
	my $termDestService    = ();
	my $termProtocolFilter = "";

	if ( $acl_line =~ /$srctext\s*$srcport\s*$dsttext\s+((lt|gt|eq|neq)\s+(\S+))/i ) {
		$dstport = $1;
		my $operator = $2;
		if ( $operator =~ /neq/ ) {
			$operator = "ne";
		}

		$ogSvcObjRef = undef;
		my $dnPort = $3;

		if ( _int_port($dnPort) =~ /^\d+$/) {
			my $thisport = {
				port     => _int_port($dnPort),
				operator => $operator,
			};
			push @{ $termDestService->{portExpression} }, $thisport;
		}

	} elsif ( $acl_line =~ /$srctext\s*$srcport\s*$dsttext\s+(range\s+(\S+)[-\s](\S+))/i ) {
		$dstport = $1;
		my $thisport = {
			portStart => _int_port($2),
			portEnd   => _int_port($3),
		};
		push @{ $termDestService->{portRange} }, $thisport;
	} elsif ( ( $termProtocol eq "icmp" ) && ( $acl_line =~ /$srctext\s*$srcport\s*$dsttext\s+((?:icmp-type\s+)?([\w-]+)\s*)/i ) ) {

		# This is either a numeric or symbolic (e.g., "echo", "timestamp-reply")
		# ICMP message type reference.  Now, this is a bit of a conundrum...we could
		# shove this in as a "port type", and the table for protocol -> numeric mapping
		# in Utils.pm that is consulted by get_port_number() would have you believe that
		# this is what you should do (e.g., there is "echo" entry there that maps to
		# 7).  On the other hand, the ONM schema has an option for "protocol-filter",
		# whose descriptive text says its use case is ICMP message types.  It is also
		# generously typed as a simple string.  We will use the latter.
		$termProtocolFilter = $2;
	}

	# Assign term fields in order of definition in the Ziptie Adapter schema
	if ($termProtocol) {
		$thisterm->{protocol} = $termProtocol;
	}

	# Initialize a filter term
	if ($termProtocolFilter) {
		$thisterm->{protocolFilter} = $termProtocolFilter;
	}

	# These should be mutually exclusive
	if( $ogSvcObjRef and not defined $thisterm->{protocol} ) {
		$thisterm->{protocolObjectGroupReference} = $ogSvcObjRef;
	}

	for my $addrTerm ( qw(host network objectGroupReference) ) {
		for (@{$termSrcAddress->{$addrTerm}}) {
			# Push the type of addr term instance being referenced as a separate sourceIpAddr instance.
			push ( @{$thisterm->{sourceIpAddr}}, { $addrTerm => $_} );
		}
		
		# Now do the same thing for dest
		for (@{$termDestAddr->{$addrTerm}}) {
			# Push the type of addr term instance being referenced as a separate destinationIpAddr instance.
			push ( @{$thisterm->{destinationIpAddr}}, { $addrTerm => $_} );
		}
	}

	if ($termSrcService) {
		$thisterm->{sourceService} = $termSrcService;
	}

	if ($termDestService) {
		$thisterm->{destinationService} = $termDestService;
	}

	# Optional modifications and side-effects to filter:
	#	-- only filter TCP RST's ("established")
	#   -- cause matching packet to be logged ("l")
	#
	#  These terms come at the end of the source/dest specs we've
	#  parsed to this point.  They can come in any order, and be
	#  interspersed with a variety of other modifiers (tos, precedence, etc.)
	#  which can show up similarly trailing on the access-control spec.
	#
	#  +++wayne bug: 'tos' and 'precedence' should be processed.  See bug 29515.
	#

	# Initialize...may be overridden
	$thisterm->{log} = 'false';

	if( $acl_line =~ /$srctext\s*$srcport\s*$dsttext\s*$dstport([ -_\w]+)/ ) {

		# Force a word start
		my $acEntry_modifiers = ' ' . $1;
		if( $acEntry_modifiers =~ /\bestablished/ and $thisterm->{protocol} eq 'tcp' ) {

			# Set optional ZED "established" filter element
			$thisterm->{established} = 'true';
		}

		if( $acEntry_modifiers =~ /\bl/ ) {
			$thisterm->{log} = 'true';
		}

		# look for the time-range instruction
		# +++wayne bug!: the "no" specifier is not taken into account here.
		#    see bug 29514.
		if( $acl_line =~ /\btime-range\s+(\S+)/i ) {
			$thisterm->{timeAllowed} = $time_range->{$1};
		}
	}

	$out->print_element( 'filterEntry', $thisterm );
	return;
}

sub parse_time_range {
	my ($in) = @_;

	my $months = {
		JAN => '01',
		FEB => '02',
		MAR => '03',
		APR => '04',
		MAY => '05',
		JUN => '06',
		JUL => '07',
		AUG => '08',
		SEP => '09',
		OCT => '10',
		NOV => '12',
		DEC => '12',
	};

	my $time_range;
	while( $in->{startup_config} =~ /^time-range\s+(\S+)(.+?)(?:!|: end)/migs ) {
		my $name = $1;
		$_ = $2;
		if( /absolute\s+start\s+(\d+:\d+)\s+(\d+)\s+(\S+)\s+(\d+)\s+end\s+(\d+:\d+)\s+(\d+)\s+(\S+)\s+(\d+)/mi ) {
			$time_range->{$name}->{startTime} = $1 . ':00';
			$time_range->{$name}->{startDate} =
			$4 . '-' . $months->{ uc( substr( $3, 0, 4 ) ) } . '-' . $2;
			$time_range->{$name}->{endTime} = $5 . ':00';
			$time_range->{$name}->{endDate} =
			$8 . '-' . $months->{ uc( substr( $7, 0, 4 ) ) } . '-' . $6;
			$time_range->{$name}->{startTime} =
			'0' . $time_range->{$name}->{startTime}
			if ( $time_range->{$name}->{startTime} =~ /^\d:/ );
			$time_range->{$name}->{endTime} =
			'0' . $time_range->{$name}->{endTime}
			if ( $time_range->{$name}->{endTime} =~ /^\d:/ );
		} elsif( /periodic\s+(\D+)(\d+:\d+)\s+to\s+(\d+:\d+)/mi ) {
			my $days = $1;
			$time_range->{$name}->{startTime} = $2 . ':00';
			$time_range->{$name}->{endTime}   = $3 . ':00';
			$time_range->{$name}->{startTime} =
			'0' . $time_range->{$name}->{startTime}
			if ( $time_range->{$name}->{startTime} =~ /^\d:/ );
			$time_range->{$name}->{endTime} =
			'0' . $time_range->{$name}->{endTime}
			if ( $time_range->{$name}->{endTime} =~ /^\d:/ );
			$days =~
			s{ (\w+) }{ ( lc ( $1 ) ne 'thursday' ? substr ( $1, 0, 3 ) : substr ( $1, 0, 4 ) ) }gex;
			push @{ $time_range->{$name}->{days} }, split( /\s+/, $days );
		}
	}

	return $time_range;
}

# checks to see if the incoming port is an integer.  If not
# it will do a lookup of the service name to convert it to an integer
sub _int_port {
	my ($port) = @_;
	if ( $port =~ /^\d+$/ ) {
		return $port;
	}
	return get_port_number($port);
}

sub parse_object_groups {
	my ($in, $out) = @_;

	# Command may not exist on certain IOS
	return unless $in->{object_groups};

	my $currentGroup;
    
	for my $line ( split /\n/, $in->{object_groups} ) {
		# There is a type of reported object group in the 12.[23]SX release line (Catalyst 6509 MFSC) only, which is
		# "IP address object group".  We will work diligently to normalize that to our notion of a 'Network' object group, 
		# since the syntax appears identical otherwise.
		if( $line =~ /^(\w+|IP address)\s+object group\s+(\S+)/ ) {
			$currentGroup = {
				type => ( $1 eq 'IP address') ? 'Network' : $1,
				id => $2
			};

			if( $currentGroup->{type} eq 'Network' ) {
				$in->{objectGroup_objects}->{networkGroup}->{ $currentGroup->{id} } = $currentGroup;
			}
		}
		elsif( $line =~ /^\s+(\S+.+)$/ ) {
			my $ogLine = $1;

			# as it's "non-blank" and doesn't open another access list, we'd better have
			# an access list we're working on
			if( not defined $currentGroup ) {
				$LOGGER->debug("Unexpected object-group definition outside of confines of parsed object group: $line");
			} elsif( $ogLine =~ /description\s+(.+)/ ) {
				# description can be for any type
				$currentGroup->{description} = $1;
			} else {

				# dispatch on object-type-specific configuration directive.
				if( $currentGroup->{type} eq 'Network' ) {

					# Host.  This should be IP addressed,
					if( $ogLine =~ /host\s+($IP_REGEX)/ ) {
						push @{ $currentGroup->{host} }, $1;
					}

					# Currently, we're covering the case where this is "host (DNS name)" by just passing 
					# the untranslated DNS name up to topology (where it will be flagged as an error, as least
					# allowing visibility to the user).  Other alternatives (e.g., doing a DNS resolve here,
					# at backup/parse time) don't seem significantly more appealing.	       
					elsif( $ogLine =~ /host\s+(\S+)/ ) {
						my $dnsHostName = $1;
						$LOGGER->debug("Found DNS object-group host name: '$dnsHostName' in object-group. Not currently suportted for translation, putting in SED anyways.");
						push @{ $currentGroup->{host} }, $dnsHostName;
					}
					# Network address
					elsif( $ogLine =~ /^\s*($IP_REGEX)\s+($IP_REGEX)/ ) {
						my ($address, $mask) = ($1, $2);
						$mask = mask_to_bits($mask);
						push @{ $currentGroup->{subnet} }, { address => $address, mask => $mask };
					}
					elsif( $ogLine =~ /^\s*($IP_REGEX)\/([\d]{1,2})/ ) {
						push @{ $currentGroup->{subnet} }, { address => $1, mask => $2 };
					}
					# range
					elsif( $ogLine =~ /range\s+($IP_REGEX)\s+($IP_REGEX)/ ) {
						push @{ $currentGroup->{range} }, { startAddress => $1, endAddress => $2 };
					}

					# any
					elsif( $ogLine =~ /^\s+any/ ) {
						push @{ $currentGroup->{subnet} }, get_any_address_encoding();
					}

					# other object ref.  This must be verified at the time this is serialized to the ZED element.
					elsif( $ogLine =~ /^\sgroup-object\s+(\S+)/ ) {
						push @{ $currentGroup->{objectGroupReference} }, $1;
					} else {
						$LOGGER->debug("Unexpected object-group 'Network' type detail: $ogLine");
					}
				} elsif( $currentGroup->{type} eq 'Service' ) {
					my $svcDetail = undef;

					# well-knonwn primary IP protocols Keep in mind that the "service" protocol can end up covering an ICMP message type.
					# oh, and while we're kvetching about it, a single spec can take on UDP and TCP.  This is why
					# "protocol" is a secondary index-listed structure in the IOSNetworkGroups representation, and
					# gets more suitably decomposed to ZED form at serialize time.

					# +++wayne we cannot deal with separate source/dest specs here in the ZED as IOS can.  We need to catch that here, until
					# we can address this in the ZED and toopology.
					if( $ogLine =~ /\s+source\s.+/ ) {
						$LOGGER->debug("Cackup cannot currently handle separate source/dest object-group network specs, discarding '$ogLine'");
						next;
					}

					if( $ogLine =~ /\s*([udtcp-]{3,})\s+(.*)/ ) {
						my $prot       = $1;
						my $protDetail = $2;
						my $operator   = undef;
						my $port;
						my $rangeDef = 0;

						# Now, to dig into what that detail is...'eq' is a default, or it can be spec'd, along with eq, lt, or gt.
						if( $protDetail =~ /^\s*(\d+)/ ) {
							$operator = "eq";
							$port     = $1;
						} elsif( $protDetail =~ /^\s*([elg][qt])\s+(\S+)/ ) {
							$operator = $1;
							$port     = $2;
						}

						# did either of those parses work?
						if( defined $operator ) {

							# Deal with the fact we may be looking atr both 
							# tcp and udp if "tcp-udp" is the protocol.
							# +++wayne I don't know if this actually shows up in 
							# 'show object-group', or whether it gets
							# split into two detail lines.  This should work 
							# either way.

							for my $thisProt ( split /-/, $prot ) {
								$svcDetail = $in->{objectGroup_objects}->{serviceGroup}->{$thisProt}->{ $currentGroup->{id} };

								if( not defined $svcDetail ) {
									$svcDetail = { id => $currentGroup->{id} };
									$in->{objectGroup_objects}->{serviceGroup}->{$thisProt}->{ $currentGroup->{id} } = $svcDetail;
								}
								push @{$svcDetail->{ports}{"ports_$operator"}}, _int_port($port);
							}
						} elsif( $protDetail =~ /^\s*range\s+(\d+)\s+(\d+)/ ) {
							$port = $1;
							my $portEnd = $2;

							for my $thisProt ( split /-/, $prot ) {
								$svcDetail =
								$in->{objectGroup_objects}->{serviceGroup}->{$thisProt}->{ $currentGroup->{id} };

								if( not defined $svcDetail ) {
									$svcDetail = { id => $currentGroup->{id} };
									$in->{objectGroup_objects}->{serviceGroup}->{$thisProt}->{ $currentGroup->{id} } = $svcDetail;
								}
								push @{ $svcDetail->{ranges} }, [_int_port($port), _int_port($portEnd)];
							}
						}
					} elsif( $ogLine =~ /\s*icmp\s+(\S+)/ ) {

						# +++wayne I've not actually seen one of these in my samples, and I don't have access to a 12.4 IOS system
						# currently, so I'm using some pretty good inference as to how this will show up in a 'show object-groups'
						# output
						my $icmpMessageType = $1;
						$svcDetail =  $in->{objectGroup_objects}->{serviceGroup}->{icmp}->{ $currentGroup->{id} };

						if( not defined $svcDetail ) {
							$svcDetail = { id => $currentGroup->{id} };
							$in->{objectGroup_objects}->{serviceGroup}->{icmp}->{ $currentGroup->{id} } = $svcDetail;
						}
						push @{ $svcDetail->{messageTypes} }, $icmpMessageType;
					} elsif( $ogLine =~ /\s*group-object\s+(\S+)/ ) {
						my $ogRef = $1;

						$svcDetail = $in->{objectGroup_objects}->{serviceGroup}->{group_ref}->{ $currentGroup->{id} };

						if ( not defined $svcDetail ) {
							$svcDetail = { id => $currentGroup->{id} };
							$in->{objectGroup_objects}->{serviceGroup}->{group_ref}->{ $currentGroup->{id} } = $svcDetail;
						}
						push @{ $svcDetail->{groupObjects} }, $ogRef;
					}

					# other IP protocol.  This is, apparently, just a single word in an entry for a pure IP
					# protocol (e.g., ESP).
					elsif( $ogLine =~ /\s+(\S+)/ ) {
						my $ipProt = $1;

						$svcDetail =
						$in->{objectGroup_objects}->{serviceGroup}->{ip}->{ $currentGroup->{id} };

						if( not defined $svcDetail ) {
							$svcDetail = { id => $currentGroup->{id} };
							$in->{objectGroup_objects}->{serviceGroup}->{ip}->{ $currentGroup->{id} } = $svcDetail;
						}
						push @{ $svcDetail->{ipProtocols} }, $ipProt;
					} else {
						$LOGGER->debug("Unable to parse object-group detail entry '$ogLine'");
					}
				}
			}
		}
	}
    
	# Even if there was no content from object_groups, call to generate object groups which have
	# been mapped into intermediate representation (in %IOSObjectGroups) from other native IOS
	# configuration elements.
	# TODO This is bogan. Mapping to an intermediary structure and then to ziptie schema just
	# creates more code (read:bugs). Remove below sub, and do it all in one go.
	_generate_ziptie_object_groups($in, $out);

	return;
}

# organization is
#  (IOSObjectGroups)
#	{network-group}	->
# 		{id} ->
# 			id = name
# 			description = description
# 			host ->
# 				w.x.y.z
# 			subnet ->
# 			 	{ address => w.x.y.z ipVrmask => a.b.c.d }
# 			 	{ address => w.x.y.z mask=nn }
# 			        ....
#	{service-group)->
#		{protocol}->     (this allows defs of a protocol separately by UDP and TCP ports)
#			{(id)}->
#				id = name
# 				description = (descrptive text)
#				ports ->
# 					eq = (port1, port2, port3)  (sorted)
# 					lt = port
# 					gt = port
# 				ranges ->
# 					(range1a, range1b)
# 					(range2a, range2b)....
#
#
#   Notes:
#    We have to do some odd checking and cross-referencing in service groups, but the idea is that network-group
#    keys have been faithfully preserved as ZED element keys, hence can have some loop-automation assigned to
#    their ZED serialization.
sub _generate_ziptie_object_groups {
	my ($in, $out) = @_;
	my $opened = 0;

	# We have to do one scan of the service groups looking just for icmp entries, since they
	# get their own ZED object group element

	if( defined $in->{objectGroup_objects}->{serviceGroup}->{icmp} ) {

		for my $group ( values %{$in->{objectGroup_objects}->{serviceGroup}->{icmp}} ) {
			if( not $opened ) {
				$out->open_element("objectGroups");
				$opened = 1;
			}

			$out->open_element('icmpGroup');
			if( defined $group->{description} ) {
				$out->print_element('description', $group->{description});
			}

			$out->print_element('id', $group->{id});

			for ( @{$group->{messageTypes}} ) {
				$out->print_element("icmpObject", $_);
			}
			$out->close_element("icmpGroup");
		}
	}
    
	if( defined $in->{objectGroup_objects}->{serviceGroup}->{ip} ) { 
		for my $group ( values %{$in->{objectGroup_objects}->{serviceGroup}->{ip}} ) {
			# Same thing for protocol groups, from ip service elements (there are raw IP protocols
			if( not $opened ) {
				$out->open_element('objectGroups');
				$opened = 1;
			}

			$out->open_element('protocolGroup');
			if( defined $group->{description} ) {
				$out->print_element('description', $group->{description});
			}

			$out->print_element('id', $group->{id});

			for my $protocol ( @{$group->{ipProtocols}} ) {
				$out->print_element('protocol', $protocol);
			}
			$out->close_element('protocolGroup');
		}
	}

	# And now, network groups, which have their own analogous ZED tree structure
	for my $group ( values %{$in->{objectGroup_objects}->{networkGroup}} ) {
		if( not $opened ) {
			$out->open_element("objectGroups");
			$opened = 1;
		}

		$out->open_element('networkGroup');

		if( defined $group->{description} ) {
			$out->print_element('description', $group->{description});
		}
		$out->print_element('id', $group->{id});

		# If there are any object group references, generate them, but check here for whether
		# they reference actual network object groups, since if they don't, undefined refs will
		# KO the topology engine if it encounters them in the ZED.

		for my $indirectNwObjGrp ( @{ $group->{objectGroupReference}} ) {
			if( defined $in->{objectGroup_objects}->{networkGroup}->{$indirectNwObjGrp} ) {
				$out->print_element("objectGroupReference", $indirectNwObjGrp);
			} else {
				$LOGGER->debug(  "Unable to find network object group \'" .
				$indirectNwObjGrp . "\', referenced in \'" .
				$group->{id} . "\', suppressing reference");
			}
		}

		# Now, all other elements should have been set up with proper sturcture to support
		# direct print element to the ZED
		for my $nwSubelt ( qw(host range subnet) ) {
			if( defined $group->{$nwSubelt} ) {
				$out->print_element($nwSubelt, $group->{$nwSubelt});
			}
		}
		$out->close_element('networkGroup');
	}

	if( defined $in->{objectGroup_objects}->{serviceGroup} ) {
		my $protocol;
		my $protocolEnt;
		my $portsEnt;                
		my $portRangesEnt;

		if( not $opened ) {
			$out->open_element('objectGroups');
			$opened = 1;
		}


		# The IOSObjectGroups are set up as 
		#  (hash root) -->
		#                 (udp) -->
		#                          (name) --> [fields, ports]
		#                          (name) --> [fields, ports]
		#                             ....
		#                 (tcp) -->
		#                          (name) --> [fields, ports]
		#                          (name) --> [fields, ports]
		#                             ....
		#
		#  HOWEVER, the ZED must have its serviceGroup names be unique, and in fact, wants multiple
		#  protocol-indexed elements within each named service group (in other words, the opposite of
		#  how the service groups have been set up.  What that means we have to do here is to do a "fixup"
		#  on the fly to combine elements where there are ports for BOTH UDP and TCP for the same named
		#  network group.
		my @protocols = qw(udp tcp);
		for my $protocol ( @protocols ) {
			$protocolEnt = $in->{objectGroup_objects}->{serviceGroup}->{$protocol};

			if( defined $protocolEnt ) {
				for my $pGroupKey ( keys %{$protocolEnt} ) {

					my $pGroupEnt = $protocolEnt->{$pGroupKey};                 	
					my @pGroupEvalSecs = ( {protocol => $protocol, nwObjectEnt => $pGroupEnt} );  # The starting section for <portExpression> generation

					# Look at remaining protocol entries to see if "fixup" is called for
					for my $alt_protocol ( @protocols ) {
						next if $protocol eq $alt_protocol;
						if (defined $in->{objectGroup_objects}->{serviceGroup}->{$alt_protocol}->{$pGroupKey}) {
							push @pGroupEvalSecs, { 
								protocol => $alt_protocol, 
								nwObjectEnt => $in->{objectGroup_objects}->{serviceGroup}->{$alt_protocol}->{$pGroupKey}
							};
							delete $in->{objectGroup_objects}->{serviceGroup}->{$alt_protocol}->{$pGroupKey};	
						}
					}

					$out->open_element('serviceGroup');

					# If there is an explicit description, slap it down
					$out->print_element( 'description', $pGroupEnt->{description} ) if defined $pGroupEnt->{description};
					$out->print_element( 'id', $pGroupEnt->{id} );

					# Handle nested group refs.
					# +++how did I end up nesting a group ref under a protocol?
					if (defined $in->{objectGroup_objects}->{serviceGroup}->{group_ref}->{$pGroupEnt->{id}})
					{
						for my $ogIndirectRef (@{$in->{objectGroup_objects}->{serviceGroup}->{group_ref}->{$pGroupEnt->{id}}->{groupObjects}})
						{
							$out->print_element("objectGroupReference", $ogIndirectRef);
						}
					}

					# Now, work through the protocol-based subelements
					my @portExpressions;
					my @portRanges;
					while (@pGroupEvalSecs) {
						my $pgSecEnt = shift @pGroupEvalSecs; 

						# see if the discrete "ports" entries exists
						$portsEnt = $pgSecEnt->{nwObjectEnt}->{ports};

						# this is in turn indexed by a number of well-known equivalence operators, with sets of ports
						# as var-lenght arrays indexed by those operators for each time they show up in the
						# group spec
						if ( defined $portsEnt ) {
							if ( defined $portsEnt->{ports_eq} ) {

								# What we want to do here is to determine the series of port scalars
								# and ranges we can take advanage of.  Port sets here are guaranteed
								# to be sorted on assignment.
								my @portSet          = @{ $portsEnt->{ports_eq} };
								my @mappedPortExps   = ();
								my @mappedPortRanges = ();

								while (@portSet) {
									my $currentPort = shift @portSet;

									if( @portSet and ( $currentPort + 1 ) == $portSet[0] ) {

										# we have a range!
										# $out->open_element("portRange");
										my $nextRange = {
											protocol  => $pgSecEnt->{protocol},
											portStart => _int_port($currentPort)
										};

										# now just count contiguous ports
										while( @portSet and ( $currentPort + 1 ) == $portSet[0] ) {
											$currentPort = shift @portSet;
										}

										# we counted to the end
										$nextRange->{portEnd} = _int_port($currentPort);
										push @mappedPortRanges, $nextRange;
									} else {
										# just a port ent
										my $nextExp = {
											port     => _int_port($currentPort),
											operator => 'eq',
											protocol => $pgSecEnt->{protocol}
										};

										push @mappedPortExps, $nextExp;
									}
								}

								# now, just walk the arrays of exps and ports.
								for my $portExp ( @mappedPortExps ) {
									push @portExpressions, $portExp;
								}

								for my $portRange ( @mappedPortRanges ) {
									push @portRanges, $portRange;
								}
							}

							# handle lt and gt expressions                           
							for my $ltPortVal ( @{$portsEnt->{ports_lt}} ) {
								my $ltElement = { port => $ltPortVal, operator => 'lt', protocol => $pgSecEnt->{protocol} };
								push @portExpressions, $ltElement;
							}

							for my $ltPortVal ( @{$portsEnt->{ports_gt}} ) {
								my $ltElement = { port => $ltPortVal, operator => 'gt', protocol => $pgSecEnt->{protocol} };
								push @portExpressions, $ltElement;
							}                                              
						}

						for my $portRangeEnt ( @{$pGroupEnt->{ranges}} ) {
							my $rangeElement = { portStart => $portRangeEnt->[0], portEnd => $portRangeEnt->[1], protocol => $pgSecEnt->{protocol} };
							push @portRanges, $rangeElement;
						}
					}

					# Print ports
					for my $portExp ( @portExpressions ) {
						$out->print_element('portExpression', $portExp);
					}
					for my $portRange ( @portRanges ) {
						$out->print_element('portRange', $portRange);
					}

					$out->close_element('serviceGroup');
				}
			}
		}
	}
    
	if ($opened) {
		$out->close_element("objectGroups");
	}
    
	# We need to clear out the objectGroup_objects from one run to the next.  This is probably the best place
	# to do it, since it would be regenerated from config again before this routine is called...
	if (defined $in->{objectGroup_objects})
	{
		delete $in->{objectGroup_objects};
	}

	return;
}

sub parse_interfaces {
	my ($in) = @_;
	my $interfaces; # array ref to be filled with model interfaces
	my $names;      # array ref to be filled with all interface names
	my $subnets;    # hash ref to be filled with interface name -> connected subnet
	my $filterAssignments; 


	# Need to perform a secondary (well, asuming parse_interfaces is the primary) pass to
	# assess interface grouping characteristics.
	_parse_interface_groupings($in);

	INTERFACE:
	while( $in->{startup_config} =~ /^interface\s+(\S+)(.+?)^!/smig ) {
		my $name      = $1;
		my $blob      = $2;
		my $interface = {
			name          => $name,
			interfaceType => get_interface_type($name),
			physical      => _is_physical($name),
		};

		push @{ $names }, $name;

		if( $blob =~ /^\s*description\s+(.+?)\s*$/mi ) {
			$interface->{description} = $1;
		}

		# Parse IP addresses
		my $ipConfiguration;
		my $order = 1;
		while( $blob =~ /^\s*ip address\s+($IP_REGEX)\s+($IP_REGEX)(\s+secondary)?/img ) {
			$ipConfiguration = {
				ipAddress  => $1,
				mask       => mask_to_bits($2),
				precedence => $order++,
			};
			push @{ $interface->{interfaceIp}->{ipConfiguration} }, $ipConfiguration;
			my $subnet = new ZipTie::Addressing::Subnet( $ipConfiguration->{ipAddress}, $ipConfiguration->{mask} );
			push @{ $subnets->{$name} }, $subnet;
		}

		$order = 1;
		while( $blob =~ /^\s*ipv6 address\s+([A-Z\d:]+)\/(\d+)/mig ) {
			$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping interface IPv6 address: $1\/$2 on $name");
		}
		while( $blob =~ /^\s*ip helper-address\s+(\S+)/mig ) {
			push @{ $interface->{interfaceIp}->{udpForwarder} }, $1;
		}
		while( $blob =~ /^\s*ip access-group\s*\b(.+?)\b\s+(in|out)\s*$/mig ) {
			my $acl       = $1;
			my $direction = $2;
			if( $direction eq 'out' ) {
				# Don't assign ACL to interface, creates duplicates in ACL page
				#$interface->{egressFilter} = $1;
				push @{$filterAssignments->{$name}}, { filterName => $1, direction => 'egress'};
			} elsif( $direction eq 'in' ) {
				#$interface->{ingressFilter} = $1;
				push @{$filterAssignments->{$name}}, { filterName => $1, direction => 'ingress'};
			}
		}

		# Process interface grouping characteristics.  We can optimize this by noting right away if this 
		# has been so marked by _parse_interface_groupings.
		for my $ifGrpIndex ( @{$in->{interfacesData}->{$name}->{ifGroupings}} ) {

			# we can iterate across primary and secondary membership lists
			for my $grpDescr ( @{$in->{interfaceGroupings}->{$ifGrpIndex}->{secondaryInterfaces}}, 
			($in->{interfaceGroupings}->{$ifGrpIndex}->{primaryInterface}) )
			{
				if ($grpDescr->{ifName} eq $name) {
					push @{$interface->{interfaceGrouping}}, {
						ifGroupName => $ifGrpIndex,
						ifGroupingCharacteristic => $grpDescr->{ifGroupCharacteristic}
					};
				}
			}
		}
		# Independently, let's see if this is marked as primary on any interface grouping
		if (exists $in->{interfaceGroupings}->{$name}) {
			push @{$interface->{interfaceGrouping}}, {
				ifGroupName => $name,
				ifGroupingCharacteristic => $in->{interfaceGroupings}->{$name}->{primaryInterface}->{ifGroupCharacteristic}
			};
		}

		# process ethernet properties
		if( $name =~ /eth/i ) {
			if( $blob =~ /^(\s+(half|full|auto).duplex|\s+duplex\s+(\S+))/mi ) {
				my $duplex = $2 // $3;
				if( defined $duplex ) {
					if( $duplex =~ /auto/i ) {
						$interface->{interfaceEthernet}->{autoDuplex} = 'true';
					} else {
						$interface->{interfaceEthernet}->{autoDuplex} = 'false';
						$interface->{interfaceEthernet}->{operationalDuplex} = $duplex;
					}
				}
			} elsif( $blob =~ /^\s+speed\s+\d+/mi ) {
				$interface->{interfaceEthernet}->{autoSpeed} = 'false';
			} elsif ( $blob =~ /^\s+speed\s+auto/mi ) {
				$interface->{interfaceEthernet}->{autoSpeed} = 'true';
			} elsif ( $blob =~ /^\s+no negotiation/mi ) {
				$interface->{interfaceEthernet}->{autoSpeed} = 'false';
			}
		}

		# see if this is a tunneling interface
		if( $blob =~ /^\s+tunnel\s+source\s+(\S+)/mi ) {
			my $tunnelLocalEndBlob = $1;
			my $tunnelLocalEndProps;
			my $tunnelIfProps;

			if($tunnelLocalEndBlob =~ /\d+\.\d+\.\d+\.\d+/) {
				$tunnelLocalEndProps->{tunnelLocalPeeringIPAddress} = $tunnelLocalEndBlob;
			} elsif(get_interface_type($tunnelLocalEndBlob) ne 'other') { # if it returns as a well known i/f type of SOME sort
				$tunnelLocalEndProps->{tunnelLocalPeeringInterface} = $tunnelLocalEndBlob;				
			} else {
				$LOGGER->debug("Cannot find local peer info in configuration for tunnel with parsed local endpoint $tunnelLocalEndBlob; bypassing interface");
				next INTERFACE;
			}

			if( defined $tunnelLocalEndBlob ) {
				$tunnelIfProps = { tunnelLocalPeering => $tunnelLocalEndProps};

				if( $blob =~ /^\s+tunnel\s+destination\s+(\d+\.\d+\.\d+\.\d+)/mi ) {
					$tunnelIfProps->{tunnelRemotePeerIPAddress} = $1;
				} else {
					$LOGGER->debug("Cannot find remote peer address in configuration for tunnel with local address $tunnelLocalEndBlob");
				}

				# Find tunnel interface encapsulation type, or rather, if it deviates from GRE
				if( $blob =~ /^\s+tunnel\s+mode\s+(\S+)/m ) {
					my $mode = $1;
					if( $mode =~ /ipsec/ ) {
						$tunnelIfProps->{encapsulationProtocol} = 'ipSec';
					} elsif( $mode =~ /ipip/ ) {
						$tunnelIfProps->{encapsulationProtocol} = 'ipv4-in-ipv4';
					} elsif( $mode =~ /ip6ip/ ) {
						$tunnelIfProps->{encapsulationProtocol} = 'ipv6-in-ipv4';
					}
				}
			} 
			$tunnelIfProps->{encapsulationProtocol} = 'gre' if not defined $tunnelIfProps->{encapsulationProtocol};
			$interface->{interfaceTunnel} = $tunnelIfProps;
		}

		# VLAN trunk references
		if ( $blob =~ /^\s+switchport\s+trunk\s+allowed\s+vlan\s+(\S+)/mi ) {
			my $vlandata = $1;
			if( $vlandata =~ /^\d+$/ ) {
				push @{ $interface->{interfaceVlanTrunks} }, { startVlan => $vlandata };
			} else {
				for my $trunkedvlan ( split /,/, $vlandata ) {
					if( $trunkedvlan =~ /-/ ) {
						my ($start, $end) = split /-/, $trunkedvlan;
						push @{ $interface->{interfaceVlanTrunks} }, { startVlan => $start, endVlan => $end };
					} else {
						push @{ $interface->{interfaceVlanTrunks} }, { startVlan => $trunkedvlan };
					}
				}
			}
		}

		# process this particular interface from "show interfaces"
		if( $in->{interfaces} =~ /^($name\s.+?)(?=^\S)/msi ) {
			my $intBlob = $1;
			if( $intBlob =~ /$name\s+is\s+(up|\S+\s+down|down),/mi ) {
				$interface->{adminStatus} = ( $1 =~ /down/i ) ? 'down' : 'up';
			}
			if( $intBlob =~ /address is ([a-f0-9A-F]{4}\.[a-f0-9A-F]{4}\.[a-f0-9A-F]{4})/mi ) {
				my $macAddress = $1;
				$interface->{interfaceEthernet}->{macAddress} = strip_mac($macAddress);
			}
			if( $intBlob =~ /MTU\s+(\d+)\s+bytes,\s+BW\s+(\d+)\s+(\S+),/mi ) {
				my $mtu   = $1;
				my $speed = $2;
				my $units = $3;
				if ( $units =~ /Mb/i ) { $speed = $speed * 1000 * 1000; }
				if ( $units =~ /Kb/i ) { $speed = $speed * 1000; }
				$interface->{speed} = $speed;
				$interface->{mtu}   = $mtu;
			}
			if( $intBlob =~ /MTU\s+(\d+)\s+bytes,\s+sub\s+MTU\s+\d+,\s+BW\s+(\d+)\s+(\S+),/ ) {
				my $mtu   = $1;
				my $speed = $2;
				my $units = $3;
				if ( $units =~ /Mbit/i ) { $speed = $speed * 1000; }
				$interface->{speed} = $speed;
				$interface->{mtu}   = $mtu;
			}
			if( $intBlob =~ /Encapsulation\s([^\s^,]+)/ ) {
				my $encaps = $1;
				if( defined $interface->{interfaceEthernet}
				or  $interface->{interfaceType} eq 'ethernet' ) {
					$interface->{interfaceEthernet}->{encapsulation} = $encaps;
				}
			}
			if( $intBlob =~ /(Half|Full|Auto)[- ]duplex,\s+(100Mb|10Mb|Auto).+,\s+(media type is |)(\S+)/i ) {
				my $duplex    = $1;
				my $autospeed = $2;
				my $media     = $4;
				if( $duplex =~ /half/i ) {
					$interface->{interfaceEthernet}->{operationalDuplex} = 'half';
				} elsif( $duplex =~ /full/i ) {
					$interface->{interfaceEthernet}->{operationalDuplex} = 'full';
				} elsif( $duplex =~ /auto/i ) {
					$interface->{interfaceEthernet}->{autoDuplex} = 'true';
				}
				if( $autospeed =~ /auto/i ) {
					$interface->{interfaceEthernet}->{autoSpeed} = 'true';
				}
				$interface->{interfaceEthernet}->{mediaType} = $media;
			}

			# Last chance for an internet address here if we don't have one (for example, an ISDN interface only activated
			# on demand where no IP address would be in running-config, but might be in 'show interfaces' state)
			if( not defined $ipConfiguration) {
				$order = 1;
				while( $intBlob =~ /Internet\s+address\s+is\s+([\d.:a-fA-F]+)\/(\d+)/gi ) {
					$ipConfiguration = {
						ipAddress  => $1,
						mask       => mask_to_bits($2),
						precedence => $order++,
					};
					push( @{ $interface->{interfaceIp}->{ipConfiguration} }, $ipConfiguration );

					my $subnet = new ZipTie::Addressing::Subnet( $ipConfiguration->{ipAddress}, $ipConfiguration->{mask} );
					push @{ $subnets->{$name} }, $subnet;
				}
			}
		}

		push @{ $interfaces }, $interface;

		if( defined $interface->{name} ) {
			# Add the interfaces to the remoteGatewayResolution object to help resolve static routes with remote gateways.
			if( defined $interface->{interfaceIp} ) {
				for my $ipConfiguration ( @{$interface->{interfaceIp}->{ipConfiguration}} ) {
					$REMOTE_GATEWAY_RESOLVER->add_local_subnet($ipConfiguration->{ipAddress}, $ipConfiguration->{mask}, $interface->{name});
				}
			}
		}
	}

	$in->{parsed_interfaces} = $interfaces;
	$in->{interface_mapper} = $subnets;
	$in->{filter_assignments} = $filterAssignments;
	$in->{interface_names} = $names;

	return;
}

sub print_interfaces {
	my ($in, $out) = @_;
	if( $in->{parsed_interfaces} ) {
		$out->print_element( 'interfaces', { interface => $in->{parsed_interfaces} } );
	}
	return;
}

# _parse_interface_groupings
#
#  Does a pass of running-config to establish the interfaces which are "grouped" along several criteria--
#  deferral of an unnumbered interface to an interface whose address it is using as endpoint, channeling
#  of multiple interfaces, etc.
#
#  This sets up a hash, $in->{interfaceGroupings}, whose index is the "primary" interface of such a 
#  relationship (or the channel name in the case of EtherChannel, for example).
#
sub _parse_interface_groupings {
	my ($in) = @_;

	while( $in->{startup_config} =~ /^interface\s+(\S+)(.+?)^!/smig ) {
		my ($ifName, $ifBlob) = ($1, $2);
		my $primaryIf;
	
		# One case  deals with unnumbered interfaces
		if ( $ifBlob =~ /^\s*ip\s+unnumbered\s+(\S+)\s*$/mi ) {
			$primaryIf = $1;

			# Assert presence of a hash bucket with the primary name of the interface
			$in->{interfaceGroupings}->{$primaryIf}->{primaryInterface} = {
				ifName => $primaryIf,
				ifGroupCharacteristic => 'proxyAuthority'
			};
			my $proxyAffiliationDesc = {
				ifName => $ifName,
				ifGroupCharacteristic => 'proxyAffiliation'
			};

			push @{$in->{interfaceGroupings}->{$primaryIf}->{secondaryInterfaces}}, $proxyAffiliationDesc;

			# Rounding this up, We'll use an extra ref off of $in to store an index *by secondary interface*
			if (defined $primaryIf) {
				push @{$in->{interfacesData}->{$ifName}->{ifGroupings}}, $primaryIf;
			}
		}

		# Another deals independently with subinterfaces, which for IOS, we'll take from the interface name
		if ($ifName =~ /^(\S+)\.\d+/) {
			$primaryIf = $1;
	
			$in->{interfaceGroupings}->{$primaryIf}->{primaryInterface} = { 
				ifName => $primaryIf,
				ifGroupCharacteristic => 'primaryInterface'
			};

			my $subInterfaceDesc = {
				ifName => $ifName,
				ifGroupCharacteristic => 'subInterface'
			};
			push @{$in->{interfaceGroupings}->{$primaryIf}->{secondaryInterfaces}}, $subInterfaceDesc;

			# Rounding this up, We'll use an extra ref off of $in to store an index *by secondary interface*
			if (defined $primaryIf) {
				push @{$in->{interfacesData}->{$ifName}->{ifGroupings}}, $primaryIf;
			}
		}
	}
	return;
}

# This routine is used to create a mapping of NAT "zones" to their corresponding interfaces
sub _parse_nat_interfaces {
	my $in = shift;
	my $natZoneMapping;

	# Commence a scan of the interface sections, looking for the "ip nat" indicator
	while ( $in->{startup_config} =~ /^interface\s+(\S+)(.+?)^!/smig ) {
		my ($name, $blob) = ($1, $2);

		if ($blob =~ /^\s*ip\s+nat\s+(\S+)/mi) {
			push @{$natZoneMapping->{$1}}, $name;
		}
	}
	return $natZoneMapping;
}

sub _is_physical {
	my $name = shift;
	my $is_physical = 'false';
	if( $name !~ /\.\d+$/ and $name =~ /seri|eth|gig|^fe|^fa|token|[bp]ri/i ) {
		$is_physical = 'true';
	}
	return $is_physical;
}

# Remove any trivial pieces of the configuration that may cause improper diffs
sub _clean_config_text {
	my $text = shift;
	$text =~ s/\n^ntp clock-period.+$//m;                 # remove the variable ntp clock-period line
	$text =~ s/\n^!\s*NVRAM config last updated at.+$//m; # remove the NVRAM updated time comment
	$text =~ s/\n^!\s*Last configuration change at.+$//m; # remove the Last config change comment
	return $text;
}

#  Presuming the interface -> subnet structure returned from parse_interfaces to find a primary host
#  address associated with the interface.
sub _mapIfNameToPrimaryAddr {
	my ($interface, $subnets) = @_;
	my $address;
	
	if( defined $subnets->{$interface} ) {
		# This only supports one IP per interface
		$address = $subnets->{$interface}->[0]->getHostIp();
	}
	
	return $address;
}

1;
