package ZipTie::Adapters::Cisco::IOS::TelemetryParsers;

use strict;
use warnings;
use 5.010_000;

use ZipTie::Adapters::Cisco::IOS::Utils qw(get_full_interface_name);
use ZipTie::Adapters::Utils qw(strip_mac trim get_interface_type get_crep);
use ZipTie::Logger;

my $LOGGER = ZipTie::Logger::get_logger();

use Exporter qw(import);
our @EXPORT_OK = qw(parse_arp parse_cdp parse_routing_neighbors parse_telemetry_interfaces parse_mac_table);

my $IP = get_crep('cipm');
my $MAC = get_crep('mac1') .'|' . get_crep('mac2') . '|' . get_crep('mac3') . '|' . get_crep('mac4');

sub parse_routing_neighbors {
	my ($in, $out) = @_;
	my $neighborData;
	my @routing_neighbors;

	while( $in->{ip_ospf_neighbors} =~ /^([\da-f.:]+)\s+\d+\s+\S+\s+\S+\s+([\da-f\.:]+)\s+(\S+)\s*$/gm ) {
		my $neighbor = {
			protocol  => 'OSPF',
			routerId  => $1,
			ipAddress => $2,
			interface => $3,
		};

		push @routing_neighbors, $neighbor;
	}		

	while( $in->{ip_bgp_neighbors} =~ /^BGP neighbor is ([\da-f:\.]+)(.+?)BGP state/gsm ) {
		my $neighbor = {
			protocol  => 'BGP',
			ipAddress => $1,
		};

		my $blob = $2;
		if ($blob =~ /router ID\s*([\da-f:.]+)/) {
			$neighbor->{routerId} = $1;
		}
		push @routing_neighbors, $neighbor;
	}

	while( $in->{ip_eigrp_neighbors} =~ /^(IP-EIGRP neigh|EIGRP-IPv4 Neigh).*?(process|AS)[ \(](\d+).*?\n(.+)$/gsim ) {
		my ($eigrpProc, $eigrpSec) = ($3, $4);

		while( $eigrpSec =~ /^(?:\d+\s+)?([\da-f:\.]{3,})\s+(\S+)/gm ) {
			# Fill in context about this route for return to caller
			$neighborData->{eigrp}->{$1} = { routeProcess => $eigrpProc, fwdInterface => get_full_interface_name($2) };

			my $neighbor = {
				protocol  => 'EIGRP',
				ipAddress => $1,
				routerId  => $1,
				interface => get_full_interface_name($2),
			};
			push @routing_neighbors, $neighbor;
		}
	}
	if( @routing_neighbors ) {
		$out->print_element('routingNeighbors', { routingNeighbor => \@routing_neighbors } );
	}

	return $neighborData;
}

sub parse_arp {
	my ($in, $out) = @_;
	my $arps;

	while ($in->{arp} =~ /^\s*\S+\s+($IP)\s+[\d\-]+\s+($MAC)\s+\S+\s+(\S+)\s*$/mg) {
		my $arp = {
			ipAddress  => $1,
			macAddress => strip_mac($2),
			interface  => $3,
		};
		push(@{$arps->{arpEntry}}, $arp);
	}

	if ($arps) {
		$out->print_element('arpTable', $arps);
	}

	return;
}

sub parse_cdp {
	my ($in, $out) = @_;
	my $opened = 0;

	while ($in->{cdp} =~ /^-+\s*$(.+?advertisement.+?)(?=^\s)/msg) {
		if (!$opened) {
			$opened = 1;
			$out->open_element('discoveryProtocolNeighbors');
		}

		my $singleCdpBlob = $1;
		my $neighbor = { protocol => 'CDP', };
		if ($singleCdpBlob =~ /Device\s*ID:\s*(\S+)/) {
			$neighbor->{sysName} = $1;
		}
		if ($singleCdpBlob =~ /IP address:\s*(\S+)/i) {
			$neighbor->{ipAddress} = $1;
		}
		if ($singleCdpBlob =~ /Interface:\s*(\S+)/m) {
			$neighbor->{localInterface} = $1;
			$neighbor->{localInterface} =~ s/,$//;
		}
		if ($singleCdpBlob =~ /\(outgoing port\):\s*(\S+)/) {
			$neighbor->{remoteInterface} = $1;
		}
		if ($singleCdpBlob =~ /Platform:\s*(\b[^,]+)/) {
			$neighbor->{platform} = $1;
		}
		if ($singleCdpBlob =~ /Version\s*:\s*(.+?)(?=^\s+)/ms) {
			$neighbor->{sysDescr} = trim($1);
		}
		$out->print_element('discoveryProtocolNeighbor', $neighbor);
	}
	if ($opened) {
		$out->close_element('discoveryProtocolNeighbors');
	}

	return;
}

sub parse_telemetry_interfaces {
	my ($in, $out) = @_;
	my $interfaces;

	while ($in->{interfaces} =~ /(\S+)\s+(is.+?)(?=^\S)/msg) {
		my $blob      = $2;
		my $name      = $1;
		my $interface = {
			name       => $name,
			type       => get_interface_type($name),
			inputBytes => 0,
		};

		if ($blob =~ /line protocol is (down|up)/i) {
			$interface->{operStatus} = ucfirst($1);
		}
		if ($blob =~ /packets input,.*?(\d+)\s*bytes/) {
			$interface->{inputBytes} = $1;
		}

		while ($blob =~ /Internet\s+address\s+is\s+([\d.:a-fA-F]+)\/(\d+)/gi) {
			my $ipEntry = {
				ipAddress => $1,
				mask      => $2,
				-virtual  => 0,
			};
			push(@{$interface->{ipEntry}}, $ipEntry);
		}

		# Now push any virtual IP addresses
		if( $in->{standby} =~ /$name\b.+?Virtual IP address is\s+(\S+).+?(?=^\S)/ms ) {
			my $ipEntry = { ipAddress => $1, -virtual => 1 };
			push @{$interface->{ipEntry}}, $ipEntry;
		}

		push(@{$interfaces->{interface}}, $interface);
	}

	# This needs to be assigned as attribute-level encoding
	$out->attributes(1);
	$out->print_element('interfaces', $interfaces);

	return $interfaces;
}

sub parse_mac_table {
	my ($in, $out) = @_;
	my $opened = 0;
	my $macEntry;

	# there are three different forms, one of which completely reorders the fields
	if ($in->{mac} =~ /Destination Address\s+Address Type\s+VLAN\s+Destination Port/i) {
		my $macAddrRe = get_crep('mac3');
		while ($in->{mac} =~ /($macAddrRe)\s+\w+\s+(\d+)\s+(\S+)/g) {
			if (!$opened) {
				$opened = 1;
				$out->open_element('macTable');
			}

			$macEntry = {
				vlan       => $2,
				macAddress => strip_mac($1),
				interface  => get_full_interface_name($3)
			};
			$out->print_element('macEntry', $macEntry);
		}
	}else{
		my $regex;
		if ($in->{mac} =~ /Vlan\s*Mac Address\s*Type\s*Ports/i) {
			$regex = '(\S+)\s*([\da-f.]{14})\s*\S+\s*(\S+)';
		} else {
			$regex = '(\S+)\s*([\da-f.]{14})\s*\S+\s*(?:Yes|No)\s*(\S+)';
		}
		while ($in->{mac} =~ /$regex/g) {
			if (!$opened) {
				$out->open_element('macTable');
				$opened = 1;
			}
			$macEntry = {
				vlan       => $1,
				macAddress => strip_mac($2),
				interface  => get_full_interface_name($3),
			};
			$out->print_element('macEntry', $macEntry);
		}
	}
	if ($opened) {
		$out->close_element('macTable');
	}
}

1;
