package ZipTie::Adapters::Cisco::NGIPS::Utils;

use strict;
use warnings;
use 5.010_000;

use ZipTie::Logger;
use ZipTie::Snort;

use Exporter qw(import);
our @EXPORT_OK = qw(get_responses);

my $LOGGER = ZipTie::Logger::get_logger();

# Returns a reference to responses from the device
sub get_responses{
	my ($cli_protocol, $device_prompt_regex, $connection_path) = @_;
	
	# Hash keys -> commands for the Sourcefire shell
	my @sourcefire_shell = (
		'version->show version',
		'network->show network',
		'interfaces->show interfaces'
	);

	# Hash keys -> commands for standard bash shell	
	my @bash_shell = (
		'hostname->hostname',
		'ip_addr->ip addr',
		'route_n->route -n',
	);

	# Send the commands and fill the responses
	my $responses;
	$LOGGER->debug('Gathering responses');

	foreach my $key_command(@sourcefire_shell){
		my ($key, $command) = split /-\>/, $key_command;
		$responses->{$key} = $cli_protocol->send_and_wait_for($command, $device_prompt_regex);

		# Magic			^[[7m--More--^[[27m
		$responses->{$key} =~ s/\e\[\d+M-+MORE-+\e\[\d+M//gi;
	}

	# Change to expert shell
	my $successful = '';
	my $removed_first_prompt = '('.substr($device_prompt_regex, index($device_prompt_regex, '|') + 1);
	while($successful !~ /$removed_first_prompt/){
		$successful = $cli_protocol->send_and_wait_for('expert', $device_prompt_regex);
	}

	# Switch user to root
	# TODO: Remove this and find privleges needed, update User Docs
	my $removed_second_prompt = '('.substr($removed_first_prompt, index($removed_first_prompt, '|') + 1);
	my $password_prompt_regex = '([Pp]assword\s*\:)|'.$removed_second_prompt;

	$successful = $cli_protocol->send_and_wait_for('sudo su -', $password_prompt_regex);
	if($successful !~ /$removed_second_prompt/){
		my $password = $connection_path->get_credentials()->{password};

		$LOGGER->debug('Sending root password quietly');
		$LOGGER->set_logging_level(1);
		$cli_protocol->send_and_wait_for($password, $removed_second_prompt);
		$LOGGER->set_logging_level(0);
	}

	# Send expert shell commands and collect their responses
	foreach my $key_command(@bash_shell){
		my ($key, $command) = split(/-\>/, $key_command);
		$responses->{$key} = $cli_protocol->send_and_wait_for($command, $device_prompt_regex);
	}

	get_snort_responses($cli_protocol, $device_prompt_regex, $responses);
	get_db_responses($cli_protocol, $device_prompt_regex, $responses);
	
	# get QRadar version - (to determine if we model userFilters)
	$responses->{qradarVersion} = get_qradar_version();

	return $responses;
}

sub get_snort_responses{
	my ($cli_protocol, $device_prompt_regex, $responses) = @_;

	# Get ROOT from /etc/sf/ims.conf this will point to where the detection engines are kept
	my $snort_root = '/usr/local/sf';
	my $snort_ims_conf = $cli_protocol->send_and_wait_for('cat /etc/sf/ims.conf', $device_prompt_regex);
	if($snort_ims_conf =~ /^ROOT="(\S+)"\s*$/mi){
		$snort_root = $1;
		while($snort_root =~ /\$\{(\S+)\}/g){
			my $variable = $1;
			my $value = '';
			if($snort_ims_conf =~ /^$variable="(\S+)"\s*$/mi){
				$value = $1;
			}
			$snort_root =~ s/\$\{$variable\}/$value/g;
		}
	}

	# Get the detection engine uuid
	my $snort_de_path;
	my $snort_affinity_conf = $cli_protocol->send_and_wait_for("cat $snort_root/fwcfg/affinity.conf", $device_prompt_regex);
	if($snort_affinity_conf =~ /DES\s*{\s*(\S+)\s*{\s*INSTANCES/ig){
		$snort_de_path = "$snort_root/detection_engines/$1";
	}else{
		$LOGGER->debug('Could not find detection engine path for master snort');
		return;
	}
	$responses->{policy_text} = $cli_protocol->send_and_wait_for("cat $snort_de_path/policyText_full.yaml", $device_prompt_regex);

	# Create callback for getting files
	my $get_config_callback = 
		sub{
			my ($file) = @_;

			if($file !~ /^\//){
				$file = "$snort_de_path/$file";
			}

			my $response = $cli_protocol->send_and_wait_for("cat $file", $device_prompt_regex);
			return($response);
		};

	my @snort_confs = ("$snort_de_path/snort.conf");
	while(scalar(@snort_confs)){
		my $snort_conf_path = shift(@snort_confs);

		# Create SNORT object
		my $snort_name = '';
		my $snort_head = $cli_protocol->send_and_wait_for("head $snort_conf_path", $device_prompt_regex);
		if($snort_head =~ /^#\s*(AC)?\s*NAME\s*\:\s*(.*?)\s*$/mi){
			$snort_name = $2;
			$snort_name =~ s/\s+/_/g;
		}
		my $snort = ZipTie::Snort->new($snort_name, $get_config_callback);

		# Parse snort.conf and any files it refrences
		$snort->parse_config($snort_conf_path);
		
		# Get SNORT uuid and put the SNORT object in the responses, support multiple snorts
		my $snort_uuid = 'MASTER';
		if($snort_conf_path =~ /SNORT\.CONF\.([0-9A-F-]+)/i){
			$snort_uuid = $1;
		}
		$responses->{snorts}->{$snort_uuid} = $snort;
		
		push(@snort_confs, @{$snort->get_bound_snorts()});
	}

	# Get securtiy intelligence lists and feeds
	my $si_files = $cli_protocol->send_and_wait_for("find $snort_root/iprep_download -name '*lf'", $device_prompt_regex);
	while($si_files =~ /^($snort_root\/iprep_download\/([0-9a-f-]+)\.[wb]lf)\s*$/img){
		my ($file, $uuid) = ($1, $2);
		$responses->{lists}->{$uuid} = $cli_protocol->send_and_wait_for("cat $file", $device_prompt_regex);
	}

	return;
}

sub get_db_responses{
	my ($cli_protocol, $device_prompt_regex, $responses) = @_;

	# Get database credentials
	my $snort_ims_data = $cli_protocol->send_and_wait_for('cat /etc/sf/ims-data.conf', $device_prompt_regex);
	while($snort_ims_data =~ /^\s*SF_MYSQL_(\S+)\s+(\S+)\;\s*$/img){
		my $key = lc $1;
		$responses->{snort_db}->{$key} = $2;
	}

	# Set up credentials with defaults if none where found
	my $snort_db_name = $responses->{snort_db}->{database} // 'sfsnort';
	my $snort_db_username = $responses->{snort_db}->{name} // 'barnyard';
	my $snort_db_password = $responses->{snort_db}->{pw} // 'barnyard';
	my $snort_db_command = "mysql -u$snort_db_username -p$snort_db_password $snort_db_name";
	my $snort_db_prompt = '\S+\>\s*$';

	# Quietly send the command to access the database
	$LOGGER->debug('Entering database quietly');
	$LOGGER->set_logging_level(1);
	my $successful = $cli_protocol->send_and_wait_for($snort_db_command, $device_prompt_regex);
	$LOGGER->set_logging_level(0);

	# If the prompt is not the expected database prompt fail
	# WARNING: This has misfired once
	if($successful !~ /$snort_db_prompt/){
		$LOGGER->fatal("Access denied from mysql db: $snort_db_name");
	}

	# Ignore apps with an id greater than 2000000000. These are "client applications", we are only interested in their "server application" counterpart.
	my $client_start = 2000000000;
	my $custom_start = 0;#1000000; 
	my @snort_db_queries = (
		"db_app_id->select appId, appName, appDescription, risk_index, productivity_index, (isServer + (isClient * 2) + (isWebApp * 4)) as typeMask from appIdInfo where appId>=$custom_start and appId<$client_start and deleted=0;",
		"db_tag->select appId, tagName from application_tag_view where appId>=$custom_start and appId<$client_start;",
		"db_category->select appId, categoryName from application_category_view where appId>=$custom_start and appId<$client_start;"
	);

	# Send queries and collect responses
	foreach my $key_query(@snort_db_queries){
		my ($key, $command) = split /-\>/, $key_query;
		$responses->{$key} = $cli_protocol->send_and_wait_for($command, $device_prompt_regex);
	}

	return;
}

sub get_qradar_version {
	my $version = `/opt/qradar/bin/myver`;
	$version =~ s/\.//g;
	$version =~ /(\d\d\d)/;
	return $1;
}

1;
