#!/usr/bin/env perl

use strict;
use warnings;
use 5.010_000;

use Data::Dumper;
use ZipTie::Adapters::Cisco::NGIPS::Disconnect qw(disconnect);
use ZipTie::Adapters::Utils qw(get_crep);
use ZipTie::CLIProtocolFactory;
use ZipTie::Client;
use ZipTie::ConnectionPath::Protocol;
use ZipTie::ConnectionPath;
use ZipTie::Credentials;
use ZipTie::Discovery qw(get_connection_path get_credentials get_devices add_or_update_devices);
use ZipTie::Logger;
use ZipTie::Recording;

$ENV{PERL_LWP_SSL_VERIFY_HOSTNAME} = 0;

my $ADAPTER = 'ZipTie::Adapters::Cisco::NGIPS';

my $PROTOCOLS = [ 'SSH' ];

my $LOGGER = ZipTie::Logger::get_logger();

execute_discovery( $ARGV[0] );

sub execute_discovery {
	my ($mgmt_server_ip) = @_;

	if( $ENV{ADAPTER_LOG_TO_FILE} ) {
		$LOGGER->enable_logging_to_file('Cisco_Firepower_Management_Center', 'fmcdiscovery', $mgmt_server_ip);
	}

	# Check parameter is a valid ip
	my $ip_regex = get_crep('cipm');
	if( $mgmt_server_ip !~ /^$ip_regex$/ ) {
		_fatal("Invalid IP address: $mgmt_server_ip");
	}

	# Get credentials for management device
	my $ziptie_client = ZipTie::Client->new();
	my ($username, $password, $port) = get_credentials($ziptie_client, $mgmt_server_ip, $PROTOCOLS);

	# Connect to the management device and get managed device IPs, also get a hash of any pre existing devices in ZTS
	$LOGGER->debug("Connecting to Firepower Management Center with username '$username' and port '$port'.");
	my $found_devices = _get_managed_devices($mgmt_server_ip, $username, $password, $port);
	$LOGGER->debug('Found ' . scalar(@$found_devices) . ' device(s) on Firepower Management Center');

	my $existing_devices = get_devices($ziptie_client);

	my $device_service = $ziptie_client->devices();

	add_or_update_devices($ziptie_client, $ADAPTER, $existing_devices, $found_devices);

	if($ENV{ENABLE_RECORDING}){
		my $recording_connection_path = ZipTie::ConnectionPath->new();
		$recording_connection_path->set_ip_address($mgmt_server_ip);
		ZipTie::Recording::get_recording()->to_file(q{Firepower_Management_Center}, q{fmcdiscovery}, $recording_connection_path);
	}

	return;
}

sub _get_managed_devices {
	my ($mgmt_server_ip, $username, $password, $port) = @_;
	my $ip_regex = get_crep('cipm');

	my $connection_path = ZipTie::ConnectionPath->new();
	$connection_path->set_management_server_ip_address($mgmt_server_ip);
	$connection_path->set_credentials(ZipTie::Credentials->new(username => $username, password => $password));
	$connection_path->set_protocols([ZipTie::ConnectionPath::Protocol->new({name => 'SSH', port => $port})]);

	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);
	$cli_protocol->connect($mgmt_server_ip, $port, $username, $password);

	# Get the prompt
	my @responses;
	push @responses, ZipTie::Response->new('(#|\$)\s*$', \&_get_prompt);
	my $response = $cli_protocol->wait_for_responses(\@responses);
	my $device_prompt_regex;
	if( $response ) {
		my $next_interaction = $response->get_next_interaction();
		$device_prompt_regex = &$next_interaction($cli_protocol, $connection_path->get_credentials());
	} else {
		_fatal('Invalid response from device encountered!');
	}
	$cli_protocol->set_prompt_by_name('prompt', $device_prompt_regex);
	
	# Run the command to look in the file with the managed devices
	my $sftunnel = $cli_protocol->send_and_wait_for('cat /etc/sf/sftunnel.conf', $device_prompt_regex);
	disconnect($cli_protocol);

	# Parse through the returned file and return an array of devices found, that we care about
	# Roles: SERVER=1,CLIENT=2,HA PRIME=3,HA SEC=4,MDC=5,FMC=6
	my @managed_devices;
	my $role;
	for my $line ( split /^/, $sftunnel ) {

		if( $line =~ /^\s+ROLE\s+(\d+)\;\s*$/i ) {
			$role = $1;
		} elsif( $role ) {
			if( $line =~ /^\s+IP\s+($ip_regex)\;\s*$/i ) {
				my $ip = $1;
				if( $role == 2 ) {
					push @managed_devices, $ip;
				} elsif( $role == 6 ) {
					$LOGGER->debug("Discovering from managed Firepower Management Center @ $ip");
					eval { execute_discovery($ip) };
				}
				$role = undef;
			}
		}
	}

	if( @managed_devices ) {
		$LOGGER->debug("Found managed devices:\n" . Dumper(\@managed_devices));
	} else {
		$LOGGER->debug('No devices found');
	}

	return \@managed_devices;
}

sub _get_prompt {
	my $cli_protocol = shift;
	my $last_match = $cli_protocol->last_match();
	$last_match =~ /([^\s][^\r\n]*)$/;

	# Clean up prompt and regi-fy
	my $device_prompt = $1;
	$device_prompt =~ s/^\s+//;
	$device_prompt =~ s/\s+$//;
	$device_prompt = quotemeta $device_prompt;

	return $device_prompt;
}

sub _fatal {
	my ($message) = @_;
	print "ERROR: $message\n";
	$LOGGER->fatal($message);
	exit 1;
}

1;
