package ZipTie::Adapters::Cisco::SecurityAppliance;

use 5.010_000;
use strict;
use warnings;

use ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin;
use ZipTie::Adapters::Cisco::SecurityAppliance::Disconnect qw(disconnect);
use ZipTie::Adapters::Cisco::SecurityAppliance::GetConfig qw(get_config);
use ZipTie::Adapters::Cisco::SecurityAppliance::Parsers qw(parse_names create_config parse_chassis parse_object_groups parse_filters parse_system parse_interfaces parse_dynamic_routes parse_static_routes parse_vlans parse_subelements parse_subelement_data);
use ZipTie::Adapters::GenericAdapter;
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::CLIProtocol;
use ZipTie::CLIProtocolFactory;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

my $LOGGER = ZipTie::Logger::get_logger();

my $SA_MORE_PROMPT_REGEX = '<?\-{2,3} ?[mM]ore ?\-{2,3}>?\s*$';

# Specifies that this adapter is a subclass of ZipTie::Adapters::BaseAdapter
use ZipTie::Adapters::BaseAdapter;
our @ISA = qw(ZipTie::Adapters::BaseAdapter);

sub backup {
	my ($package_name, $backup_doc) = @_;

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ( $connection_path ) = ZipTie::Typer::translate_document( $backup_doc, 'connectionPath' );

	# Connect to the device and get the prompt
	my ( $cli_protocol, $enable_prompt_regex ) = _connect( $connection_path );

	# Setup the model printer
	my $filehandle = get_model_filehandle( 'SecurityAppliance', $connection_path->get_ip_address() );
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

	my $responses;

	# Configure the terminal page to be unlimited in length
	$cli_protocol->send_and_wait_for( 'terminal pager 0', $enable_prompt_regex );

	# Removed - see defect 103591
	# If that fails we need to watch for a more prompt
	#if( $tpResp =~ /% Invalid input detected|for a list of available commands/ ) {
	#	$cli_protocol->set_more_prompt( '<--- More --->\s*$', '20');
	#}

	$responses->{names} = $cli_protocol->send_and_wait_for( 'show names', $enable_prompt_regex );
	$responses->{names} =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt
	parse_names($responses);

	# Now output the configuration
	$responses->{running_config} = get_config( $cli_protocol, $connection_path, 'running-config' );
	if( $responses->{running_config} =~ /% Invalid input detected|for a list of available commands/ ) {
		$LOGGER->fatal_error_code($INSUFFICIENT_PRIVILEGE, $cli_protocol->get_ip_address(), "Unable to issue 'show running-config'");
	}

	$responses->{version} = $cli_protocol->send_and_wait_for( 'show version', $enable_prompt_regex );
	$responses->{version} =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt
	parse_system( $responses, $printer );
	parse_chassis( $responses, $printer );

	# Is this a transparent firewall?
	if( $responses->{running_config} =~ /^firewall transparent/mi ) {  
		$responses->{firewallTransparent} = 1;
	}

	$responses->{show_access_list} = $cli_protocol->send_and_wait_for( 'show access-list', $enable_prompt_regex );

	# Crypto maps only exist on single context devices, only need to issue this command once.
	$responses->{crypto_ipsec_sa} = $cli_protocol->send_and_wait_for('show crypto ipsec sa', $enable_prompt_regex);

	# Gather information on eigrp routes
	$responses->{show_route} = $cli_protocol->send_and_wait_for('show route', $enable_prompt_regex);
	$responses->{show_eigrp_topology} = $cli_protocol->send_and_wait_for('show eigrp topology', $enable_prompt_regex);
	$responses->{show_eigrp_neighbors} = $cli_protocol->send_and_wait_for('show eigrp neighbors', $enable_prompt_regex);

	$responses->{show_shun} = $cli_protocol->send_and_wait_for('show shun', $enable_prompt_regex);
	$responses->{show_shun} =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt

	# change to system context where available (FWSMs)
	my $changeSystem = $cli_protocol->send_and_wait_for( 'changeto system', $enable_prompt_regex.'|[#>]\s*$' );
	
	# If the prompt changed from what it was, then we made something happen.
	if ($changeSystem !~ /$enable_prompt_regex/)
	{
		my $systemPrompt = ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin::_get_prompt( $cli_protocol );
		$LOGGER->debug("Found system prompt '$systemPrompt'");

		$responses->{na_running_config} = get_config( $cli_protocol, $connection_path, 'running-config', $systemPrompt );
		$responses->{na_startup_config} = get_config( $cli_protocol, $connection_path, 'startup-config', $systemPrompt );

		# Collect context firewall modes
		$responses->{show_firewall} = $cli_protocol->send_and_wait_for( 'show firewall', $systemPrompt );

		my ($adminContext) = $responses->{na_running_config} =~ /^admin-context\s+(\S+)/mi;
		if ($adminContext) 
		{
			# The top-level parsing logic for user-level security contexts will simply create sufficient, er,
			# internal context in $responses to be able to allow us to scan for them
			parse_subelements($responses, $printer);
				
			foreach my $uContext ( @{$responses->{userContexts}} )
			{
				# pass on the admin context, we're have that as the top level config elements ane are or will be
				# good with that, it's user contexts we're after here
				if( $uContext->{name} eq $adminContext ) {
					next;
				}
				
				# context prompt should be a slash followed by the context name
				my $uContextPrompt = "\\/$uContext->{name}";
				my $changeSystem = $cli_protocol->send_and_wait_for( "changeto context $uContext->{name}", $uContextPrompt  );
				
				if ( $changeSystem !~ /invalid/i)
				{
					# Names are inherited from the admin context
					$uContext->{names} = $cli_protocol->send_and_wait_for( 'show names', $uContextPrompt );
					$uContext->{show_access_list} = $cli_protocol->send_and_wait_for( 'show access-list', $uContextPrompt );
					$uContext->{running_config} = get_config( $cli_protocol, $connection_path, 'running-config', $uContextPrompt );

					# Collect interfaces in either of two ways....
					$uContext->{interfaces} = $cli_protocol->send_and_wait_for( "show interface detail", $uContextPrompt );
					if ($uContext->{interfaces} =~ /incorrect|invalid input|Usage:|% Error/i ) {
						$uContext->{interfaces} = $cli_protocol->send_and_wait_for( "show interface", $uContextPrompt );						
					}
					$uContext->{show_shun} = $cli_protocol->send_and_wait_for('show shun', $uContextPrompt);

					for ($uContext->{names}, $uContext->{interfaces}, $uContext->{show_shun})
					{
						$_ =~ s/$uContextPrompt//;
						$_ =~ s/<---\s+More\s+--->\s*//mig;			
					}						
					# Store off the running config to be able to create a composite return configuration
					# store element...
					$LOGGER->debug("SecurityAppliance::backup() : storing config for context $uContext->{name} ");			
					$responses->{context_configs}->{$uContext->{name}} = $uContext->{running_config};	
				}
				else
				{
					$LOGGER->debug("SecurityAppliance::backup() : ** unable to changeto context $uContext->{name} ; error: " .
						$changeSystem);
				}
				# Either way, back to the changeable point of system context +++waynet is it "changeto context admin"
				# I actually have to issue here?
				$cli_protocol->send_and_wait_for( 'changeto system', $enable_prompt_regex.'|[#>]\s*$' );				
			} 
			
			$LOGGER->debug("SecurityAppliance::backup() : changing to admin-context \'$adminContext\'");			
			$cli_protocol->send_and_wait_for( 'changeto context '.$adminContext, $enable_prompt_regex );			
		}
		else
		{
			$LOGGER->debug("Avoiding admin-context change!");
			$adminContext = $responses->{na_startup_config} =~ /admin-context(.*)/mi;
			$LOGGER->debug("Found in startup config: '$adminContext'.");
			$adminContext = $responses->{na_running_config} =~ /admin-context(.*)/mi;
			$LOGGER->debug("Found in running config: '$adminContext'.");
		}
	}

	$responses->{startup_config} = get_config( $cli_protocol, $connection_path, 'startup-config' );
	create_config( $responses, $printer );
	delete $responses->{version};
	delete $responses->{startup_config};
	delete $responses->{na_running_config};
	delete $responses->{na_startup_config};
	delete $responses->{context_configs};

	$responses->{interfaces} = $cli_protocol->send_and_wait_for( "show interface detail", $enable_prompt_regex );
	if ( $responses->{interfaces} =~ /incorrect|invalid input|Usage:|% Error/mi ) {
		$responses->{interfaces} = $cli_protocol->send_and_wait_for( "show interface", $enable_prompt_regex );
	}
	$responses->{interfaces} =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt

	$LOGGER->debug("Parse Object Groups.");	
	parse_object_groups( $responses, $printer );
	parse_dynamic_routes( $responses, $printer );
	$LOGGER->debug("Parse Filters.");	
	parse_filters( $responses, $printer );

	delete $responses->{names};

	$LOGGER->debug("Parse Interfaces.");	
	parse_interfaces( $responses, $printer );
	delete $responses->{interfaces};
	$LOGGER->debug("Parse Static Routes.");	
	parse_static_routes( $responses, $printer );

	# Go back to parse the subelement data collected previously
	$LOGGER->debug("Parse Subelement.");
	parse_subelement_data($responses, $printer);
	delete $responses->{userContexts};
	delete $responses->{running_config};

	# Vlan information was parsed in parse_interafces, but is printed out here.
	$LOGGER->debug("Parse Vlans.");
	parse_vlans($responses, $printer);
	delete $responses->{_vlans};

	$printer->close_model();
	close_model_filehandle($filehandle);
	disconnect($cli_protocol);

	return;
}

sub _connect {
	my $connection_path = shift;

	# Create a new CLI protocol object
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);
	#$cli_protocol->set_more_prompt( $SA_MORE_PROMPT_REGEX, '20');

	# Make a connection to and successfully authenticate with the Security Appliance device
	my $enable_prompt_regex = ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin::execute( $cli_protocol, $connection_path );

	# Store the found prompt as "enablePrompt" on the specified CLI protocol.
	$cli_protocol->set_prompt_by_name( "enablePrompt", $enable_prompt_regex );
	
	# Return the created ZipTie::CLIProtocol object and the enable prompt encountered after successfully connecting to a device.
	return ( $cli_protocol, $enable_prompt_regex );
}

sub _get_file_systems
{
	# give the output of "show file systems", this method returns
	# and array of the file systems
	my $show_file_systems_response = shift;
	my @fs;
	while ( $show_file_systems_response =~ /^\*?\s+\d+\s+\d+\s+\b(\S+)\b\s+[A-Za-z]+\s+(\S+):/mg )
	{
		my $type = $1;
		my $name = $2;
		if ( $type !~ /opaque|nvram/i )
		{
			push( @fs, $name );
		}
	}
	return @fs;
}

1;
