package ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin;

use strict;
use warnings;
use 5.010_000;

use ZipTie::CLIProtocol;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Recording;
use ZipTie::Response;

my $LOGGER = ZipTie::Logger::get_logger();

sub execute {
	my ($cli_protocol, $connection_path) = @_;

	my $credentials = $connection_path->get_credentials();

	# Store all the information needed to connect to the device
	my $ip_address = $connection_path->get_ip_address();
	my $port       = $cli_protocol->get_port();
	my $username   = $credentials->{username};
	my $password   = $credentials->{password};

	my $protocol_name = $cli_protocol->get_protocol_name();
	$LOGGER->debug("Protocol in use: $protocol_name");

	# Attempt to connect to the device
	$cli_protocol->connect( $ip_address, $port, $username, $password );
	$LOGGER->debug('Verifying the initial connection ...');

	# If all goes well during the login process, the last method to be called will be "_get_prompt"
	# and its return value will be returned all the way to here
	my $device_prompt_regex = _initial_connection( $cli_protocol, $credentials );

	$LOGGER->debug("Login has successfully completed!\n");

	return $device_prompt_regex;
}

sub _initial_connection {
	my ($cli_protocol, $credentials) = @_;

	# Specify the responses to check for
	my @responses = ();
	push @responses, ZipTie::Response->new( 'sername:|^\s*[Ll]ogin:', \&_send_username );
	push @responses, ZipTie::Response->new( '[Pp]assword:',           \&_send_password );
	push @responses, ZipTie::Response->new( '>\s*$',                  \&_send_enable );
	push @responses, ZipTie::Response->new( '#\s*$',                  \&_get_prompt ); # auto-enable case

	my $response = $cli_protocol->wait_for_responses( \@responses );

	if( not $response ) {
		$LOGGER->fatal('Invalid response from device encountered!');
	}

	# Based on the response of the device, determine the next interaction that should be executed.
	my $next_interaction = $response->get_next_interaction();
	return &$next_interaction( $cli_protocol, $credentials );
}

sub _send_username {
	my ($cli_protocol, $credentials) = @_;

	# Specify the responses to check for
	my @responses = ();
	push @responses, ZipTie::Response->new( '[Pp]assword:', \&_send_password );
	push @responses, ZipTie::Response->new( 'incorrect',      undef, $INVALID_CREDENTIALS );
	push @responses, ZipTie::Response->new( 'sername:|ogin:', undef, $INVALID_CREDENTIALS );

	# Send username credential
	my $username = $credentials->{username};
	$LOGGER->debug("Sending username credential ...");
	$cli_protocol->send($username, 0);
	my $response = $cli_protocol->wait_for_responses( \@responses );

	if( not $response ) {
		$LOGGER->fatal('Invalid response from device encountered!');
	}

	my $next_interaction = $response->get_next_interaction();
	return &$next_interaction( $cli_protocol, $credentials );
}

sub _send_password {
	my ($cli_protocol, $credentials) = @_;

	# Specify the responses to check for
	my @responses = ();
	push @responses, ZipTie::Response->new( 'sername:|ogin:',    undef, $INVALID_CREDENTIALS );
	push @responses, ZipTie::Response->new( 'assword:',          undef, $INVALID_CREDENTIALS );
	push @responses, ZipTie::Response->new( 'incorrect|invalid', undef, $INVALID_CREDENTIALS );
	push @responses, ZipTie::Response->new( '>\s*$',      \&_send_enable );
	push @responses, ZipTie::Response->new( '(#|\$)\s*$', \&_get_prompt );

	# Send username credential
	my $password = $credentials->{password};
	$LOGGER->debug("Sending password credential ...");
	$cli_protocol->send($password, 1);
	my $response = $cli_protocol->wait_for_responses( \@responses );

	if( not $response ) {
		$LOGGER->fatal('Invalid response from device encountered!');
	}

	my $next_interaction = $response->get_next_interaction();
	return &$next_interaction( $cli_protocol, $credentials );
}

sub _send_enable {
	my ($cli_protocol, $credentials) = @_;
	
	if ( $credentials->{enablePassword} ) {    
		# Specify the responses to check for
		my @responses = ();
		push @responses, ZipTie::Response->new( 'incorrect|invalid', undef, $INVALID_CREDENTIALS );
		push @responses, ZipTie::Response->new( '[Pp]assword:', \&_send_enable_password );
		push @responses, ZipTie::Response->new( '(#|\$)\s*$',   \&_get_prompt );

		# First look for the enable username and if valid level send 'enable <level>'
		if( $credentials->{enableUsername} && $credentials->{enableUsername} =~ /^level-([1-9]|1[0-5])$/){
				my $enableLevel = $1;
				$cli_protocol->send("enable $enableLevel");
		} else {
			$cli_protocol->send('enable'); # default to 'enable'
		}
		
		my $response = $cli_protocol->wait_for_responses( \@responses );

		if( not $response ) {
			$LOGGER->fatal("Invalid response from device encountered!");
		}

		my $next_interaction = $response->get_next_interaction();
		return &$next_interaction( $cli_protocol, $credentials );
	}

	return _get_prompt($cli_protocol);
}

sub _send_enable_password {
	my ($cli_protocol, $credentials) = @_;

	# Specify the responses to check for
	my @responses = ();
	push( @responses, ZipTie::Response->new( '[Pp]assword:|incorrect|invalid', undef, $INVALID_CREDENTIALS ) );
	push( @responses, ZipTie::Response->new( '(#|\$)\s*$', \&_get_prompt ) );

	# Send password credential
	my $password = $credentials->{enablePassword};
	$LOGGER->debug("Sending enable password credential ...");
	$cli_protocol->send($password, 1);
	my $response = $cli_protocol->wait_for_responses( \@responses );

	if( not $response ) {
		$LOGGER->fatal('Invalid response from device encountered!');
	}

	my $next_interaction = $response->get_next_interaction();
	return &$next_interaction( $cli_protocol, $credentials );
}

sub _get_prompt {
	my $cli_protocol = shift;

	# Grab the last match that the CLIProtocol implementation captured
	my $last_match = $cli_protocol->last_match();
	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[LAST MATCH CAPTURED]");
	$LOGGER->debug($last_match);

	# Use a regular expression to strip the prompt from the match
	$last_match =~ /([^\s][^\r\n]*)$/;

	# Grab the extracted device prompt
	my $device_prompt = $1;

	# Register the device prompt with the ZipTie::Recording module
	my $recording = ZipTie::Recording::get_recording();
	$recording->set_device_prompt($device_prompt);

	# Clean up the device prompt for regular expression use
	$device_prompt =~ s/^\s+//;
	$device_prompt =~ s/\s+$//;
	$device_prompt = quotemeta $device_prompt;

	# Construct the regular expression for the prompt so that it will only match the prompt that is displayed
	# as the very last thing.  This will get around the issue of a device possibly echoing back it's prompt when
	# a command is sent.
	my $devicePromptRegEx = $device_prompt . "\\s*\$";

	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[REGULAR EXPRESSION TO MATCH DEVICE PROMPT]");
	$LOGGER->debug($devicePromptRegEx);
	$LOGGER->debug("---------------------------------------------------------");

	return $devicePromptRegEx;
}

1;
