package ZipTie::Adapters::Cisco::SecurityAppliance::Telemetry;

use 5.010_000;
use strict;
use warnings;

use ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin;
use ZipTie::Adapters::Cisco::SecurityAppliance::Disconnect qw(disconnect);
use ZipTie::Adapters::Cisco::SecurityAppliance::TelemetryParsers qw(parse_arp parse_telemetry_interfaces parse_mac_table parse_routing_neighbors);
use ZipTie::Adapters::Utils qw(choose_admin_ip get_model_filehandle close_model_filehandle);
use ZipTie::CLIProtocolFactory;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

my $LOGGER = ZipTie::Logger::get_logger();

sub invoke {
	my ($pkg, $connection_path_doc) = @_;

	# Create CLIProtocol object for communicating
	my ($connection_path, $discoveryParams) = ZipTie::Typer::translate_document( $connection_path_doc, 'connectionPath' );
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	# Connect
	my $prompt = ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin::execute( $cli_protocol, $connection_path );

	# Setup the model printer
	my $filehandle = get_model_filehandle( "SecurityAppliance", $connection_path->get_ip_address() );
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'telemetry');

	# Attempt to turn off paging and if that fails set the more prompt
	my $paging_response = $cli_protocol->send_and_wait_for('terminal pager 0', $prompt);       
	if( $paging_response =~ /% Invalid input detected|for a list of available commands/ ) {
		$cli_protocol->set_more_prompt( '<--- More --->\s*$', '20');
	}

	# Get responses from the main context
	my $responses = _get_responses( $cli_protocol, $prompt );

	# Get responses from possible subcontexts
	my $changeSystem = $cli_protocol->send_and_wait_for('changeto system', $prompt . '|[#>]\s*$');
	if( $changeSystem !~ /$prompt/ ) {

		my $is_subcontext = 1;

		my $system_prompt = ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin::_get_prompt($cli_protocol);

		my $show_context = $cli_protocol->send_and_wait_for('show context', $system_prompt);
		while( $show_context =~ /^\s{1,4}([^\*\s#]+)\s(?!\s*active Security Contexts)(?!\s*Name\s+Class)/img ) {
			my $context = $1;
			my $context_prompt = "\\/\Q$context\E[#>]\\s*\$";
			$cli_protocol->send_and_wait_for("changeto context $context", $context_prompt);
			$responses->{context}->{$context} = _get_responses( $cli_protocol, $context_prompt, $is_subcontext );
		}
	}


	# Start parsing and printing
	$printer->attributes(1);
	$printer->open_discovery_event();

	$printer->open_element('neighbors');

	parse_arp($responses, $printer);
	parse_mac_table($responses, $printer);
	parse_routing_neighbors($responses, $printer);
	
	$printer->close_element('neighbors');

	my $interfaces = parse_telemetry_interfaces($responses, $printer);
	
	my $adminIp = $connection_path->get_ip_address();
	if( $discoveryParams->{calculateAdminIp} eq 'true' ) {
		$adminIp = choose_admin_ip($connection_path->get_ip_address(), $interfaces);
	}
	$printer->print_element( 'adminIp', $adminIp );

	if( $responses->{context} ) {
		$printer->open_element('subdevices');

		for my $context ( sort keys %{ $responses->{context} } ) {
			my $context_responses = $responses->{context}->{$context};

			$printer->open_element('subdevice');
			$printer->print_element('subdeviceName', $context);

			$printer->open_element('neighbors');

			parse_arp($context_responses, $printer);
			parse_mac_table($context_responses, $printer);
			parse_routing_neighbors($context_responses, $printer);

			$printer->close_element('neighbors');

			parse_telemetry_interfaces($context_responses, $printer);

			$printer->close_element('subdevice');
		}

		$printer->close_element('subdevices');
	}

	# Cleanup	
	$printer->close_element('DiscoveryEvent');
	close_model_filehandle($filehandle);
	disconnect($cli_protocol);

	return;
}

sub _get_responses {
	my ($cli_protocol, $prompt, $is_subcontext) = @_;

	my @commands = (
		'show arp',
		'show mac-address-table',
		'show ospf neighbor',
		'show route',
		'show interface',
#		'show ipv6 interface'
	);

	my $responses;
	for my $command ( @commands ) {
		next if $command eq 'show ospf neighbors' and $is_subcontext; # ospf is not available in subcontexts

		my $key = $command;
		$key =~ s/ |-/_/g;
		$responses->{$key} = $cli_protocol->send_and_wait_for($command, $prompt);
	}

	return $responses;
}

1;
