package ZipTie::Adapters::Cisco::SecurityAppliance::TelemetryParsers;

use 5.010_000;
use strict;
use warnings;

use ZipTie::Adapters::Cisco::SecurityAppliance::Utils qw(get_asa_interface_type);
use ZipTie::Adapters::Utils qw(mask_to_bits strip_mac get_crep);

use Exporter qw(import);
our @EXPORT_OK = qw(parse_arp parse_telemetry_interfaces parse_mac_table parse_routing_neighbors);

my $LOGGER = ZipTie::Logger::get_logger();

my $IP_REGEX = get_crep('cipm');
my $V6_REGEX = get_crep('ipv6');

sub parse_routing_neighbors {
	my ( $in, $out ) = @_;
	my @neighbors;

	while( $in->{show_ospf_neighbor} =~ /^\s*([\da-f.:]+)\s+\d+\s+\S+(?:\s+\S+)?\s+\S+\s+($IP_REGEX)\s+(\S+)\s*$/gm ) {
		push @neighbors, {
			protocol  => 'OSPF',	
			routerId  => $1,	
			ipAddress => $2,	
			interface => $3
		};
	}

	my %protocol_codes = ( D => 'EIGRP', R => 'RIP', B => 'BGP', i => 'ISIS', I => 'IGRP', S => 'STATIC' );
	while( $in->{show_route} =~ /^(I|R|B|D|i|S)\*?\s+[\da-f:.]+\s+(?:[\da-f:.]+)\s+\S+\s+via\s+([\da-f:.]+),.*?(\S+)\s*$/gm ) {
		my ($code, $address, $interface) = ($1, $2, $3);

		my $protocol = $protocol_codes{$code};

		push @neighbors, {
			protocol  => $protocol,
			ipAddress => $address,
                        interface => $interface
		};
	}

	if( @neighbors ) {
		$out->print_element( 'routingNeighbors', { routingNeighbor => \@neighbors } );
	}

	return;
}

sub parse_arp {
	my ( $in, $out ) = @_;
	my @arps;

	while( $in->{show_arp} =~ /^\s*(\S+)\s+($IP_REGEX)\s+(\S+)/mg ) {
		push @arps, {
			interface  => $1,
			ipAddress  => $2,
			macAddress => strip_mac($3)
		};
	}

	if( @arps ) {
		$out->print_element( 'arpTable', { 'arpEntry' => \@arps } );
	}

	return;
}

sub parse_mac_table {
	my ( $in, $out ) = @_;
	my @macs;

	# Skip past the column headers
	$in->{show_mac_address_table} =~ /^\s*-{10,}\s*$/mg;

	# Example mac table entry taken from the Cisco documentation:
	# outside                 0009.7cbe.2100                 static               -

	# Example mac table entry taken from a device with bridge groups
	# inside                     0050.5691.664c          dynamic      4         1
	while( $in->{show_mac_address_table} =~ /^\s*(\S+)\s+(\S+)/mg ) {
		my $interface = $1;
		my $mac = strip_mac($2);
		push @macs, {
			macAddress => $mac,
			interface  => $interface
		};
	}

	if( @macs ) {
		$out->print_element('macTable', { macEntry => \@macs } );
	}

	return;
}

sub parse_telemetry_interfaces {
	my ( $in, $out ) = @_;

	my @interfaces = ();

	while ( $in->{show_interface} =~ /(\S+)\s+"([^"]+)",?\s+(is.+?)(?=^\S)/msg ) {
		my ($type, $name, $blob) = ($1, $2, $3);

		my $interface = { 
			name => $name,
			type => get_asa_interface_type($type, $blob, $in, $name),
			inputBytes => 0
		};

		if( $blob =~ /line protocol is (down|up)/mi ) {
			$interface->{operStatus}= ucfirst $1;
		}

		if( $blob =~ /\d+ packets input,\s+(\d+)\s*bytes/mi ) {
			$interface->{inputBytes} = $1;
		}

		while( $blob =~ /IP\s+address\s+($IP_REGEX),\s+subnet\s+mask\s+($IP_REGEX)/mgi ) {
			push @{ $interface->{ipEntry} }, {
				ipAddress => $1,	
				mask => mask_to_bits($2)
			};
		}
		push @interfaces, $interface;
	}

	if( @interfaces ) {
		$out->print_element( 'interfaces', { interface => \@interfaces } );
	}

	# Return the interfaces for choose_admin_ip
	return { interface => \@interfaces };
}

1;
