package ZipTie::Adapters::Cisco::SecurityAppliance::Utils;

use strict;
use warnings;

use ZipTie::Adapters::Utils qw(get_interface_type);
use Exporter 'import';
our @EXPORT_OK = qw(get_asa_interface_type break_network_object_into_hosts_subnets_ranges);

# The following acts as a front-end, of sorts, to Utils::get_interface_type.
# it determines if there's something PIX/ASA specific to trigger a typing
# that we'd want to see here.
#
# PARAMS:
#   $type - "type" of interface, more likely interface name such as ethernet0, 
#        serial0.0/0, those sorts of things.
#   $interaceDetail - section of 'get interfaces full' that gives full device
#         definition to this interface.
#   $confiParams - output] from startup/running config 
#
# RETURNS:
#      returns a string, with an exact SED-assignable interface type value
#          from ONM 'interfaceType' enumeration.
#
sub get_asa_interface_type {
	my ($type, $interfaceDetail, $configParams, $name) = @_;

	# First, let's see if there's anything in the interface detail that's a 
	# giveaway here...
	if( $interfaceDetail ) {
		if( $interfaceDetail =~ /^\s+management\-only\sinterface/mi ) {
			return 'softwareLoopback';
		}
	}
	
	# NOTE: _get_asa_interface_type is called both during telemetry collection and
	# during a backup.  When collecting telemetry the running_config is
	# not available.  This means the failover link will not be set to
	# softwareLoopback during telemetry collection, but this should be
	# ok because topology does not use telemetry to build its model.
	if( $configParams->{running_config} ) {
		# Check for failover link.  This can be either cable-based or LAN-based.  The statements 
		# we're looking for assign a name which is logical in the context of the failover config
		# to the *physical* interface (name passed) in the context of subsequent failover
		# configuration.  That logical name is the "\S+" below
		if( $configParams->{running_config} =~ /^failover (link\s\S+|lan\sinterface\s\S+) $type/m ) {
			# This is a failover link, which we're currently using as a 
			# "loopback" inteface, to mitigate ambitious attempts to use
			# for routing/bridged traffic egress
			return 'softwareLoopback';
		}
	}

	if( $type =~ /^Internal-Control/mi ) {
		return 'softwareLoopback';
	}

	if( $name and $name eq '_internal_loopback' ) {
		return 'softwareLoopback';
	}

	# bvi is a type specific to the ASA
	if( $type =~ /^bvi/mi ) {
		return 'virtual';
	}

	return get_interface_type($type);
}

sub break_network_object_into_hosts_subnets_ranges
{
	my ($networkGroups, $id, $seen) = @_;

	my $hosts = [];
	my $subnets = [];
	my $ranges = [];

	# Seen will be used to keep track of the object groups we have already broken.
	if (!defined($seen))
	{
		$seen = {};
	}

	# If we have already boke this object group then return now, otherwise we may recurse forever.
	if (defined($seen->{$id}))
	{
		return ($hosts, $ranges);
	}

	# Mark this object group as seen.
	$seen->{$id} = 1;

	# If this isn't a network group, return empty arrays.
	if (!defined($networkGroups->{$id}))
	{
		return ($hosts, $ranges);
	}

	# Add hosts.
	if (defined($networkGroups->{$id}->{host}))
	{
		push(@{$hosts}, @{$networkGroups->{$id}->{host}});
	}

	# Add subnets.
	if (defined($networkGroups->{$id}->{subnet}))
	{
		push(@{$subnets}, @{$networkGroups->{$id}->{subnet}});
	}

	# Add ranges.
	if (defined($networkGroups->{$id}->{range}))
	{
		push(@{$ranges}, @{$networkGroups->{$id}->{range}});
	}

	# If it's an object group reference, then break that into hosts and ranges, and add them.
	if (defined($networkGroups->{$id}->{objectGroupReference}))
	{
		foreach my $objectGroupReference (@{$networkGroups->{$id}->{objectGroupReference}})
		{
			my ($refHosts, $refSubnets, $refRanges) = break_network_object_into_hosts_subnets_ranges($networkGroups, $objectGroupReference, $seen);
			push(@{$hosts}, @{$refHosts});
			push(@{$subnets}, @{$refSubnets});
			push(@{$ranges}, @{$refRanges});
		}
	}

	return ($hosts, $subnets, $ranges);
}

1;
