package ZipTie::Adapters::Juniper::JUNOS::ParseInterfaces;

use strict;
use warnings;
use 5.010_000;

use XML::LibXML;
use ZipTie::Adapters::Juniper::JUNOS::Parsers qw(get_dom_from_xml);
use ZipTie::Adapters::Utils qw(get_crep);
use ZipTie::Logger;

use Exporter qw(import);
our @EXPORT_OK = qw(parse_interfaces _get_config_interfaces);

my $LOGGER = ZipTie::Logger::get_logger();

my $IP_REGEX = get_crep('cipm');

# Raw interfaces are the current runnnig interfaces
# Config interfaces are interface defined in the config
#
# For each physical interface
#  parse and write phyiscal interface (No layer 3 data)
#  parse and write logical interfaces using config data (called units)
sub parse_interfaces {
	my ($in, $out) = @_;

	my $is_firewall_enabled = 1;
	if( $in->{showFirewallPlain} and $in->{showFirewallPlain} =~ /Warning: configuration block ignored/ ) {
		$is_firewall_enabled = 0;
	}

	# Parses 'show interfaces' and returns an array of XPathContexts
	my ($raw_xpc, $raw_interfaces) =  _get_raw_interfaces($in);

	# Parses 'show configuration interfaces' and returns an array of XPathContexts
	my $config_interfaces = _get_config_interfaces($in);

	$out->open_element('interfaces');
	for my $raw_interface ( @{ $raw_interfaces } ) {

		my $name = $raw_xpc->findvalue('./if:name', $raw_interface);

		my $status = $raw_xpc->findvalue('./if:oper-status', $raw_interface) || 'N/A';

		my $type = $raw_xpc->findvalue('./if:link-level-type', $raw_interface) || $raw_xpc->findvalue('./if:if-type', $raw_interface);

		_parse_physical_interface($out, $raw_xpc, $raw_interface, $status, $type, $name);

		_parse_logical_interfaces($in, $out, $raw_xpc, $raw_interface, $config_interfaces, $status, $type, $name, $is_firewall_enabled);
	}
	$out->close_element('interfaces');

	return;
}

sub _get_raw_interfaces {
	my ($in) = @_;

	my ($xml) = $in->{interfaces} =~ /(<rpc-reply.+?<\/rpc-reply>)/sm;
	my ($interface_namespace) = $xml =~ /interface-information\s+xmlns="([^"]+)"/;

	my $dom = get_dom_from_xml($xml, 'show interfaces');

	my $xpc = XML::LibXML::XPathContext->new($dom);
	$xpc->registerNs('if',  $interface_namespace);

	my @raw_interfaces = $xpc->findnodes('/rpc-reply/if:interface-information/if:physical-interface');

	return ($xpc, \@raw_interfaces);
}

# Needs to now return a hash of physical names to array of units
sub _get_config_interfaces {
	my ($in) = @_;

	my ($xml) = $in->{showConfigInterfaces} =~ /(<rpc-reply.+?<\/rpc-reply>)/sm;

	my $dom = get_dom_from_xml($xml, 'show configuration interfaces');

	my ($config_interfaces) = $dom->findnodes('/rpc-reply/configuration/interfaces');

	return $config_interfaces;
}

sub _parse_physical_interface {
	my ($out, $xpc, $raw_interface, $status, $type, $name) = @_;

	my $model = {
		adminStatus => $status,
		name => $name,
		physical => 'true'
	};

	# snmp-index
	my $index = $xpc->findvalue('./if:snmp-index', $raw_interface);
	if( $index ) {
		$model->{ifIndex} = $index;
	}

	my $mac = $xpc->findvalue('./if:hardware-physical-address', $raw_interface);
	if( $type and $type =~ /ethernet/i and $mac ) {
		$mac =~ s/://g;
		$model->{interfaceEthernet} = { macAddress => $mac };
	}

	$model->{interfaceType} = _resolve_interface_type($type);

	my $mtu = $xpc->findvalue('./if:mtu', $raw_interface);
	if( $mtu ) {
		$model->{mtu} = _resolve_mtu($mtu);
	}

	my $speed = _process_speed_rating($xpc->findvalue('./if:speed', $raw_interface));
	if( $speed ) {
		$model->{speed} = $speed;
	}

	$out->print_element('interface', $model);

	return;
}

sub _parse_logical_interfaces {
	my ($in, $out, $raw_xpc, $raw_interface, $config_interfaces, $status, $type, $parent_name, $is_firewall_enabled) = @_;

	my @units = $config_interfaces->findnodes("./interface[name='$parent_name']/unit");
	for my $unit ( @units ) {

		my $name = $parent_name . '.' . $unit->findvalue('./name');

		my $model = {
			adminStatus => $status,
			name => $name,
			interfaceType => _resolve_interface_type($type),
			physical => 'false'
		};

		# Only attach filters to interfaces if firewall is active
		if( $is_firewall_enabled ) {
			my $egress_filter = $unit->findvalue('./family/inet/filter/output/filter-name')
			                 || $unit->findvalue('./family/inet/filter/output');
			if($egress_filter){
				$model->{egressFilter} = $egress_filter;
			}

			my $ingress_filter = $unit->findvalue('./family/inet/filter/input/filter-name')
			                  || $unit->findvalue('./family/inet/filter/input');
			if($ingress_filter){
				$model->{ingressFilter} = $ingress_filter;
			}	
		}
		
		my $index = $raw_xpc->findvalue("./if:logical-interface[if:name='$name']/if:snmp-index", $raw_interface) ||
			    $raw_xpc->findvalue('./if:snmp-index', $raw_interface);	
		if( $index ) {
			$model->{ifIndex} = $index;
		}
	
		if( $in->{parsed} and $in->{parsed}->{zones} ) {

			my $zones = $in->{parsed}->{zones};
			for my $zone_name ( keys %{ $zones } ) {

				my $zone = $zones->{$zone_name};
				for my $zone_interface ( @{ $zone->{interfaces} } ) {

					if( $name eq $zone_interface ) {
						push @{ $model->{interfaceGrouping} }, {
							ifGroupName => $zone_name,
							ifGroupingCharacteristic => 'securityZoneMember'
						};
					}
				}
			}
		}

		my $ip_addresses = _get_ip_addresses($name, $raw_xpc, $raw_interface, $unit);
		if( $ip_addresses ) {
			my @ip_configurations;
			for my $address ( @{ $ip_addresses } ) {

				my ($ip, $mask) = split /\//, $address;

				push @ip_configurations, {
					ipAddress => $ip,
					mask => $mask
				};

				my $subnet_junos = new ZipTie::Adapters::Juniper::JUNOS::Addressing::SubnetJunos($ip, $mask);
				push @{ $in->{IFIpForwarding}->{$name} }, $subnet_junos;
			}
			$model->{interfaceIp}->{ipConfiguration} = \@ip_configurations;
		}

		my ($mtu_node) = $raw_xpc->findnodes("./if:logical-interface[if:name='$name']/if:address-family/if:mtu", $raw_interface);
		if( $mtu_node ) {
			my $mtu = $mtu_node->to_literal();
			$model->{mtu} = _resolve_mtu($mtu);
		}

		$model->{speed} = _process_speed_rating($unit->findvalue('./bandwidth'));

		$out->print_element('interface', $model);
	}

	return;
}

sub _get_ip_addresses {
	my ($name, $raw_xpc, $raw_interface, $unit) = @_;
	my $addresses;

	my @address_nodes = $unit->findnodes('./family/inet/address/name');
	for my $address_node ( @address_nodes ) {
		my $address = $address_node->textContent();
		$address =~ s/^\s+|\s+$//g;
		push @{ $addresses }, $address;
	}
	
	if( not $addresses and $unit->exists('./family/inet/dhcp-client') ) {
		my @raw_ip_nodes = $raw_xpc->findnodes("./if:logical-interface[if:name='$name']/if:address-family/if:interface-address", $raw_interface);

		for my $raw_ip_node ( @raw_ip_nodes ) {
			my $address = $raw_xpc->findvalue('./if:ifa-local', $raw_ip_node);
			my $network = $raw_xpc->findvalue('./if:ifa-destination', $raw_ip_node);
			my $mask = 32;
			if( $network =~ /\/(\d+)/ ) {
				$mask = $1;
			}
			if( $address ) {
				push @{ $addresses }, "$address/$mask";
			}
		}
	}

	my @ipv6_addresses;
	my @ipv6_address_nodes = $unit->findnodes('./family/inet6/address/name');
	for my $address_node ( @ipv6_address_nodes ) {
		my $address = $address_node->textContent();
		$address =~ s/^\s+|\s+$//g;
		push @ipv6_addresses, $address;
	}
	if( @ipv6_addresses ) {
		$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping interface IP addresses: " . join ', ', @ipv6_addresses); 
	}

	return $addresses;
}

#T1= 1544kbps
#T2=6312 kbps 
#T3= 44736 kbps 
#
#Relevant to Australia are:
#E1 = 2,048 Mbit/s
#E2 = 4 x E1 = 8,448 Mbit/s
#E3 = 4 x E2 = 34,368 Mbit/s
#E4 = 4 x E3 = 139,264 Mbit/s
#E5 = 4 x E4 = 565 Mbit/
sub _process_speed_rating{
	my $speed = shift;

	if( $speed =~ /^\d+$/ ) {
		return $speed;

	} elsif( $speed =~ /(\d+)m/i ) {
		return $1 * 1000 * 1000;

	} elsif( $speed =~ /(\d+)k/i ) {
		return $1 * 1000;

	} elsif( $speed =~ /T3/i ) {
		return 44736000;

	} elsif( $speed =~ /T2/i ) {
		return 6312000;

	} elsif( $speed =~ /T1/i ) {
		return 1544000;

	} elsif( $speed =~ /E1/i ) {
		return 2048000;

	} elsif( $speed =~ /E2/i ) {
		return 8448000;

	} elsif( $speed =~ /E3/i ) {
		return 34368000;

	} elsif( $speed =~ /E4/i ) {
		return 139264000;

	} elsif( $speed =~ /E5/i ) {
		return 565000000;

	} elsif( $speed =~ /(\d+)/i ) {
		return $1;

	} elsif( $speed eq 'Unlimited' ) {
		# SNMP values Unlimited at 2147483647
		return '2147483647';
	}
	
	return 0;
}

sub _resolve_mtu {
	my ($mtu) = @_;

	$mtu = $mtu eq 'Unlimited' ? 2147483647 : $mtu;

	return $mtu;
}

sub _resolve_interface_type {
	my ($type) = @_;

	$type =   $type =~ /ethernet/i ? 'ethernet'
		: $type =~ /loopback/i ? 'softwareLoopback'
		: $type =~ /frame/i    ? 'frameRelay'
		: $type =~ /sonet/i    ? 'sonet'
		: $type =~ /token/i    ? 'tokenRing'
		: $type =~ /atm/i      ? 'atm'
		: $type =~ /ppp/i      ? 'ppp'
		:                        'other';

	return $type;
}

1;
