package ZipTie::Adapters::Juniper::JUNOS::Parsers;

use strict;
use warnings;
use 5.010_000;

use Data::Dumper;
use MIME::Base64 qw(encode_base64);
use Net::IP;
use XML::LibXML;
use XML::Twig;

use ZipTie::Adapters::Utils qw(seconds_since_epoch strip_mac get_mask mask_to_bits get_port_number trim get_interface_type getUnitFreeNumber parseCIDR get_port_expression get_crep ncname_from_string get_protocol_name);
use ZipTie::Adapters::Juniper::JUNOS::SecurityApps qw($securityApps);

use ZipTie::Adapters::Juniper::JUNOS::Addressing::SubnetJunos;
use ZipTie::Logger;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

use Exporter qw(import);
our @EXPORT_OK = qw(parse_dynamic_routes create_config parse_chassis parse_filters parse_static_routes parse_system parse_object_groups clear_parsing_state get_dom_from_xml);

my $ip_regex = get_crep('cipm');
my $ipv6_regex = get_crep('ipv6');

sub clear_parsing_state {
	my ($in) = @_;

	if(defined $in->{IFIpForwarding}){
		delete($in->{IFIpForwarding});
	}
}

sub parse_chassis{
	my ($in, $out) = @_;

	$out->open_element('chassis');

	my $chassisAsset = {
			'core:assetType' => 'Chassis',
			'core:factoryinfo' => {
				'core:make' => 'Juniper'
			}
	};

	if($in->{showVer} =~ /^\s*Model:\s+(\S+)/im){
		$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
	}

	my (@powerSupplies, @cards, @cpus);
	my ($hardwareSection) = $in->{showHardware} =~ /(<chassis[\s|>].+?<\/chassis>)/s;

	if($hardwareSection){
		my $twig = XML::Twig->new();
		$twig->parse($hardwareSection);
		undef($hardwareSection);

		my $rootElement = $twig->root();

		if( $rootElement->first_child('serial-number') ) {
			$chassisAsset->{'core:factoryinfo'}->{'core:serialNumber'} = $rootElement->first_child('serial-number')->text();
		} elsif( $in->{showVer} =~ /^\s*Serial\s+Number:\s+(\S+)\s*$/im ){
			$chassisAsset->{'core:factoryinfo'}->{'core:serialNumber'} = $1;
		}

		$out->print_element('core:asset', $chassisAsset);
		$out->print_element('core:description', $rootElement->first_child('description')->text());

		foreach my $chassisModule ($rootElement->descendants('chassis-module')){
			my $module = _parse_module($chassisModule);

			my $name   = $chassisModule->first_child('name')->text();
			if($name =~ /Power\s+Supply\s+(\S)/i){
				my $number = ord($1) - 64;
				$module->{'core:asset'}->{'core:assetType'} = 'PowerSupply';
				$module->{number} = $number;
				push(@powerSupplies, $module);
			}elsif($name =~ /^FPC\s+(\d+)/ ){
				$module->{'core:asset'}->{'core:assetType'} = 'Card';
				$module->{'slotNumber'} = $1;

				# Daughter cards
				foreach my $daughterCard ($chassisModule->descendants('chassis-sub-module')){
					my $dc = _parse_module($daughterCard);
					$dc->{'core:asset'}->{'core:assetType'} = 'Card';

					my $name = $daughterCard->first_child('name')->text();
					if($name =~ /(\d+)$/){
						$dc->{slotNumber} = $1;
					}

					push(@{$module->{daughterCard}}, $dc);
				}
				push(@cards, $module);
			}elsif($name =~ /^FEB/){
				$module->{'core:asset'}->{'core:assetType'} = 'CPU';
				push(@cpus, $module);
			}
		}
	}

	foreach my $card (@cards){
		$out->print_element('card', $card);
	}
	foreach my $cpu (@cpus) {
		$out->print_element('cpu', $cpu);
	}

	if($in->{showMAC} && $in->{showMAC} =~ /Public base address\s+((?:\w{2}\:){5}\w{2})/){
		$out->print_element('macAddress', strip_mac($1));
	}

	if($in->{showRen} =~ /^\s*DRAM\s+(\d+)\s+(\S+)/mi){
		my $memory = {
			'core:description' => 'RAM',
			kind               => 'RAM',
			size               => getUnitFreeNumber($1, $2, 'byte'),
		};
		$out->print_element('memory', $memory);
	}

	foreach my $power_supply (@powerSupplies){
		$out->print_element('powersupply', $power_supply);
	}

	$out->close_element('chassis');

	return;
}

sub parse_system {
	my ($in, $out) = @_;

	if($in->{showVer} =~ /^\s*Hostname:\s+(\S+)/im){
		$out->print_element('core:systemName', $1);
	}elsif($in->{snmp} =~ /<system-name>(.*)<\/system-name>/im){
		$out->print_element('core:systemName', $1);
	}

	$out->open_element('core:osInfo');

	if(defined $in->{showFirmware} and $in->{showFirmware} =~ /Juniper ROM Monitor Version\s(\d+\.\w+)/s){
		$out->print_element('core:fileName', $1);
	}

	$out->print_element('core:make', 'Juniper');

	my $version;
	if($in->{showVer} =~ /^\s*JUNOS (Base OS Software Suite|Software Release|Base Version) \[(.*)\]/im){
		$version = $2;
		$out->print_element('core:version', $version);
	}

	$out->print_element('core:osType', 'JUNOS');
	$out->close_element('core:osInfo');

	if($version){
		$out->print_element('core:biosVersion', $version);
	}

	my $applianceType = 'Firewall';
	my $verString = $in->{showVer};
	if($verString =~ /^\s*Model:\s+(\S+)/im){
		my $modelStr = $1;
		my $patHash = _get_patterns();

		foreach my $patKey (keys(%{$patHash})){
			my $pat = $patHash->{$patKey};

			if($modelStr =~ /$pat/i){
				$applianceType = $patKey;
				last;
			}
		}
	}
	$out->print_element('core:deviceType', $applianceType);

	# System booted: 2007-03-08 10:09:09 CST (6w6d 03:03 ago)
	if($in->{showUptime} =~ /^System booted:\s+(\d+)-(\d+)-(\d+)\s+(\d{1,2}:\d{1,2}:\d{1,2})\s+(\S+)/mi){
		my ($year, $month, $day, $time, $timezone) = ($1, $2, $3, $4, $5);
		my ($hour, $min, $sec) = $time =~ /(\d+):(\d+):(\d+)/;
		$out->print_element('core:lastReboot', seconds_since_epoch($sec, $min, $hour, $day, $month, $year, $timezone));
	}

	return;
}

# Populates the configuration entity for the main Juniper config
sub create_config{
	my ($in, $out) = @_;

	# The name of the repository
	my $repository = {'core:name' => '/'};

	my $active = {
		'core:name'       => 'active-config',
		'core:textBlob'   => encode_base64($in->{active}),
		'core:mediaType'  => 'text/plain',
		'core:context'    => 'active',
		'core:promotable' => 'false',
	};

	# Push the configuration into the repository
	push(@{$repository->{'core:config'}}, $active);

	# Print the repository
	$out->print_element('core:configRepository', $repository);

	return;
}

sub parse_static_routes{
	my ($in, $out) = @_;
	return unless ($in->{showRoute} =~ /rpc-reply/);

	my @static_routes;
	my $vrfRoutes;

	my ($xml) = $in->{showRoute} =~ /(<rpc-reply.+?<\/rpc-reply>)/s;

	my $dom = get_dom_from_xml($xml, 'show configuration routing-options');

	my $default_metric = $dom->findvalue('/rpc-reply/configuration/routing-options/static/defaults/metric/metric-value');
	my $default_preference = $dom->findvalue('/rpc-reply/configuration/routing-options/static/defaults/preference/metric-value');

	my @routes = $dom->findnodes('/rpc-reply/configuration/routing-options/static/route');
	for my $route ( @routes ) {
		my $cidr = $route->findvalue('./name');
		next unless $cidr;

		if( $cidr =~ /^$ip_regex$/ and $cidr =~ /^0\./ and $cidr !~ /^0\.0\.0\.0/ ) {
			next;
		}

		if( $cidr eq 'default' ) {
			$cidr = '0.0.0.0/0';
		}

		my $is_discard = $route->exists('./discard');
		if( $is_discard ) {
			my $static_route = _create_route($cidr);
			$static_route->{discard} = 'true';
			$static_route->{interface} = 'n/a';
			push @static_routes, $static_route;
		} else {
			# These hold the matrix of ('interface' x 'gateway addresses') the route can be applied to
			my %interfaces; # will have 1..N interfaces
			my %gateways;   # will have 1..M gateway addresses, giving N * M routes created for this route

			# If this route's definition is in a VRF table
			my $next_table = $route->findvalue('./next-table');
			my @next_hops = $route->findnodes('./next-hop');
			if( $next_table ) {

				# Get the first VRF table name in a potentially dot-delimited path
				my ($vrf_name) = $next_table =~ /^([^\s\.]+)\.?.*/;
				if( exists $in->{showNamedVrfs}->{$vrf_name} ) {

					# If this vrf table has been XML parsed, don't call again
					if( not defined $vrfRoutes->{$vrf_name} ) {
						my ($vrf_xml) = $in->{showNamedVrfs}->{$vrf_name} =~ /(<instance.+?<\/instance>)/s;
						my $vrf_dom = get_dom_from_xml($vrf_xml, "show configuration routing-instances $vrf_name");
						$vrfRoutes->{$vrf_name} = $vrf_dom;
					}

					my @vrf_routes = $vrfRoutes->{$vrf_name}->findnodes("/instance/routing-options/static/route[name='$cidr']");
					if( @vrf_routes ) {
						my @interfaces = $vrfRoutes->{$vrf_name}->findnodes('/instance/interface');
						for my $interface ( @interfaces ) {
							my $interface_name = $interface->findvalue('./name');
							$interfaces{$interface_name} = 1;
						}
						my @vrf_next_hops = $vrfRoutes->{$vrf_name}->findnodes("/instance/routing-options/static/route[name='$cidr']/next-hop");
						for my $vrf_next_hop ( @vrf_next_hops ) {
							my $gateway = $vrf_next_hop->to_literal();
							$gateways{$gateway} = 1;
						}
					} else {
						$LOGGER->debug("WARNING: Cannot find VRF entry for route $cidr in VRF '$vrf_name'");
						next;
					}
				} else {
					$LOGGER->debug("WARNING: Cannot find VRF '$vrf_name' in routing config.");
					next;
				}
			# else the gateway is given in this route
			} elsif( @next_hops ) {

				# Can be multiple next hops
				for my $next_hop ( @next_hops ) {
					my $gateway = $next_hop->to_literal();

					if( $gateway =~ /^$ip_regex$/) {
						$gateways{$gateway} = 1;

					# When next hop is an interface name, resolve gateway to connected subnets
					} elsif( $in->{IFIpForwarding}->{$gateway} ) {
						for my $interface_subnet ( @{ $in->{IFIpForwarding}->{$gateway} } ) {
							$gateway = $interface_subnet->get_original_host();
							$gateways{$gateway} = 1;
						}
					}
				}
			}

			for my $gateway ( keys %gateways ) {
				# Search interface subnets for our next hop, and add to interface list if found
				if( not %interfaces ) {
					for my $interface_name ( keys %{ $in->{IFIpForwarding} } ) {
						my $interface_subnets = $in->{IFIpForwarding}->{$interface_name};
						for my $interface_subnet ( @{ $interface_subnets } ) {
							if($interface_subnet->contains($gateway)){
								$interfaces{$interface_name} = 1;
								next;
							}
						}
					}
				}
				$interfaces{'n/a'} = 1 unless %interfaces;

				for my $interface ( keys %interfaces ) {
					my $static_route = _create_route($cidr);
					$static_route->{gatewayAddress} = $gateway;
					$static_route->{interface} = $interface;

					# Use metric from block OR preference from block OR the default metric OR the default preference
					my $this_metric = $route->findvalue('./metric/metric-value');
					my $this_preference = $route->findvalue('./preference/metric-value');
					my $metric = $this_metric || $this_preference || $default_metric || $default_preference;
					$static_route->{routeMetric} = $metric if $metric;

					push @static_routes, $static_route;
				}
			}
		}
	}

	if( @static_routes ) {
		$out->print_element('staticRoutes', { staticRoute => \@static_routes } );
	}

	return;
}


sub parse_dynamic_routes{
	my ($in, $out) = @_;
	my @dynamic_routes;

	# Parse OSPF from one place
	if( $in->{showOspfRoute} and $in->{showOspfRoute} =~ /rpc-reply/ ) {
		my ($xml) = $in->{showOspfRoute} =~ /(<rpc-reply.+?<\/rpc-reply>)/s;
		my ($routing_namespace) = $xml =~ /ospf-route-information\s+xmlns="([^"]+)"/;

		# The response can be essentially empty if ospf is not turned on
		if( $routing_namespace ) {

			my $dom = get_dom_from_xml($xml, 'show ospf route');

			my $xpc = XML::LibXML::XPathContext->new($dom);
			$xpc->registerNs('rt',  $routing_namespace);

			my @routes = $xpc->findnodes('/rpc-reply/rt:ospf-route-information/rt:ospf-topology-route-table/rt:ospf-route');
			for my $route ( @routes ) {

				my @route_entries = $xpc->findnodes('./rt:ospf-route-entry', $route);
				for my $route_entry ( @route_entries ) {
					my $dynamic_route = _parse_ospf_route_entry($xpc, $route_entry);
					if( $dynamic_route ) {
						push @dynamic_routes, $dynamic_route;
					}
				}
			}
		}
	}

	# Parse BGP from somewhere completely different
	if( $in->{showRouteProtocolBgp} and $in->{showRouteProtocolBgp} =~ /rpc-reply/ ) {
		my ($xml) = $in->{showRouteProtocolBgp} =~ /(<rpc-reply.+?<\/rpc-reply>)/s;
		my ($routing_namespace) = $xml =~ /route-information\s+xmlns="([^"]+)"/;

		if( $routing_namespace ) {
			my $dom = get_dom_from_xml($xml, 'show route protocol bgp');

			my $xpc = XML::LibXML::XPathContext->new($dom);
			$xpc->registerNs('rt',  $routing_namespace);

			my @route_tables = $xpc->findnodes('/rpc-reply/rt:route-information/rt:route-table');
			for my $route_table ( @route_tables ) {

				# Only take from the default IPv4 table
				my $is_default_table = ( $xpc->findvalue('./rt:table-name', $route_table) eq 'inet.0' );
				next unless $is_default_table;

				my @routes = $xpc->findnodes('./rt:rt', $route_table);
				for my $route ( @routes ) {

					my @route_entries = $xpc->findnodes('./rt:rt-entry', $route);
					for my $route_entry ( @route_entries ) {

						my @next_hops = $xpc->findnodes('./rt:nh', $route_entry);
						for my $next_hop ( @next_hops ) {
							my $dynamic_route = _parse_bgp_route($xpc, $next_hop, $route_entry);
							if( $dynamic_route ) {
								push @dynamic_routes, $dynamic_route;
							}
						}
					}
				}
			}
		}
	}

	if( @dynamic_routes ) {	
		$out->print_element('dynamicRoutes', { dynamicRoute => \@dynamic_routes });
	}

	return;
}

sub _create_route{
	my ($cidr) = @_;

	my ($addr, $mask) = split(/\//, $cidr);

	my $static_route = {
		destinationAddress => $addr,
		destinationMask    => $mask,
		defaultGateway     => $cidr eq '0.0.0.0/0' ? 'true' : 'false',
	};

	return $static_route;
}

sub _get_applies_to {
	my ($in) = @_;
	my $applies_to;

	# Get the interfaces that have lists of filters defined
	my ($xml) = $in->{showConfigInterfaces} =~ /(<interfaces.+?<\/interfaces>)/sm;

	my $dom = get_dom_from_xml($xml, 'show configuration interfaces');

	my @interfaces = $dom->findnodes('/interfaces/interface');
	for my $interface ( @interfaces ) {
		my $interface_name = $interface->findvalue('./name');

		my @units = $interface->findnodes('./unit');
		for my $unit ( @units ) {
			my $unit_name = $unit->findvalue('./name');
			my $name = "$interface_name.$unit_name";

			my @output_lists = $unit->findnodes('./family/inet/filter/output-list');
			for my $list ( @output_lists ) {
				my $list_name = $list->to_literal();
				push @{ $applies_to->{$list_name}->{out} }, $name;
			}

			my @input_lists = $unit->findnodes('./family/inet/filter/input-list');
			for my $list ( @input_lists ) {
				my $list_name = $list->to_literal();
				push @{ $applies_to->{$list_name}->{in} }, $name;
			}
		}
	}

	return $applies_to;
}

# Uses partial loading of the firewall XML output to process all filters
sub parse_filters{
	my ($in, $out) = @_;

	$out->open_element('filterLists');

	my ($firewallSection) = $in->{showFirewall} =~ /(<firewall.+?<\/firewall>)/s;

	# If the inet6 block is ignore then lets strip it out so that it isn't parsed
	if($in->{showFirewallInet6} && $in->{showFirewallInet6} =~ /Warning: configuration block ignored/){
		$firewallSection =~ s/<inet6>.+?<\/inet6>//s;
	}

	if($in->{showFirewallPlain} && $in->{showFirewallPlain} =~ /Warning: configuration block ignored/m){
		$firewallSection =~ s/<firewall>.+?<\/firewall>//s;
	}

	if($firewallSection){
		my $filterListHash = {};
		my $twig = XML::Twig->new(
			twig_roots => {
				'filter' => _process_filter($in, $out, _get_applies_to($in), $filterListHash)
			}
		);
		$twig->parse($firewallSection);
		undef $firewallSection;
	}

	if(defined($in->{showPolicy}) and $in->{showPolicy} =~ /configuration/){
		_parse_srx_filters($in, $out);
	}

	$out->close_element('filterLists');

	return;
}

sub _parse_zone_interfaces{
	my ($in, $out) = @_;
	return unless ($in->{showZones} =~ /rpc-reply/);
	my $result;

	my ($xml) = $in->{showZones} =~ /(<rpc-reply.+?<\/rpc-reply>)/s;

	my $dom = get_dom_from_xml($xml, 'show configuration security zones');
	
	my @zones = $dom->findnodes('/rpc-reply/configuration/security/zones/security-zone');
	for my $zone ( @zones ) {
		my $name = $zone->findvalue('./name');
		my @interfaces = $zone->findnodes('./interfaces');
		for my $interface ( @interfaces ) {
			push @{ $result->{$name}->{interfaces} }, $interface->findvalue('./name');
		}
	}

	$in->{parsed}->{zones} = $result;

	return $result;
}

# SRX IPS configuration...
sub _parse_srx_filters{
	my ($in, $out) = @_;
	return unless ($in->{showPolicy} =~ /rpc-reply/);

	my ($xml) = $in->{showPolicy} =~ /(<rpc-reply.+?<\/rpc-reply>)/s;

	my $dom = get_dom_from_xml($xml, 'show configuration security');

	#  Return if SRX policy rules aren't applicable.
	return unless $dom->exists('/rpc-reply/configuration/security/policies');

	my $zones_interfaces = _parse_zone_interfaces($in, $out);
	my $applications = _parse_applications($in, $out);

	my $is_permit_all = $dom->exists('/rpc-reply/configuration/security/policies/default-policy/permit-all');
	my $default_policy = $is_permit_all ? 'permit' : 'deny';

	my @policy_sets = $dom->findnodes('/rpc-reply/configuration/security/policies/policy');
	if( @policy_sets ) {
		for my $policy_set ( @policy_sets ) {
			$out->open_element('filterList');

			# Only set the applies to if there are interfaces assigned to both the from-zone
			# and the to-zone.  In SRX the appliesTo must be complete with source and destination
			# interfaces to be valid
			my $policy_from_zone = $policy_set->findvalue('./from-zone-name');
			my $policy_to_zone = $policy_set->findvalue('./to-zone-name');
			my $from_zone_interfaces = $zones_interfaces->{$policy_from_zone};
			my $to_zone_interfaces = $zones_interfaces->{$policy_to_zone};
			if( $from_zone_interfaces and $from_zone_interfaces->{interfaces}
			and $to_zone_interfaces and $to_zone_interfaces->{interfaces} ) {
				$out->open_element('appliesTo');
				$out->print_element('sourceGroup', $policy_from_zone);
				$out->print_element('destinationGroup', $policy_to_zone);
				$out->close_element('appliesTo');
			}

			my $currentProcessOrder = 0;
			my @policies = $policy_set->findnodes('./policy');

			POLICY_LOOP:
			for my $policy ( @policies ) {
				my $name  = $policy->findvalue('./name');
				my ($match) = $policy->findnodes('./match');
				my ($then)  = $policy->findnodes('./then');

				my $action = 'permit';
				if( $then->exists('./discard') ) {
					$action = 'drop';
				} elsif( $then->exists('./reject') || $then->exists('./deny') ) {
					$action = 'deny';
				}

				my $log = ($then->exists('./log') || $then->exists('./syslog') ) ? 'true' : 'false';

				my $fe = {
					log           => $log,
					primaryAction => $action,
					name          => $name
				};

				for my $direction ('source', 'destination') {
					my @addresses = $match->findnodes("./$direction-address");
					for my $address( @addresses ) {
						$address = $address->to_literal();
						next if $address eq 'any';

						if( $address =~ /$ipv6_regex/ ) {
							$LOGGER->warning($RULE_DROPPED, "IPv6 address '$address' not supported for policy '$name'");
							next POLICY_LOOP;
						}

						my $zone_name = $direction eq 'source' ? $policy_from_zone : $policy_to_zone;
						my $ref = '_ENTRY_' . ncname_from_string($address);
						my $zone_ref = $zone_name . $ref;
						my $global_ref = 'global' . $ref;

						if( $in->{parsed}->{object_groups}->{$zone_ref} ) {
							push @{ $fe->{$direction . 'IpAddr'} }, { objectGroupReference => $zone_ref };
						} elsif( $in->{parsed}->{object_groups}->{$global_ref} ) {
							push @{ $fe->{$direction . 'IpAddr'} }, { objectGroupReference => $global_ref };
						} else {
							$LOGGER->warning($RULE_DROPPED, "Object group '$zone_ref' or '$global_ref' not recognised in policy '$name'");
							next POLICY_LOOP;
						}
					}
				}

				my @destination_services;
				my @applications = $match->findnodes('./application');
				for my $app ( @applications ) {
					$app = $app->to_literal();
					next if $app eq 'any';

					if( $app =~ /^junos-(.+)$/ ) {
						my $port_expressions = get_port_expression($1);
						if( $port_expressions ) {
							push @destination_services, @{ $port_expressions };
						} else {
							$LOGGER->debug("WARNING: Dropping rule, destination service '$app' not recognised in policy '$name'");
							next POLICY_LOOP;
						}
					} elsif( $applications->{$app} ) {
						push @destination_services, @{ $applications->{$app} };
					} else {
						$LOGGER->debug("WARNING: Dropping rule, destination service '$app' not recognised in policy '$name'");
						next POLICY_LOOP;
					}
				}
				if( @destination_services ) {
					# TODO dedup cause there can be duplicated services
					$fe->{destinationService} = \@destination_services;
				}

				$fe->{processOrder} = ++$currentProcessOrder;
				$out->print_element('filterEntry', $fe);
			}

			$out->print_element('filterEntry', {
					log           => 'false',
					primaryAction => $default_policy,
					processOrder  => ++$currentProcessOrder,
					name          => 'default-policy',
				}
			);

			$out->print_element('mode', 'stateless');
			$out->print_element('name', "fromZone_$policy_from_zone-toZone_$policy_to_zone");
			$out->close_element('filterList');
		}
	}

	return;
}

sub _parse_applications {
	my ($in, $out) = @_;

	if( $in->{showApplications} ) {
		my ($xml) = $in->{showApplications} =~ /(<rpc-reply.+?<\/rpc-reply>)/s;

		if( $xml ) {
			my $dom = get_dom_from_xml($xml, 'show configuration applications');

			my @applications = $dom->findnodes('/rpc-reply/configuration/applications/application');
			for my $application ( @applications ) {
				my $name = $application->findvalue('./name');
				$securityApps->{$name} = _parse_application($name, $application);
			}

			my @application_sets = $dom->findnodes('/rpc-reply/configuration/applications/application-set');
			for my $application_set ( @application_sets ) {
				my $name = $application_set->findvalue('./name');

				my @services;
				my @members = $application_set->findnodes('./application');
				for my $member ( @members ) {
					my $member_name = $member->findvalue('./name');

					if( $member_name =~ /^junos-(.+)$/ ) {
						my $port_expressions = get_port_expression($1);
						if( $port_expressions ) {
							push @services, @{ $port_expressions };
						} else {
							$LOGGER->debug("WARNING: Dropping application-set memeber, service '$member_name' not recognised in application-set '$name'");
						}
					} else {
						for my $lookup ( @{ $securityApps->{$member_name} } ) {
							push @services, $lookup;
						}
					}
				}
				$securityApps->{$name} = \@services;
			}
		}
	}
	return $securityApps;
}

sub _parse_application {
	my ($name, $application) = @_;
	my @services;

	my $port = $application->findvalue('./destination-port');
	my $protocol = $application->findvalue('./protocol');

	if( defined $protocol and $protocol =~ /^\d+$/) {
		if( $protocol == 41 ) { #ipv6 encapsulation
			$protocol = 'ipv6'
		}
	}

	if( $port and $port =~ /(\d+)\-(\d+)/ ) {
		push @services,	{
			portRange => {
				portStart => $1,
				portEnd   => $2,
				protocol  => $protocol,
			},
		};
	} elsif( $port and $port =~ /,/ ) {
		my @ports = split /,/, $port;
		for my $portNo ( @ports ) {
			push @services,	{
				portExpression => {
					operator => 'eq',
					port     => get_port_number($portNo),
					protocol => $protocol,
				},
			};
		}
	} elsif( defined $port ) {
		my $lookup = get_port_number($port);
		if( $lookup =~ /^\d+/ ) {
			push @services,	{
				portExpression => {
					operator => 'eq',
					port     => get_port_number($port),
					protocol => $protocol,
				},
			};
		}
	} elsif( $protocol ) {

		# Service group cannot be created without a port number so this is a failure of the ZED
		# We'll 'fake' it with the full port range, clients will have to be aware.
		push @services,	{
			portExpression => {
				operator => 'gt',
				port     => 0,
				protocol => $protocol,
			},
		};
	}
	return \@services;
}

sub parse_object_groups{
	my ($in, $out) = @_;
	my %network_groups = ();

	# Parse each zones' address book
	if( $in->{showZones} and $in->{showZones} =~ /rpc-reply/ ) {
		my ($xml) = $in->{showZones} =~ /(<rpc-reply.+?<\/rpc-reply>)/s;

		my $dom = get_dom_from_xml($xml, 'show configuration security zones');

		my @zone_nodes = $dom->findnodes('/rpc-reply/configuration/security/zones/security-zone');

		for my $zone_node ( @zone_nodes ) {
			my $name = $zone_node->findvalue('./name');
			my ($address_book) = $zone_node->findnodes('./address-book');
			if( $address_book ) {
				my $zone_groups = _parse_address_book($name, $address_book);
				if( $zone_groups ) {
					%network_groups = ( %network_groups, %{ $zone_groups } );
				}
			}
		}
	}

	# Parse the global address book
	if( $in->{showPolicy} and $in->{showPolicy} =~ /rpc-reply/ ) {
		my ($xml) = $in->{showPolicy} =~ /(<rpc-reply.+?<\/rpc-reply>)/s;

		my $dom = get_dom_from_xml($xml, 'show configuration security');

		my ($global_address_book) = $dom->findnodes('/rpc-reply/configuration/security/address-book');
		if( $global_address_book ) {
			my $global_groups = _parse_address_book('global', $global_address_book);
			if( $global_groups ) {
				%network_groups = ( %network_groups, %{ $global_groups } );
			}
		}
	}

	# Parse the global prefix lists
	if($in->{active} && $in->{active} =~ /^policy-options/m){
		my $prefix_groups = _parse_prefix_lists( $in->{active} );
		if( $prefix_groups ) {
			%network_groups = ( %network_groups, %{ $prefix_groups } );
		}
	}

	if( %network_groups ) {
		$out->open_element('objectGroups');
		for my $id ( keys %network_groups ) {
			_print_network_group($out, $network_groups{$id});
		}
		$out->close_element('objectGroups');
	}

	$in->{parsed}->{object_groups} = \%network_groups;

	return;
}

sub _parse_address_book {
	my ($name, $address_book_node) = @_;
	my $network_groups;

	# Parse single addresses that have been given a name
	my @address_nodes = $address_book_node->findnodes('./address');
	for my $address_node ( @address_nodes ) {
		my $address_name = $address_node->findvalue('./name');
		my $id = $name . '_ENTRY_' . ncname_from_string($address_name);

		my $address = $address_node->findvalue('./ip-prefix');
		if( $address ) {
			if( $address !~ /$ipv6_regex/ ) {
				my $net = parseCIDR( $address );
				my $host = $net->{host};
				my $network = mask_to_bits( $net->{network} );

				if( $host ){
					my $network_group = { id => $id };

					if( $network eq 32 ){
						$network_group->{host} = $host;
					}else{
						$network_group->{subnet} = {address => $host, mask => $network};
					}

					$network_groups->{$id} = $network_group;
				} else {
					$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Host IP address not found for address '$id'");
				}
			} else {
				$LOGGER->warning($NETWORK_OBJECT_DROPPED, "IPv6 address not supported: '$address'");
			}
		}
	}

	my $count = scalar keys %{ $network_groups };
	$LOGGER->debug("Parsed $count addresses to network groups in address book '$name'");

	# Parse address sets
	my %unresolved;
	my @address_set_nodes = $address_book_node->findnodes('./address-set');
	for my $address_set_node ( @address_set_nodes ) {
		my $set_name = $address_set_node->findvalue('./name');
		my $id = $name . '_ENTRY_' . ncname_from_string($set_name);
		my $network_group = { id => $id };

		my @member_nodes = $address_set_node->findnodes('./address');

		for my $member_node ( @member_nodes ) {
			my $member_name = $member_node->findvalue('./name');
			my $member_id = $name . '_ENTRY_' . ncname_from_string($member_name);

			if( $network_groups->{$member_id} ) {
					push @{ $network_group->{objectGroupReference} }, $member_id;;
			} else {
				push @{ $unresolved{$member_id} }, $id;
			}
		}

		$network_groups->{$id} = $network_group;
	}

	# Any address references dropped eariler because they had not yet been parsed are retrospectively added
	for my $id ( keys %unresolved ) {
		if( $network_groups->{$id} ) {
			for my $set_id ( @{ $unresolved{ $id } } ) {
				$LOGGER->verbose("Retrospectively adding reference '$id' to address set '$set_id'");
				push @{ $network_groups->{$set_id}->{objectGroupReference} }, $id;
			}
		} else {
			my $unresolved_sets = join ', ', @{ $unresolved{ $id } };
			$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Unrecognised address object '$id' in address sets: $unresolved_sets");
		}
	}

	$count -= scalar keys %{ $network_groups }; $count *= -1;
	$LOGGER->debug("Parsed $count address sets to network objects in address book '$name'");

	return $network_groups;
}

# Parses prefix lists defined in the running config
sub _parse_prefix_lists {
	my ($config) = @_;
	my $network_groups;

	my ($policy_options) = $config =~ /(^policy-options\s+.+?^})/sm;
	my @prefix_lists = $policy_options =~ /(^\s+prefix-list\s+\S+\s+{.+?;\s*^\s+})/smg;
	for my $prefix_list ( @prefix_lists ) {
		if( $prefix_list =~ /prefix-list\s+(\S+)/ ) {
			my $id = $1;
			my $network_group = { id => $id };
			while( $prefix_list =~ /^\s+($ip_regex)\/(\d+);\s*$/gm ) {
				my ($address, $mask) = ($1, $2);
				if($mask eq 32){
					push @{ $network_group->{host} }, $address;
				}else{
					push @{ $network_group->{subnet} }, {address => $address, mask => $mask};
				}
			}
			$network_groups->{$id} = $network_group;
		}
	}

	my $count = scalar keys %{ $network_groups };
        $LOGGER->debug("Parsed $count prefix-lists to network objects");

	return $network_groups;
}

sub _print_network_group{
	my ($out, $network_group) = @_;

	$out->open_element('networkGroup');
	$out->print_element('id', $network_group->{id});

	for my $element ( 'host', 'objectGroupReference', 'subnet' ) {
		if(defined $network_group->{$element}){
			$out->print_element($element, $network_group->{$element});
		}
	}

	$out->close_element('networkGroup');

	return;
}

sub _parse_module{
	my ($raw_module) = @_;
	my $module;

	$module->{'core:asset'}->{'core:factoryinfo'}->{'core:make'} = 'Juniper';
	$module->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = 'Unknown';

	if(defined($raw_module->first_child('version'))){
		$module->{'core:asset'}->{'core:factoryinfo'}->{'core:hardwareVersion'} = $raw_module->first_child('version')->text();
	}

	if(defined($raw_module->first_child('part-number'))){
		$module->{'core:asset'}->{'core:factoryinfo'}->{'core:partNumber'} = $raw_module->first_child('part-number')->text();
	}

	if(defined($raw_module->first_child('serial-number'))){
		$module->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = $raw_module->first_child('serial-number')->text();
	}

	if(defined($raw_module->first_child('description'))){
		$module->{'core:description'} = $raw_module->first_child('description')->text();
	}

	return($module);
}

sub _process_filter{
	my ($in, $out, $appliesTo, $filterListHash) = @_;
	return sub {

		# Processes a single firewall filter as an XML::Twig
		my ($twig, $element) = @_;
		my $filterName = _get_element_text($element, 'name');
		if($filterName){
			$out->open_element('filterList');

			my $lists = $appliesTo->{$filterName};
			if($lists && $lists->{in}){
				my $list = $lists->{in};
				foreach my $int (@$list){
					$out->open_element('appliesTo');
					$out->print_element('source', $int);
					$out->close_element('appliesTo');
				}
			}
			if($lists && $lists->{out}){
				my $list = $lists->{out};
				foreach my $int (@$list){
					$out->open_element('appliesTo');
					$out->print_element('destination', $int);
					$out->close_element('appliesTo');
				}
			}
		}

		my $currentProcessOrder = 0;
		foreach my $filterEntryElement ($element->descendants('term')){
			my $filterEntry = {
				name          => _get_element_text($filterEntryElement, 'name'),
				log           => 'false',
				primaryAction => 'permit',
			};

			# Process filter references
			my $filterRef = _get_element_text($filterEntryElement, 'filter');
			if($filterRef){
				my $inlineEntries = $filterListHash->{$filterRef};
				foreach my $inlineEntry (@$inlineEntries){
					$out->print_element('filterEntry', $inlineEntry);
				}
				next;
			}

			# Process the address information
			my $from = $filterEntryElement->first_child('from');
			if($from){
				my $address = $from->first_child('address');
				if($address){
					my $parsed_address = _process_address($address);
					if( $parsed_address ) {
						push @{$filterEntry->{sourceIpAddr}}, $parsed_address;
					}
				}

				for my $direction ( 'destination', 'source' ) {
					my $element = $direction . 'IpAddr';

					my @addresses = $from->children("$direction-address");
					if(@addresses){
						my $parsed_addresses = _process_addresses(@addresses);
						if( $parsed_addresses ) {
							push @{$filterEntry->{$element}}, $parsed_addresses;
						}
					}

					my @ranges = $from->children("$direction-address-range");
					if(@ranges){
						my $parsed_ranges = _process_address_ranges(@ranges);
						if( $parsed_ranges ) {
							push @{$filterEntry->{$element}}, $parsed_ranges;
						}
					}

					my @lists = $from->children("$direction-prefix-list");
					if( $direction eq 'source' and not @lists) {
						@lists = $from->children("prefix-list");
					}
					if(@lists){
						my $parsed_lists = _process_address_lists($in, @lists);
						if( $parsed_lists ) {
							push @{$filterEntry->{$element}}, $parsed_lists;
						}
					}
				}

				# Process the port information
				my @ports = $from->children('port');
				if(@ports){
					$filterEntry->{sourceService} = _process_port(@ports);
				}

				my @dstPorts = $from->children('destination-port');
				if(@dstPorts){
					$filterEntry->{destinationService} = _process_port(@dstPorts);
				}

				my @srcPorts = $from->children('source-port');
				if(@srcPorts){
					$filterEntry->{sourceService} = _process_port(@srcPorts)
				}

				my @exceptSrcPorts = $from->children('source-port-except');
				if(@exceptSrcPorts){
					$filterEntry->{sourceService} = _process_port_negate(@exceptSrcPorts);
				}

				my @exceptDestPorts = $from->children('destination-port-except');
				if(@exceptDestPorts){
					$filterEntry->{destinationService} = _process_port_negate(@exceptDestPorts);
				}

				foreach my $prot ($from->children('protocol')){
					my $protocol = $prot->text();

					if($protocol =~ /^\d+$/){
						$protocol = get_protocol_name($protocol);

						if(!$protocol){
							$LOGGER->debug("[WARNING] Unable to resolve protocol number '" . $prot->text() . "'");
							next;
						}
					}

					push(@{$filterEntry->{protocol}}, $protocol);
				}
			}

			my $then = $filterEntryElement->first_child('then');
			if($then){
				if($then->first_child('accept')){
					$filterEntry->{primaryAction} = 'permit';
				}elsif($then->first_child('reject')){
					$filterEntry->{primaryAction} = 'deny';
				}elsif($then->first_child('discard')){
					$filterEntry->{primaryAction} = 'drop';
				}elsif($then->first_child('next')){
					$filterEntry->{primaryAction} = 'next';
				}

				if($then->first_child('log') || $then->first_child('syslog')){
					$filterEntry->{log} = 'true';
				}
			}

			$filterEntry->{processOrder} = ++$currentProcessOrder;

			if($filterName){
				push(@{$filterListHash->{$filterName}}, $filterEntry);
			}
			$out->print_element('filterEntry', $filterEntry);
		}

		if($filterName){
			$out->print_element('mode', 'stateless');
			$out->print_element('name', $filterName);
			$out->close_element('filterList');
		}
	}
}

# If the element is defined this returns the text value of the element.
# If the element is not defined this returns undef;
sub _get_element_text{
	my ($twig, $elementName) = @_;
	my $child = $twig->first_child($elementName);
	if($child){
		return $child->text();
	}else{
		return undef;
	}
}

sub _process_addresses{
	my @twigs = @_;

	my @addresses;
	foreach my $twig (@twigs){
		my $parsed_address = _process_address($twig);
		if( $parsed_address ) {
			push @addresses, $parsed_address;
		}
	}
	return \@addresses;
}

sub _process_address{
	my $twig = shift;

	my $addresses;
	foreach my $child ($twig->children('name')){
		my ($addr, $mask) = split(/\//, $child->text());

		if($addr !~ /$ipv6_regex/){
			if($mask eq 32){
				push(@{$addresses}, {host => $addr});
			}else{
				push(@{$addresses}, {network => {address => $addr, mask => $mask,}});
			}
		} else {
			$LOGGER->warning($RULE_DROPPED, "IPv6 addresses are not supported: $addr");
		}
	}
	return $addresses;
}

sub _process_address_ranges{
	my @twigs = @_;

	my @ranges;
	foreach my $twig (@twigs){
		my ($low) = $twig->children('low');
		$low = $low->text();
		my ($high) = $twig->children('high');
		$high = $high->text();

		if( $low !~ /$ipv6_regex/ and $high !~ /$ipv6_regex/ ) { 
			push @ranges, {range => {startAddress => $low, endAddress => $high}};
		} else {
			$LOGGER->warning($RULE_DROPPED, "IPv6 ranges are not supported: $low - $high");
		}
	}

	return \@ranges;
}

sub _process_address_lists{
	my ($in, @twigs) = @_;

	my @lists;
	for my $twig (@twigs){
		for my $twig_ref ( $twig->children('name') ) {
			my $ref = $twig_ref->text();
			if( $in->{parsed}->{object_groups}->{$ref} ) {
				push @lists, {objectGroupReference => $ref};
			} else {
				$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Reference '$ref' is not known, dropping from rule");
			}
		}
	}

	return \@lists;
}

sub _process_port_negate{
	my $portArray = _process_port(@_);

	foreach my $port (@$portArray){
		$port->{negate} = 'true';
	}

	return $portArray;
}

sub _process_port{
	my $portArray;
	foreach my $port (@_){
		my $p = $port->text();
		if($p =~ /^(\d+)\-(\d+)$/){
			push(@{$portArray}, {portRange => {portStart => $1, portEnd => $2,}});
		}elsif($p =~ /^\d+$/){
			push(@{$portArray}, {portExpression => {operator => 'eq', port => $p,}});
		}elsif($p =~ /\S+/){
			push(@{$portArray}, {portExpression => {operator => 'eq', port => get_port_number($p),}});
		}
	}
	return $portArray;
}

sub _get_patterns{
	return({
#		Router => '^(acx|ctp|e|jcs|ln|m|ptx|srx|t)',
		Router => '^(mx)',
		Switch => '^(ex|qfx|wl)',
		IDP => '^(idp)',
	});
}

# Given the name of a juniper configuration section, returns an array of each section
sub _get_section{
	my ($section, $input) = @_;
	my @results;

	while ($input =~ /^((\s+)$section[^\n]*?{(.+?)^\2})/msig){
		push(@results, $1);
	}

	return(\@results);
}

sub _parse_ospf_route_entry {
	my ($xpc, $route_entry) = @_;

	my $route_type = $xpc->findvalue('./rt:next-hop-type', $route_entry);
	if( $route_type ne 'IP' ) { 
		return;  # only parse IP routes
	}

	my $dynamic_route = { advertisingRoutingProcess => 'ospf' };

	my $destination = $xpc->findvalue('./rt:address-prefix', $route_entry);
	my $mask;
	if( $destination =~ /^($ip_regex)\/(\d+)$/ ) {
		$destination = $1;
		$mask = $2;
	}

	$dynamic_route->{routeNetwork} = {
		address => $destination,
		mask => $mask // 32
	};

	my @next_hops = $xpc->findnodes('./rt:ospf-next-hop', $route_entry);
	for my $next_hop ( @next_hops ) {
		my $routeNextHop = _parse_ospf_next_hop($xpc, $next_hop, $route_entry);
		push @{ $dynamic_route->{routeNextHop} }, $routeNextHop;
	}
    
	return $dynamic_route;
}

sub _parse_ospf_next_hop {
	my ($xpc, $next_hop, $route_entry) = @_;

	my ($interface_node) = $xpc->findnodes('./rt:next-hop-name', $next_hop);
	my $interface = $xpc->findvalue('./rt:interface-name', $interface_node);
	my ($next_hop_address) = $xpc->findnodes('./rt:next-hop-address', $next_hop);
	my $address = $xpc->findvalue('./rt:interface-address', $next_hop_address);
	my $cost = $xpc->findvalue('./rt:interface-cost', $route_entry);
	
	my $model = {
		adjacencyInterface => $interface,
		adjacencyMetric => $cost
	};
	$model->{adjacencyNode} = $address if $address;

	return $model;
}

sub _parse_bgp_route {
	my ($xpc, $next_hop, $route_entry) = @_;
	my $dynamic_route;

	my $is_selected_hop = $xpc->exists('./rt:selected-next-hop', $next_hop);
	if( $is_selected_hop ) {

		my $to = $xpc->findvalue('./rt:to', $next_hop);
		my $via = $xpc->findvalue('./rt:via', $next_hop);
		my $metric = $xpc->findvalue('./rt:local-preference', $route_entry);

		$dynamic_route = {
			routeNetwork => {
				address => '0.0.0.0',
				mask    => '0'
			},
			routeNextHop => {
				adjacencyNode      => $to,
				adjacencyInterface => $via,
				adjacencyMetric    => $metric
			},
			advertisingRoutingProcess => 'bgp'
		};
	}

	return $dynamic_route;
}

sub get_dom_from_xml {
	my ($xml, $response_name) = @_;
	my $dom;

	eval { $dom = XML::LibXML->load_xml(string => $xml); };
	if( $@ ) {
		$LOGGER->fatal("Error parsing '$response_name' xml:\n$@");
	}

	return $dom;
}

1;
