package ZipTie::Adapters::Juniper::JUNOS::Telemetry;

use strict;
use warnings;
use 5.010_000;

use ZipTie::Adapters::Juniper::JUNOS::AutoLogin;
use ZipTie::Adapters::Juniper::JUNOS::Disconnect qw(disconnect);
use ZipTie::Adapters::Utils qw(choose_admin_ip get_model_filehandle close_model_filehandle open_telemetry_neighbors close_telemetry_neighbors get_crep strip_mac get_interface_type);
use ZipTie::CLIProtocolFactory;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

use Net::IP;
use XML::Twig;

use Exporter qw(import);
our @EXPORT_OK = qw(parse_arp parse_routing_neighbors parse_telemetry_interfaces);

my $LOGGER = ZipTie::Logger::get_logger();
my $ip_regex = get_crep('cipm');
my $ipv6_regex = get_crep('ipv6');

sub invoke{
	my ($pkg, $connectionPathDoc) = @_;
	
	# Setup work
	my ($connectionPath, $discoveryParams) = ZipTie::Typer::translate_document($connectionPathDoc, 'connectionPath');
	my $filehandle = get_model_filehandle('JUNOS', $connectionPath->get_ip_address());
	my $cliProtocol = ZipTie::CLIProtocolFactory::create($connectionPath);
	my $responses = {};
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'telemetry');

	# Make a connection to device and get the data necessary to fill out the discovery event document 
	my $promptRegex = ZipTie::Adapters::Juniper::JUNOS::AutoLogin::execute( $cliProtocol, $connectionPath );
	$cliProtocol->send_and_wait_for('set cli screen-length 0', $promptRegex);
	$responses->{arp}        = $cliProtocol->send_and_wait_for('show arp no-resolve',           $promptRegex);
	$responses->{ospf}       = $cliProtocol->send_and_wait_for('show ospf neighbor',            $promptRegex);
	$responses->{rip}        = $cliProtocol->send_and_wait_for('show rip neighbor',             $promptRegex);
	$responses->{bgp}        = $cliProtocol->send_and_wait_for('show bgp neighbor',             $promptRegex);
	$responses->{interfaces} = $cliProtocol->send_and_wait_for('show interfaces | display xml', $promptRegex);
	
	if($responses->{interfaces} !~ /rpc-reply/){
		die('Access to the command \'show interfaces | display xml\' is compulsory for neigbour data to succeed');
	}

	$printer->open_discovery_event();

	open_telemetry_neighbors(1, $printer);
	parse_arp($responses, $printer);
	parse_routing_neighbors($responses, $printer);
	close_telemetry_neighbors(1, $printer);

	# Process the interfaces and their status
	my $interfacesHash = parse_telemetry_interfaces($responses, $printer);
	my $adminIp = $connectionPath->get_ip_address();
	if ($discoveryParams->{calculateAdminIp} eq 'true'){
		$adminIp = choose_admin_ip($connectionPath->get_ip_address(), $interfacesHash);
	}
	$printer->print_element('adminIp', $adminIp);
	
	# Tear down work
	$printer->close_element('DiscoveryEvent');
	close_model_filehandle($filehandle);
	disconnect($cliProtocol);
}

sub parse_arp {
	my ($in, $out) = @_;
	my $opened = 0;

	if($in->{arp}){
		while ($in->{arp} =~ /^([\da-f:]+)\s+([\da-f.:]+)\s+(\S+)/mgi){
			my ($mac, $ip, $interface) = ($1, $2, $3);

			if( $ip =~ /$ip_regex/ ) {
				if( !$opened ) {
					$out->open_element('arpTable');
					$opened = 1;
				}
				my $arp = {
					ipAddress  => $ip,
					macAddress => strip_mac($mac),
					interface  => $interface
				};
				$out->print_element('arpEntry', $arp);

			# TODO add IPv6 support
			} elsif( $ip =~ /$ipv6_regex/i ) {
				$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping ARP entry: $ip : $mac on $interface");
			}
		}
		if( $opened ) {
			$out->close_element('arpTable');
		}
	}
}

sub parse_routing_neighbors{
	my ($in, $out) = @_;

	my %protocols = (
		ospf => {
			regex => q{^([\da-f.:]+)\s+(\S+)\s+\S+\s+([\da-f.:]+)\s+\d+\s+\d+\s*$},
			a => 'ipAddress',
			b => 'interface',
			c => 'routerId'
		},
		bgp => {
			regex => q{^\s*Peer\s+ID:\s+([\da-f.:]+)\s+Local\s+ID:\s+([\da-f.:]+)},
			a => 'ipAddress',
			b => 'routerId'
		},
		rip => {
			regex => q{^\s*(\S+)\s+\S+\s+([\da-f.:]+)\s+([\da-f.:]+)},
			a => 'interface',
			b => 'routerId',
			c => 'ipAddress'
		}
	);

	my $neighbors;
	foreach my $protocol ( keys( %protocols ) ){
		if($in->{$protocol}){

			my $template = $protocols{$protocol};
			NEIGHBOR: while ($in->{$protocol} =~ /$template->{regex}/gmi){
				my $neighbor = {
					protocol => uc($protocol), 
					$template->{a} => $1,
					$template->{b} => $2
				};

				if($template->{c}){
					$neighbor->{$template->{c}} = $3;
				}

				foreach my $possible_ipv6 ( 'ipAddress', 'routerId' ) {
					# TODO add IPv6 support
					if($neighbor->{$possible_ipv6} && $neighbor->{$possible_ipv6} =~ /$ipv6_regex/i ){
						$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping $neighbor->{protocol} route: $possible_ipv6 = $neighbor->{$possible_ipv6}");
						next(NEIGHBOR);
					}
				}

				push(@{$neighbors->{routingNeighbor}}, $neighbor);
			}
		}
	}

	if(exists $neighbors->{routingNeighbor}){
		$out->print_element('routingNeighbors', $neighbors);
	}
}

# Creates the ZipTie interfaces model by reading the Juniper interface-information
# XML with XML::Twig and then traversing the twigs.  This could be streamlined
# by reading only physical-interface blocks at a time and using a twig callback
# to a method that parses that single interface
sub parse_telemetry_interfaces{
	my ($in, $out) = @_;
	my $interfaces;

	my ($intSection) = $in->{interfaces} =~ /(<interface-information.+?<\/interface-information>)/s;
	my $interfacesTwig = XML::Twig->new();
	$interfacesTwig->parse($intSection);

	foreach my $physicalInt ($interfacesTwig->descendants('physical-interface')){
		my $name       = $physicalInt->first_child('name')->text();
		my $operStatus = lc($physicalInt->first_child('oper-status')->text());

		my $type       = $name;
		if($physicalInt->first_child('if-type')){
			$type = $physicalInt->first_child('if-type')->text();
		}
		$type = get_interface_type($type);

		my $trafficStatsElement = $physicalInt->first_child('traffic-statistics');
		my $inputBytes = 0;
		if(defined($trafficStatsElement)){
			my $inputPacketsElement = $trafficStatsElement->first_child('input-packets');
			if(defined $inputPacketsElement){
				$inputBytes = $inputPacketsElement->text();
			}
		}

		my $majorInt = {
			name       => $name,
			operStatus => ucfirst($operStatus),
			type       => $type,
			inputBytes => $inputBytes,
		};

		push(@{$interfaces->{interface}}, $majorInt);

		foreach my $logicalInt ($physicalInt->descendants('logical-interface')){

			my $subName;
			if($logicalInt->first_child('name')){
				$subName = $logicalInt->first_child('name')->text();
			}else{
				next;
			}

			my $subInt = {
				name       => $subName,
				operStatus => ucfirst($operStatus),
				type       => $type,
				inputBytes => 0,
			};

			foreach my $fam ($logicalInt->descendants('address-family')){
				foreach my $addr ($fam->descendants('interface-address')){
					my $child = $addr->first_child('ifa-local');
					if($child){
						my $net_ip = new Net::IP($addr->first_child('ifa-local')->text());
						if(defined($net_ip)){

							# TODO add IPv6 support
							if($net_ip->ip() =~ /$ipv6_regex/i){
								$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping sub interface $subName");
								next;
							}
							my $ipConfiguration = {ipAddress => $net_ip->ip()};
							if(defined($addr->first_child('ifa-destination'))){
								my ($mask) = $addr->first_child('ifa-destination')->text() =~ /\/(\d+)/;
								$ipConfiguration->{mask} = $mask;
							}
							push(@{$subInt->{ipEntry}}, $ipConfiguration);
						}
					}
				}
			}
			push(@{$interfaces->{interface}}, $subInt);
		}
	}

	$out->attributes(1);
	$out->print_element('interfaces', $interfaces);
	return $interfaces;
}

1;
